// create.c
// Copyright : 2018-10-30 Yutaka Sawada
// License : GPL

#ifndef _UNICODE
#define _UNICODE
#endif
#ifndef UNICODE
#define UNICODE
#endif
#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0501
#endif

#include <stdio.h>

#include <windows.h>

#include "common2.h"
#include "phmd5.h"
#include "md5_crc.h"
#include "version.h"
#include "create.h"

//#define TIMER // p

// \[gɍڂr
static int sort_cmp(const void *elem1, const void *elem2)
{
	const file_ctx_c *file1, *file2;
	int i;

	file1 = elem1;
	file2 = elem2;

	// TCY 0 ̋t@CtH_ non-recovery set Ɋ܂߂
	if (file1->size == 0){
		if (file2->size > 0)
			return 1;
	} else {
		if (file2->size == 0)
			return -1;
	}

	// File ID ̏ɕёւ
	for (i = 15; i >= 0; i--){
		if (file1->id[i] > file2->id[i]){
			return 1; // 1 > 2
		} else if (file1->id[i] < file2->id[i]){
			return -1; // 1 < 2
		}
	}
	return 0; // 1 = 2
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

// PAR 2.0 ̃pPbgEwb_[쐬
void set_packet_header(
	unsigned char *buf,			// wb_[i[ 64oCg̃obt@[
	unsigned char *set_id,		// Recovery Set IDA16oCg
	int type_num,				// ̃pPbǧ^ԍ
	unsigned int body_size)	// pPbg̃f[^ETCY
{
	// Magic sequence
	memcpy(buf, "PAR2\0PKT", 8);
	// Length of the entire packet
	body_size += 64;	// pPbgŜ̃TCYɂ
	memcpy(buf + 8, &body_size, 4);
	memset(buf + 12, 0, 4);
	// MD5 Hash of packet ͌ŌvZ
	// Recovery Set ID
	memcpy(buf + 32, set_id, 16);
	// Type
	memcpy(buf + 48, "PAR 2.0\0", 8);
	memset(buf + 56, 0, 8);
	switch (type_num){
	case 1:		// Main packet
		memcpy(buf + 56, "Main", 4);
		break;
	case 2:		// File Description packet
		memcpy(buf + 56, "FileDesc", 8);
		break;
	case 3:		// Input File Slice Checksum packet
		memcpy(buf + 56, "IFSC", 4);
		break;
	case 4:		// Recovery Slice packet
		memcpy(buf + 56, "RecvSlic", 8);
		break;
	case 5:		// Creator packet
		memcpy(buf + 56, "Creator", 7);
		break;
	case 10:	// Unicode Filename packet
		memcpy(buf + 56, "UniFileN", 8);
		break;
	case 11:	// ASCII Comment packet
		memcpy(buf + 56, "CommASCI", 8);
		break;
	case 12:	// Unicode Comment packet
		memcpy(buf + 56, "CommUni", 7);
		break;
	}
}

// \[XEt@C̏W߂
int get_source_files(file_ctx_c *files)
{
	unsigned char work_buf[16 + 8 + (MAX_LEN * 3)];
	unsigned char hash0[16] = {0xd4, 0x1d, 0x8c, 0xd9, 0x8f, 0x00, 0xb2, 0x04,
							0xe9, 0x80, 0x09, 0x98, 0xec, 0xf8, 0x42, 0x7e};
	wchar_t file_name[MAX_LEN], file_path[MAX_LEN];
	int i, list_off = 0, num, len;
	WIN32_FILE_ATTRIBUTE_DATA AttrData;

	wcscpy(file_path, base_dir);
	for (num = 0; num < file_num; num++){
		files[num].name = list_off;	// t@C̈ʒuL^
		wcscpy(file_name, list_buf + list_off);
		//list_off += (wcslen(file_name) + 1);
		while (list_buf[list_off] != 0)
			list_off++;
		list_off++;

		// 𒲂ׂ
		wcscpy(file_path + base_len, file_name);
		if (!GetFileAttributesEx(file_path, GetFileExInfoStandard, &AttrData)){
			print_win32_err();
			printf_cp("GetFileAttributesEx, %s\n", file_name);
			return 1;
		}
		if (AttrData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY){	// tH_Ȃ
			memcpy(files[num].hash, hash0, 16);
			files[num].size = 0;
		} else {	// t@CȂ
			files[num].size = ((__int64)AttrData.nFileSizeHigh << 32) | (unsigned __int64)AttrData.nFileSizeLow;
			if (files[num].size == 0){	// ̃t@CȂ
				memcpy(files[num].hash, hash0, 16);
			} else {
				// t@C̐擪 16KB ̃nbVlvZ
				if (file_md5_16(file_path, files[num].hash)){
					printf_cp("file_md5_16, %s\n", file_name);
					return 1;
				}
			}
		}

		// File ID vZ
		memcpy(work_buf, files[num].hash, 16);
		memcpy(work_buf + 16, &(files[num].size), 8);
		unix_directory(file_name);	// L^̃fBNgĹu/vɂ
		utf16_to_utf8(file_name, work_buf + 24);
		len = (int)strlen(work_buf + 24);	// File ID vZ̃t@C̃TCY 4̔{ɂȂ
		data_md5(work_buf, 24 + len, files[num].id);
		// ̃t@CƓ File ID ɂȂĂȂׂ
		do {
			for (i = 0; i < num; i++){
				if (memcmp(files[num].id, files[i].id, 16) == 0){	// File ID Ȃ
					files[num].id[0] += 1;	// ŉʃoCgω
					break;
				}
			}
		} while (i < num);	// File ID Kj[NȒlɂ
	}

	return 0;
}

// ʃpPbg쐬
int set_common_packet(
	unsigned char *buf,
	int *packet_num,			// ʃpPbg̐
	int switch_u,				// jR[h̃t@CL^
	file_ctx_c *files)
{
	char ascii_buf[MAX_LEN * 3];
	unsigned char set_id[16], file_hash[16], *main_packet_buf;
	unsigned char hash0[16] = {0xd4, 0x1d, 0x8c, 0xd9, 0x8f, 0x00, 0xb2, 0x04,
							0xe9, 0x80, 0x09, 0x98, 0xec, 0xf8, 0x42, 0x7e};
	wchar_t file_name[MAX_LEN], file_path[MAX_LEN];
	int i, err, off, off2, num, len, data_size, main_packet_size;
	unsigned int time_last;
	__int64 prog_now = 0;

#ifdef TIMER
unsigned int time_start = GetTickCount();
#endif
	printf("\n");
	print_progress_text(0, "Computing file hash");
	wcscpy(file_path, base_dir);

	// t@C File ID ̏ɕёւ
	qsort(files, file_num, sizeof(file_ctx_c), sort_cmp);

	// ŏ Main packet 쐬Ă
	main_packet_size = 12 + (file_num * 16);
	main_packet_buf = (unsigned char *)malloc(main_packet_size);
	if (main_packet_buf == NULL){
		printf("malloc, %d\n", main_packet_size);
		return 1;
	}
	memcpy(main_packet_buf, &block_size, 4);
	memset(main_packet_buf + 4, 0, 4);
	memcpy(main_packet_buf + 8, &entity_num, 4);
	for (num = 0; num < file_num; num++)	// ёւꂽ File ID Rs[
		memcpy(main_packet_buf + (12 + (16 * num)), files[num].id, 16);
	data_md5(main_packet_buf, main_packet_size, set_id);	// Recovery Set ID vZ
	*packet_num = 1;

	// t@CƂ̃pPbg쐬
	time_last = GetTickCount();
	off = 0;
	for (num = 0; num < file_num; num++){
		// File Description packet
		off2 = off;	// ʒuL^Ă
		memcpy(buf + (off + 64), files[num].id, 16);
		memcpy(buf + (off + 64 + 32), files[num].hash, 16);	// Ŝ̃nbVl͌ŏ
		memcpy(buf + (off + 64 + 48), &(files[num].size), 8);
		// t@C UTF-8 ɕϊ
		wcscpy(file_name, list_buf + files[num].name);
		wcscpy(file_path + base_len, file_name);
		unix_directory(file_name);	// L^̃fBNgĹu/vɂ
		utf16_to_utf8(file_name, ascii_buf);
		len = (int)strlen(ascii_buf);
		for (i = len + 1; (i < len + 4) || (i < MAX_LEN * 3); i++)
			ascii_buf[i] = 0;
		len = (len + 3) & 0xFFFFFFFC;
		memcpy(buf + (off + 64 + 56), ascii_buf, len);
		data_size = 56 + len;
		set_packet_header(buf + off, set_id, 2, data_size);	// pPbgEwb_[쐬
		//data_md5(buf + (off + 32), 32 + data_size, buf + (off + 16));	// pPbg MD5 vZ
		off += (64 + data_size);
		(*packet_num)++;

		if (files[num].size > 0){	// Input File Slice Checksum packet
			// QuickPar ͂ȂTCY 0 ̃t@Cɑ΂Ă`FbNT̃pPbgKvEEE
			// ӖȃpPbg邽߂ non-recovery set g?
			memcpy(buf + (off + 64), files[num].id, 16);
			// t@C MD5 nbVlƃubÑ`FbNT𓯎ɌvZ
			err = file_hash_crc(file_path, files[num].size, file_hash, buf + (off + 64 + 16), &time_last, &prog_now);
			if (err){
				if (err == 1)
					printf_cp("file_hash_crc, %s\n", file_name);
				off = err;
				goto error_end;
			}
			len = (int)((files[num].size + (__int64)block_size - 1) / block_size);
			// t@C MD5-16k ͂svȂ̂ŁAubN CRC-32 ɕύX
			memset(files[num].hash, 0, 16);
			for (i = 0; i < len; i++){	// XOR  16oCgɌ炷
				memcpy(&err, buf + (off + 64 + 16 + 20 * i + 16), 4);
				((unsigned int *)files[num].hash)[i & 3] ^= err;
			}
//			printf("%d: 0x%08X %08X %08X %08X\n", num,
//				((unsigned int *)files[num].hash)[0], ((unsigned int *)files[num].hash)[1],
//				((unsigned int *)files[num].hash)[2], ((unsigned int *)files[num].hash)[3]);
			i = 16 + 20 * len;	// pPbgẽTCY
			set_packet_header(buf + off, set_id, 3, i);	// pPbgEwb_[쐬
			data_md5(buf + (off + 32), 32 + i, buf + (off + 16));	// pPbg MD5 vZ
			off += (64 + i);
			(*packet_num)++;
		} else {
			memcpy(file_hash, hash0, 16);
		}

		// File Description packet ̑
		memcpy(buf + (off2 + 64 + 16), file_hash, 16);	// vZnbVlŏ
		data_md5(buf + (off2 + 32), 32 + data_size, buf + (off2 + 16));	// pPbg MD5 vZ

		if (switch_u != 0){	// jR[h̃t@C
			// w肪Ăt@C ASCII Ȃ烆jR[h̃pPbg͍Ȃ
			len = (int)wcslen(file_name);
			for (i = 0; i < len; i++){
				if (file_name[i] != ascii_buf[i])
					break;
			}
			if (i < len){
				// Unicode Filename packet
				memcpy(buf + (off + 64), files[num].id, 16);
				len = (int)wcslen(file_name) * 2;
				memcpy(buf + (off + 64 + 16), file_name, len);
				if (len & 3)
					memset(buf + (off + 64 + 16 + len), 0, 2);
				data_size = 16 + ((len + 3) & 0xFFFFFFFC);
				set_packet_header(buf + off, set_id, 10, data_size);	// pPbgEwb_[쐬
				data_md5(buf + (off + 32), 32 + data_size, buf + (off + 16));	// pPbg MD5 vZ
				off += (64 + data_size);
				(*packet_num)++;
			}
		}
	}
	print_progress_done();	// sčs̐擪ɖ߂Ă

	// Main packet
	set_packet_header(buf + off, set_id, 1, main_packet_size);	// pPbgEwb_[쐬
	memcpy(buf + (off + 64), main_packet_buf, main_packet_size);
	data_md5(buf + (off + 32), 32 + main_packet_size, buf + (off + 16));	// pPbg MD5 vZ
	off += (64 + main_packet_size);

#ifdef TIMER
time_start = GetTickCount() - time_start;
printf("hash %d.%03d sec\n", time_start / 1000, time_start % 1000);
#endif

error_end:
	free(main_packet_buf);
	return off;
}

// \[XEt@C̏񂩂狤ʃpPbg̃TCYvZ
int measure_common_packet(
	int *packet_num,			// ʃpPbg̐
	int switch_u)				// jR[h̃t@CL^
{
	char ascii_buf[MAX_LEN * 3];
	wchar_t *file_name, file_path[MAX_LEN];
	int i, list_off = 0, off = 0, len, data_size;
	__int64 file_size;
	WIN32_FILE_ATTRIBUTE_DATA AttrData;

	wcscpy(file_path, base_dir);

	// ŏ Main packet 쐬Ă
	data_size = 12 + (file_num * 16);
	off += (64 + data_size);
	*packet_num = 1;

	// Item index for debug output
	if (split_size >= 4){
		printf("\n         Size  Slice Split :  Filename\n");
	} else {
		printf("\n         Size  Slice :  Filename\n");
	}

	// t@CƂ̃pPbg쐬
	while (list_off < list_len){
		file_name = list_buf + list_off;
		//list_off += (wcslen(file_name) + 1);
		while (list_buf[list_off] != 0)
			list_off++;
		list_off++;

		// 𒲂ׂ
		wcscpy(file_path + base_len, file_name);
		if (!GetFileAttributesEx(file_path, GetFileExInfoStandard, &AttrData)){
			print_win32_err();
			printf_cp("GetFileAttributesEx, %s\n", file_name);
			return 1;
		}
		if (AttrData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY){	// tH_Ȃ
			file_size = 0;
		} else {	// t@CȂ
			file_size = ((__int64)AttrData.nFileSizeHigh << 32) | (unsigned __int64)AttrData.nFileSizeLow;
		}
		// Line for debug output
		i = 0;
		if (file_size > 0)
			i = (int)((file_size + block_size - 1) / block_size);	// ubN
		printf("%13I64d %6d", file_size, i);
		if (split_size >= 4){
			i = (int)((file_size + split_size - 1) / split_size);	// 
			printf(" %5d", i);
		}
		printf_cp(" : \"%s\"\n", file_name);

		// File Description packet
		// t@C UTF-8 ɕϊ
		utf16_to_utf8(file_name, ascii_buf);
		len = (int)strlen(ascii_buf);
		len = (len + 3) & 0xFFFFFFFC;
		data_size = 56 + len;
		off += (64 + data_size);
		(*packet_num)++;

		if (file_size > 0){	// Input File Slice Checksum packet
			data_size = 16;
			data_size += (int)(((file_size + block_size - 1) / block_size) * 20);
			off += (64 + data_size);
			(*packet_num)++;
		}

		if (switch_u != 0){	// jR[h̃t@C
			// w肪Ăt@C ASCII Ȃ烆jR[h̃pPbg͍Ȃ
			len = (int)wcslen(file_name);
			for (i = 0; i < len; i++){
				if (file_name[i] != ascii_buf[i]){
					break;
				}
			}
			if (i < len){
				// Unicode Filename packet
				len = (int)wcslen(file_name) * 2;
				data_size = 16 + ((len + 3) & 0xFFFFFFFC);
				off += (64 + data_size);
				(*packet_num)++;
			}
		}
	}

	return off;
}

// pPbg쐬
int set_footer_packet(
	unsigned char *buf,
	wchar_t *par_comment,		// Rg
	unsigned char *set_id)		// Recovery Set ID
{
	unsigned char ascii_buf[COMMENT_LEN * 3];
	int off = 0, len, data_size;

	if (par_comment[0] != 0){	// ASCII  Unicode ̂ǂ炩L^
		data_size = 0;
		// 7-bit ASCII ȊO݂邩ǂ𒲂ׂ
		len = 0;
		while (par_comment[len] != 0){
			if ((par_comment[len] & 0xFF80) != 0){
				data_size = 3;
				break;
			}
			len++;
		}
		if (data_size == 0){
			// ASCII ւ̕ϊɎsꍇ̓jR[hEpPbgL^
			if (!WideCharToMultiByte(CP_ACP, 0, par_comment, -1, ascii_buf, COMMENT_LEN * 3, NULL, &len) || len){
				data_size = 1;
			} else if (wcslen(par_comment) != strlen(ascii_buf)){	// ƃoCgvȂȂ
				data_size = 2;
			}
		}
		if (data_size == 0){
			// ASCII Comment packet
			len = (int)strlen(ascii_buf);
			memcpy(buf + (off + 64), ascii_buf, len);
			if (len & 3)
				memset(buf + (off + 64 + len), 0, 3);
			data_size = (len + 3) & 0xFFFFFFFC;
			set_packet_header(buf + off, set_id, 11, data_size);	// pPbgEwb_[쐬
			data_md5(buf + (off + 32), 32 + data_size, buf + (off + 16));	// pPbg MD5 vZ
			off += (64 + data_size);
		} else {
			// Unicode Comment packet
			memset(buf + (off + 64), 0, 16);
			len = (int)wcslen(par_comment) * 2;
			memcpy(buf + (off + 64 + 16), par_comment, len);
			if (len & 3)
				memset(buf + (off + 64 + 16 + len), 0, 2);
			data_size = 16 + ((len + 3) & 0xFFFFFFFC);
			set_packet_header(buf + off, set_id, 12, data_size);	// pPbgEwb_[쐬
			data_md5(buf + (off + 32), 32 + data_size, buf + (off + 16));	// pPbg MD5 vZ
			off += (64 + data_size);
		}
	}

	// Creator packet
	sprintf(ascii_buf, "par2j v%x.%x.%x\0", CLIENT_VERSION >> 12,
		(CLIENT_VERSION >> 8) & 0xF, (CLIENT_VERSION >> 4) & 0xF);
	len = (int)strlen(ascii_buf);
	memcpy(buf + (off + 64), ascii_buf, len);
	if (len & 3)
		memset(buf + (off + 64 + len), 0, 3);
	data_size = (len + 3) & 0xFFFFFFFC;
	//strcpy(buf + (off + 64), "par2j v1.1.6\0");
	//data_size = 12;
	set_packet_header(buf + off, set_id, 5, data_size);	// pPbgEwb_[쐬
	data_md5(buf + (off + 32), 32 + data_size, buf + (off + 16));	// pPbg MD5 vZ
	off += (64 + data_size);

	return off;
}

// pPbg̃TCYvZ
int measure_footer_packet(
	wchar_t *par_comment)		// Rg
{
	unsigned char ascii_buf[COMMENT_LEN * 3];
	int off = 0, len, data_size;

	if (par_comment[0] != 0){	// ASCII  Unicode ̂ǂ炩L^
		data_size = 0;
		// ASCII ւ̕ϊɎsꍇ̓jR[hEpPbgL^
		if (!WideCharToMultiByte(CP_ACP, 0, par_comment, -1, ascii_buf, COMMENT_LEN * 3, NULL, &len) || len){
			data_size = 1;
		} else if (wcslen(par_comment) != strlen(ascii_buf)){	// ƃoCgvȂȂ
			data_size = 2;
		}

		if (data_size == 0){
			// ASCII Comment packet
			len = (int)strlen(ascii_buf);
			data_size = (len + 3) & 0xFFFFFFFC;
			off += (64 + data_size);
		} else {
			// Unicode Comment packet
			len = (int)wcslen(par_comment) * 2;
			data_size = 16 + ((len + 3) & 0xFFFFFFFC);
			off += (64 + data_size);
		}
	}

	// Creator packet
	//len = strlen("par2j v*.*.*\0");
	//data_size = (len + 3) & 0xFFFFFFFC;
	off += (64 + 12);

	return off;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

// {[ԍ̌߂
static int calc_max_num(
	int block_distri,	// peBEubN̕z@ (3-bitڂ͔ԍ̕t)
	int *max_num2)		// őXCX̌
{
	int i, j, num;
	int exp_num, block_start, block_count, block_start_max, block_count_max = 0;

	if (block_distri >> 2){	// t@CԍɂAvol_1, vol_2, vol_3, ...
		i = recovery_num;
		for (j = 1; j < 10; j++){
			i /= 10;
			if (i == 0)
				break;
		}
		return j;	// t@Č
	}

	block_start = 0;
	block_count = 0;
	exp_num = 1;
	for (num = 0; num < recovery_num; num++){
		// JoEt@C̃t@C
		switch (block_distri & 3){
		case 0:	// Ȃ
			block_count = parity_num / recovery_num;
			j = parity_num % recovery_num;	// ؂Ȃꍇ
			if ((j > 0) && (num < j))
				block_count++;
			break;
		case 1:	// {XňقȂ鐔ɂ
			if (num == recovery_num - 1){	// Ō̃t@Cɗ]
				block_count = parity_num - block_start;
			} else {
				exp_num = recovery_limit & 0xFFFF;
				if (num >= 16){
					block_count = exp_num;
				} else {
					block_count = ((unsigned int)recovery_limit >> 16) << num;
					if (block_count > exp_num)
						block_count = exp_num;
				}
			}
			break;
		case 2:	// 1,2,4,8,16 2̏搔Ȃ
			block_count = exp_num;
			if (block_count >= recovery_limit){
				block_count = recovery_limit;
			} else {
				exp_num *= 2;
			}
			break;
		case 3:	// 1,1,2,5,10,10,20,50 Ƃ decimal weights sizing scheme Ȃ
			block_count = exp_num;
			if (block_count >= recovery_limit){
				block_count = recovery_limit;
			} else {
				switch (num % 4){
				case 1:
				case 3:
					exp_num = exp_num * 2;
					break;
				case 2:
					exp_num = (exp_num / 2) * 5;
					break;
				}
			}
			break;
		}
		if (block_start + block_count > parity_num)
			block_count = parity_num - block_start;
		if (block_count_max < block_count)
			block_count_max = block_count;
		block_start += block_count;
	}
	block_start_max = first_num + block_start - block_count;	// Ōɑ
	for (j = 1; j < 10; j++){
		block_start_max /= 10;
		if (block_start_max == 0)
			break;
	}
	block_start_max = j;
	for (j = 1; j < 10; j++){
		block_count_max /= 10;
		if (block_count_max == 0)
			break;
	}
	*max_num2 = j;

	return block_start_max;
}

// _ɏނɂ͊ǗҌKv
//#define RANDOM_ACCESS

// JoEt@C쐬ċʃpPbgRs[
int create_recovery_file(
	wchar_t *recovery_path,		// Ɨp
	int packet_limit,			// JoEt@C̃pPbgJԂ̐
	int block_distri,			// peBEubN̕z@ (3-bitڂ͔ԍ̕t)
	int packet_num,				// ʃpPbg̐
	unsigned char *common_buf,	// ʃpPbg̃obt@[
	int common_size,			// ʃpPbg̃obt@[ETCY
	unsigned char *footer_buf,	// pPbg̃obt@[
	int footer_size,			// pPbg̃obt@[ETCY
	HANDLE *rcv_hFile,			// eJoEt@C̃nh
	parity_ctx_c *p_blk)		// epeBEubN̏
{
	wchar_t recovery_base[MAX_LEN], file_ext[EXT_LEN], *tmp_p;
	int i, j, num;
	int exp_num, block_start, block_count, block_start_max, block_count_max;
	int repeat_max, packet_to, packet_from, packet_size, common_off;
	unsigned int time_last, prog_num = 0;
	__int64 file_off;
#ifdef RANDOM_ACCESS
	HANDLE hToken;
#endif

	print_progress_text(0, "Constructing recovery file");
	time_last = GetTickCount();

#ifdef RANDOM_ACCESS
	// SE_MANAGE_VOLUME_NAME Lɂ
	if (OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken) != 0){
		LUID luid;
		if (LookupPrivilegeValue(NULL, SE_MANAGE_VOLUME_NAME, &luid ) != 0){
			TOKEN_PRIVILEGES tp;
			tp.PrivilegeCount = 1;
			tp.Privileges[0].Luid = luid;
			tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
			AdjustTokenPrivileges(hToken, FALSE, &tp, sizeof(TOKEN_PRIVILEGES), (PTOKEN_PRIVILEGES)NULL, (PDWORD)NULL);
			if (GetLastError() != ERROR_SUCCESS){
				print_win32_err();
			} else {
				printf("\nSE_MANAGE_VOLUME_NAME: ok\n");
			}
		}
		CloseHandle(hToken);
	}
#endif

	// JoEt@C̊gqɂ͎w肳ꂽ̂g
	file_ext[0] = 0;
	wcscpy(recovery_base, recovery_file);
	tmp_p = offset_file_name(recovery_base);
	tmp_p = wcsrchr(tmp_p, '.');
	if (tmp_p != NULL){
		if (wcslen(tmp_p) < EXT_LEN){
			wcscpy(file_ext, tmp_p);	// gqL^Ă
			*tmp_p = 0;	// gq菜
		}
	}

	// {[ԍ̌߂
	block_start_max = calc_max_num(block_distri, &block_count_max);

	// JoEt@C쐬ċʃpPbg
	block_start = 0;
	block_count = 0;
	exp_num = 1;
	for (num = 0; num < recovery_num; num++){
		// JoEt@C̃t@C
		switch (block_distri & 3){
		case 0:	// Ȃ
			block_count = parity_num / recovery_num;
			j = parity_num % recovery_num;	// ؂Ȃꍇ
			if ((j > 0) && (num < j))
				block_count++;
			break;
		case 1:	// {XňقȂ鐔ɂ
			if (num == recovery_num - 1){	// Ō̃t@Cɗ]
				block_count = parity_num - block_start;
			} else {
				exp_num = recovery_limit & 0xFFFF;
				if (num >= 16){
					block_count = exp_num;
				} else {
					block_count = ((unsigned int)recovery_limit >> 16) << num;
					if (block_count > exp_num)
						block_count = exp_num;
				}
			}
			break;
		case 2:	// 1,2,4,8,16 2̏搔Ȃ
			block_count = exp_num;
			if (block_count >= recovery_limit){
				block_count = recovery_limit;
			} else {
				exp_num *= 2;
			}
			break;
		case 3:	// 1,1,2,5,10,10,20,50 Ƃ decimal weights sizing scheme Ȃ
			block_count = exp_num;
			if (block_count >= recovery_limit){
				block_count = recovery_limit;
			} else {
				switch (num % 4){
				case 1:
				case 3:
					exp_num = exp_num * 2;
					break;
				case 2:
					exp_num = (exp_num / 2) * 5;
					break;
				}
			}
			break;
		}
		if (block_start + block_count > parity_num)
			block_count = parity_num - block_start;
		if (block_distri >> 2){
			// t@CԍɂAvol_1, vol_2, vol_3, ...
			swprintf(recovery_path, MAX_LEN, L"%s.vol_%0*d%s", recovery_base, block_start_max, num + 1, file_ext);
		} else {
			// QuickParAvolXX+YY
			swprintf(recovery_path, MAX_LEN, L"%s.vol%0*d+%0*d%s", recovery_base, block_start_max,
				first_num + block_start, block_count_max, block_count, file_ext);
		}
		// JoEt@CJ
		//move_away_file(recovery_path);	// ̃t@Cǂ
		rcv_hFile[num] = CreateFile(recovery_path, GENERIC_READ | GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (rcv_hFile[num] == INVALID_HANDLE_VALUE){
			print_win32_err();
			rcv_hFile[num] = NULL;
			printf_cp("cannot create file, %s\n", recovery_path);
			return 1;
		}

		// pPbǧJԂ񐔂vZ
		repeat_max = 1;
		for (j = 2; j <= block_count; j *= 2)	// JԂ񐔂 log2(block_count)
			repeat_max++;
		if ((packet_limit > 0) && (repeat_max > packet_limit))
			repeat_max = packet_limit;	// JԂ񐔂𐧌
		// JoEt@C̑傫vZ
		file_off = (__int64)(68 + block_size) * block_count;
		file_off += (common_size * repeat_max) + footer_size;
		// ̃TCYɂ
		if (!SetFilePointerEx(rcv_hFile[num], *((PLARGE_INTEGER)&file_off), NULL, FILE_BEGIN)){
			print_win32_err();
			return 1;
		}
		if (!SetEndOfFile(rcv_hFile[num])){
			print_win32_err();
			return 1;
		}
#ifdef RANDOM_ACCESS
		if (!SetFileValidData(rcv_hFile[num], file_off)){
			print_win32_err();
			return 1;
		}
#endif

		file_off = 0;
		repeat_max *= packet_num;	// JoEt@C̋ʃpPbg̐
		packet_from = 0;
		common_off = 0;

		// Recovery Slice packet ͌ォ珑
		for (j = block_start; j < block_start + block_count; j++){
			prog_num++;
			// oߕ\
			if (GetTickCount() - time_last >= UPDATE_TIME){
				if (print_progress((prog_num * 1000) / parity_num))
					return 2;
				time_last = GetTickCount();
			}

			// Recovery Slice packet
			file_off += 68;
			p_blk[j].file = num;
			p_blk[j].off = file_off;	// JnʒuL^Ă
			file_off += block_size;

			// ǂꂾ̋ʃpPbgނ
			packet_size = 0;
			packet_to = (int)((__int64)repeat_max * (j - block_start + 1) / block_count);
			while (packet_to - packet_from > 0){
				memcpy(&i, common_buf + (common_off + (packet_size + 8)), 4);	// ̃pPbg̃f[^ETCY𒲂ׂ
				packet_size += i;
				packet_from++;
			}
			if (packet_size > 0){	// ʃpPbg
				if (file_write_data(rcv_hFile[num], file_off, common_buf + common_off, packet_size)){
					printf_cp("file_write_data, %s\n", recovery_path);
					return 1;
				}
				// ItZbg𒴂߂Ă
				common_off += packet_size;
				if (common_off >= common_size)
					common_off -= common_size;
				file_off += packet_size;
			}
		}
		block_start += block_count;

		// pPbg
		if (file_write_data(rcv_hFile[num], file_off, footer_buf, footer_size)){
			printf_cp("file_write_data, %s\n", recovery_path);
			return 1;
		}
	}
	print_progress_done();	// sčs̐擪ɖ߂Ă

	return 0;
}

// 쐬̃JoEt@C폜
void delete_recovery_file(
	wchar_t *recovery_path,		// Ɨp
	int block_distri,			// peBEubN̕z@ (3-bitڂ͔ԍ̕t)
	int switch_p,				// CfbNXEt@CȂ
	HANDLE *rcv_hFile)			// eJoEt@C̃nh
{
	wchar_t recovery_base[MAX_LEN], file_ext[EXT_LEN], *tmp_p;
	int j, num;
	int exp_num, block_start, block_count, block_start_max, block_count_max;

	if (switch_p == 0)
		DeleteFile(recovery_file);	// CfbNXEt@C폜

	// JoEt@C̊gqɂ͎w肳ꂽ̂g
	file_ext[0] = 0;
	wcscpy(recovery_base, recovery_file);
	tmp_p = offset_file_name(recovery_base);
	tmp_p = wcsrchr(tmp_p, '.');
	if (tmp_p != NULL){
		if (wcslen(tmp_p) < EXT_LEN){
			wcscpy(file_ext, tmp_p);	// gqL^Ă
			*tmp_p = 0;	// gq菜
		}
	}

	// {[ԍ̌߂
	block_start_max = calc_max_num(block_distri, &block_count_max);

	// JoEt@C폜
	block_start = 0;
	block_count = 0;
	exp_num = 1;
	for (num = 0; num < recovery_num; num++){
		// JoEt@C̃t@C
		switch (block_distri & 3){
		case 0:	// Ȃ
			block_count = parity_num / recovery_num;
			j = parity_num % recovery_num;	// ؂Ȃꍇ
			if ((j > 0) && (num < j))
				block_count++;
			break;
		case 1:	// {XňقȂ鐔ɂ
			if (num == recovery_num - 1){	// Ō̃t@Cɗ]
				block_count = parity_num - block_start;
			} else {
				exp_num = recovery_limit & 0xFFFF;
				if (num >= 16){
					block_count = exp_num;
				} else {
					block_count = ((unsigned int)recovery_limit >> 16) << num;
					if (block_count > exp_num)
						block_count = exp_num;
				}
			}
			break;
		case 2:	// 1,2,4,8,16 2̏搔Ȃ
			block_count = exp_num;
			if (block_count >= recovery_limit){
				block_count = recovery_limit;
			} else {
				exp_num *= 2;
			}
			break;
		case 3:	// 1,1,2,5,10,10,20,50 Ƃ decimal weights sizing scheme Ȃ
			block_count = exp_num;
			if (block_count >= recovery_limit){
				block_count = recovery_limit;
			} else {
				switch (num % 4){
				case 1:
				case 3:
					exp_num = exp_num * 2;
					break;
				case 2:
					exp_num = (exp_num / 2) * 5;
					break;
				}
			}
			break;
		}
		if (block_start + block_count > parity_num)
			block_count = parity_num - block_start;
		if (block_distri >> 2){
			// t@CԍɂAvol_1, vol_2, vol_3, ...
			swprintf(recovery_path, MAX_LEN, L"%s.vol_%0*d%s", recovery_base, block_start_max, num + 1, file_ext);
		} else {
			// QuickParAvolXX+YY
			swprintf(recovery_path, MAX_LEN, L"%s.vol%0*d+%0*d%s", recovery_base, block_start_max,
				first_num + block_start, block_count_max, block_count, file_ext);
		}
		block_start += block_count;

		if (rcv_hFile[num] != NULL){	// t@C쐬ς݂Ȃ (JĂ)
			CloseHandle(rcv_hFile[num]);
			rcv_hFile[num] = NULL;
			DeleteFile(recovery_path);	// r܂ł̃JoEt@C폜
		}
	}
}

// JoEt@C̃TCYvZ
void measure_recovery_file(
	wchar_t *recovery_path,		// Ɨp (ŏ̓RgĂ)
	int packet_limit,			// JoEt@C̃pPbgJԂ̐
	int block_distri,			// peBEubN̕z@
	int packet_num,				// ʃpPbg̐
	int common_size,			// ʃpPbg̃obt@[ETCY
	int footer_size,			// pPbg̃obt@[ETCY
	int switch_p)				// CfbNXEt@CȂ
{
	char ascii_buf[MAX_LEN * 3];
	wchar_t recovery_base[MAX_LEN], file_ext[EXT_LEN], *tmp_p;
	int j, num;
	int exp_num, block_start, block_count, block_start_max, block_count_max;
	int footer_num, packet_count, repeat_max;
	__int64 file_size;

	get_file_name(recovery_file, recovery_base);	// t@Cɂ
	footer_num = 1;
	if (recovery_path[0] != 0)
		footer_num++;

	total_file_size = 0;	// JoEt@C̍vTCYvZ
	if (switch_p == 0){
		// Index File ɕ\
		file_size = common_size + footer_size;
		total_file_size += file_size;
		packet_count = packet_num + footer_num;
		utf16_to_cp(recovery_base, ascii_buf, cp_output);
		printf("%13I64d %7d      0 : \"%s\"\n", file_size, packet_count, ascii_buf);
	}
	if (parity_num == 0)	// peBEubNȂꍇ͂ŏI
		return;

	// JoEt@C̊gqɂ͎w肳ꂽ̂g
	file_ext[0] = 0;
	tmp_p = offset_file_name(recovery_base);
	tmp_p = wcsrchr(tmp_p, '.');
	if (tmp_p != NULL){
		if (wcslen(tmp_p) < EXT_LEN){
			wcscpy(file_ext, tmp_p);	// gqL^Ă
			*tmp_p = 0;	// gq菜
		}
	}

	// {[ԍ̌߂
	block_start_max = calc_max_num(block_distri, &block_count_max);

	// JoEt@C쐬ċʃpPbg
	block_start = 0;
	block_count = 0;
	exp_num = 1;
	for (num = 0; num < recovery_num; num++){
		// JoEt@C̃t@C
		switch (block_distri & 3){
		case 0:	// Ȃ
			block_count = parity_num / recovery_num;
			j = parity_num % recovery_num;	// ؂Ȃꍇ
			if ((j > 0) && (num < j))
				block_count++;
			break;
		case 1:	// {XňقȂ鐔ɂ
			if (num == recovery_num - 1){	// Ō̃t@Cɗ]
				block_count = parity_num - block_start;
			} else {
				exp_num = recovery_limit & 0xFFFF;
				if (num >= 16){
					block_count = exp_num;
				} else {
					block_count = ((unsigned int)recovery_limit >> 16) << num;
					if (block_count > exp_num)
						block_count = exp_num;
				}
			}
			break;
		case 2:	// 1,2,4,8,16 2̏搔Ȃ
			block_count = exp_num;
			if (block_count >= recovery_limit){
				block_count = recovery_limit;
			} else {
				exp_num *= 2;
			}
			break;
		case 3:	// 1,1,2,5,10,10,20,50 Ƃ decimal weights sizing scheme Ȃ
			block_count = exp_num;
			if (block_count >= recovery_limit){
				block_count = recovery_limit;
			} else {
				switch (num % 4){
				case 1:
				case 3:
					exp_num = exp_num * 2;
					break;
				case 2:
					exp_num = (exp_num / 2) * 5;
					break;
				}
			}
			break;
		}
		if (block_start + block_count > parity_num)
			block_count = parity_num - block_start;
		if (block_distri >> 2){
			// t@CԍɂAvol_1, vol_2, vol_3, ...
			swprintf(recovery_path, MAX_LEN, L"%s.vol_%0*d%s", recovery_base, block_start_max, num + 1, file_ext);
		} else {
			// QuickParAvolXX+YY
			swprintf(recovery_path, MAX_LEN, L"%s.vol%0*d+%0*d%s", recovery_base, block_start_max,
				first_num + block_start, block_count_max, block_count, file_ext);
		}
		utf16_to_cp(recovery_path, ascii_buf, cp_output);

		// pPbǧJԂ񐔂vZ
		repeat_max = 1;
		for (j = 2; j <= block_count; j *= 2)	// JԂ񐔂 log2(block_count)
			repeat_max++;
		if ((packet_limit > 0) && (repeat_max > packet_limit))
			repeat_max = packet_limit;	// JԂ񐔂𐧌
		// JoEt@C̑傫vZ
		file_size = (__int64)(68 + block_size) * block_count;
		file_size += (common_size * repeat_max) + footer_size;
		total_file_size += file_size;
		packet_count = repeat_max * packet_num;	// JoEt@C̋ʃpPbg̐
		packet_count += footer_num + block_count;
		printf("%13I64d %7d %6d : \"%s\"\n", file_size, packet_count, block_count, ascii_buf);

		block_start += block_count;
	}
}

// \[XEt@C𕪊 (TCY split_size Ŏw肷)
int split_files(
	file_ctx_c *files,
	int *cur_num, int *cur_id)	// G[́A̎_ł̃t@Cԍƕԍ߂
{
	unsigned char buf[IO_SIZE];
	wchar_t file_path[MAX_LEN], recovery_dir[MAX_LEN];
	wchar_t *file_name, *file_ext;
	int num, id, split_num, split_max, total_num, dir_len;
	int name_len, ext_len, num2;
	unsigned int rv, len, split_left;
	unsigned int time_last, prog_num = 0;
	__int64 num8, file_left;
	HANDLE hFile, hFile_src;

	// t@Cǂɕۑ邩
	get_base_dir(recovery_file, recovery_dir);
	if (compare_directory(base_dir, recovery_dir) == 0){	// 悪ꏊȂRs[Ȃ
		recovery_dir[0] = 0;
		dir_len = base_len;
	} else {
		dir_len = (int)wcslen(recovery_dir);
		//printf_cp("\n save_path = %s \n", recovery_dir);
	}

	// ܂ł邩m߂
	total_num = 0;
	for (num = 0; num < file_num; num++){
		if (files[num].size > (__int64)split_size){
			num8 = (files[num].size + (__int64)split_size - 1) / split_size;
			if (num8 > 99999){	//  99999𒴂G[
				printf("too many split file, %d\n", num);
				*cur_num = -1;
				*cur_id = 0;
				return 1;
			}
			total_num += (int)num8;
		} else if (recovery_dir[0] != 0){	// 悪قȂꍇRs[
			total_num++;
		}
	}
	//printf("split total num = %d\n", total_num);
	if (total_num == 0)
		return 0;	// KvȂ

	// t@CA\[XEt@C㏑Ȃ`FbN
	for (num = 0; num < file_num; num++){
		// gq{
		ext_len = 0;
		file_name = list_buf + files[num].name;
		file_ext = wcsrchr(file_name, '.');
		if (file_ext != NULL){
			name_len = (int)(file_ext - file_name);
			ext_len = (int)wcslen(file_ext);
			if ((ext_len >= 4) && (ext_len <= 6)){	// Œ ".99999" ɂȂ
				split_num = 0;
				for (id = 1; id < ext_len; id++){
					if ((file_ext[id] < '0') || (file_ext[id] > '9')){
						ext_len = 0;
						break;
					}
					split_num = split_num * 10 + (file_ext[id] - '0');
				}
				// 悪قȂꍇ́A\[XEt@C㏑Ȃ
				if ((ext_len > 0) && (recovery_dir[0] != 0) && (files[num].size > split_size))
					ext_len = 0;
			} else {
				ext_len = 0;
			}
		}
		if (ext_len > 0){	// SĐ̊gq\[XEt@CȂ
			//printf_cp("\n risky name = %s \n", file_name);
			for (num2 = 0; num2 < file_num; num2++){
				if (num2 == num)
					continue;
				if (_wcsnicmp(list_buf + files[num2].name, file_name, name_len) == 0){
					//printf_cp(" match name = %s \n", list_buf + files[num2].name);
					num8 = (files[num2].size + (__int64)split_size - 1) / split_size;
					split_max = (int)num8;
					//printf("split_num = %d, split max = %d, ext_len = %d \n", split_num, split_max, ext_len);
					// IɖȂΏO
					if ((split_max == 1) || (split_num > split_max))
						continue;	// TCYȃt@ĆÅgqtȂ
					if (((split_max < 1000) && (ext_len >= 5)) || ((split_max < 10000) && (ext_len >= 6)))
						continue;	// gq̌قȂ
					// ㏑댯̂ŃG[ɂ
					printf_cp("split bad file, %s\n", file_name);
					*cur_num = -1;
					*cur_id = 0;
					return 1;
				}
			}
		}
	}

	// TCY傫ȃt@C
	printf("\n");
	print_progress_text(0, "Spliting file");
	time_last = GetTickCount();
	for (num = 0; num < file_num; num++){
		if (files[num].size > (__int64)split_size){	// 
			file_left = files[num].size;
			num8 = (file_left + (__int64)split_size - 1) / split_size;
			split_num = (int)num8;
			if (split_num <= 999){
				split_max = 3;	// Wł 001`999  999܂ł 3
			} else if (split_num <= 9999){
				split_max = 4;
			} else {	// \[XEubN 32768ȉ番ȉ̂͂
				split_max = 5;
			}

			// \[XEt@CJ
			wcscpy(file_path, base_dir);
			wcscpy(file_path + base_len, list_buf + files[num].name);
			hFile_src = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
			if (hFile_src == INVALID_HANDLE_VALUE){
				print_win32_err();
				printf_cp("cannot open file, %s\n", list_buf + files[num].name);
				*cur_num = num;
				*cur_id = 0;
				return 1;
			}
			if (recovery_dir[0] != 0)
				wcscpy(file_path, recovery_dir);	// قȂꏊɕۑ
			for (id = 1; id <= split_num; id++){
				swprintf(file_path + dir_len, _countof(file_path) - dir_len, L"%s.%0*d", list_buf + files[num].name, split_max, id);
				hFile = CreateFile(file_path, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
				if (hFile == INVALID_HANDLE_VALUE){
					print_win32_err();
					CloseHandle(hFile_src);
					printf_cp("cannot create file, %s\n", file_path);
					*cur_num = num;
					*cur_id = id - 1;
					return 1;
				}
				// t@C𕪊TCYƂɏł
				split_left = split_size;
				if ((__int64)split_left > file_left)
					split_left = (unsigned int)file_left;
				file_left -= split_left;
				while (split_left){
					len = IO_SIZE;
					if (split_left < IO_SIZE)
						len = split_left;
					split_left -= len;
					if (!ReadFile(hFile_src, buf, len, &rv, NULL) || (len != rv)){
						print_win32_err();
						CloseHandle(hFile_src);
						CloseHandle(hFile);
						printf("ReadFile, input file %d\n", num);
						*cur_num = num;
						*cur_id = id;
						return 1;
					}
					if (!WriteFile(hFile, buf, len, &rv, NULL)){
						print_win32_err();
						CloseHandle(hFile_src);
						CloseHandle(hFile);
						printf("WriteFile, split file %d.%03d", num, id);
						*cur_num = num;
						*cur_id = id;
						return 1;
					}
				}
				CloseHandle(hFile);

				// oߕ\
				prog_num++;
				if (GetTickCount() - time_last >= UPDATE_TIME){
					if (print_progress((prog_num * 1000) / total_num)){
						CloseHandle(hFile_src);
						*cur_num = num;
						*cur_id = id;
						return 2;
					}
					time_last = GetTickCount();
				}
			}
			CloseHandle(hFile_src);

		} else if (recovery_dir[0] != 0){	// ȃt@C͂̂܂܃Rs[
			// \[XEt@C̃pX
			wcscpy(file_path, base_dir);
			wcscpy(file_path + base_len, list_buf + files[num].name);
			// Rs[̃pX
			wcscpy(recovery_dir + dir_len, list_buf + files[num].name);
			rv = CopyFile(file_path, recovery_dir, FALSE);
			if (rv != 0){	// ǂݎpĂ
				len = GetFileAttributes(recovery_dir);
				if ((len & (FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_READONLY)) != 0){
					len &= ~(FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_READONLY);
					rv = SetFileAttributes(recovery_dir, len);
				}
			}
			if (rv == 0){
				print_win32_err();
				printf_cp("cannot copy file, %s\n", recovery_dir);
				*cur_num = num;
				*cur_id = 0;
				return 1;
			}
			recovery_dir[dir_len] = 0;	// fBNgɖ߂Ă

			// oߕ\
			prog_num++;
			if (GetTickCount() - time_last >= UPDATE_TIME){
				if (print_progress((prog_num * 1000) / total_num)){
					*cur_num = num;
					*cur_id = 0;
					return 2;
				}
				time_last = GetTickCount();
			}
		}
	}
	print_progress_done();	// sčs̐擪ɖ߂Ă

	return 0;
}

// ꂽ\[XEt@C폜
void delete_split_files(
	file_ctx_c *files,
	int max_num, int max_id)	// G[̃t@Cԍƕԍ
{
	wchar_t file_path[MAX_LEN], recovery_dir[MAX_LEN];
	int num, id, split_num, split_max, dir_len;
	__int64 num8, file_left;

	// t@Cǂɕۑ邩
	get_base_dir(recovery_file, recovery_dir);
	if (compare_directory(base_dir, recovery_dir) == 0){	// 悪ꏊȂRs[Ȃ
		recovery_dir[0] = 0;
		dir_len = base_len;
	} else {
		dir_len = (int)wcslen(recovery_dir);
	}

	// TCY傫ȃt@C
	for (num = 0; num <= max_num; num++){
		if (files[num].size > (__int64)split_size){	// ꂽt@C폜
			file_left = files[num].size;
			num8 = (file_left + (__int64)split_size - 1) / split_size;
			split_num = (int)num8;	//  99999 ȉ̂͂
			if (split_num <= 999){
				split_max = 3;	// Wł 001`999  999܂ł 3
			} else if (split_num <= 9999){
				split_max = 4;
			} else if (split_num <= 99999){
				split_max = 5;
			}
			if (num == max_num)
				split_num = max_id;	// f܂ō폜
			for (id = 1; id <= split_num; id++){
				if (recovery_dir[0] != 0){
					wcscpy(file_path, recovery_dir);	// قȂꏊɕۑ
				} else {
					wcscpy(file_path, base_dir);
				}
				swprintf(file_path + dir_len, _countof(file_path) - dir_len, L"%s.%0*d", list_buf + files[num].name, split_max, id);
				if (!DeleteFile(file_path))	// ɕĂ\[XEt@C폜
					return;	// 폜Ɏso
			}

		} else if (recovery_dir[0] != 0){	// ʂ̏ꏊɃRs[ꍇ폜
			wcscpy(recovery_dir + dir_len, list_buf + files[num].name);	// Rs[̃pX
			if (!DeleteFile(recovery_dir))	// Rs[\[XEt@C폜
				return;	// 폜Ɏso
			recovery_dir[dir_len] = 0;	// fBNgɖ߂Ă
		}
	}
}

// JoEt@C\[XEt@C̖ɂ
// tH[}bgɂĂ̓\[XEt@CgȂȂ鋰ꂪ̂ŁAH
// OǂݍłXg[`ƖȂAtb^[ɓǂތ`Ɩ
// 7z ɂ͖ȂAZIP ɂ͖HvȂƂȂB
// bz2  gz  7-Zip œWJɃG[\EEE
// EXE  DLL t@C̓`FbNTς̂ŁAǂݍ݃G[邩B
// JPEG, GIF, PNG, MP3 Ȃǂ͖ȂۂǁAJoER[h𖄂ߍޕKvB
int append_recovery_file(
	wchar_t *file_name,	// \[XEt@C̖O
	HANDLE *rcv_hFile,	// eJoEt@C̃nh
	int no_index)		// CfbNXEt@C݂Ȃ
{
	unsigned char buf[IO_SIZE], header[22];
	wchar_t file_path[MAX_LEN], save_path[MAX_LEN];
	int num;
	unsigned int rv, len, attach_flag;
	unsigned int time_last;
	HANDLE hFileWrite = NULL, hFileRead = NULL;
	LARGE_INTEGER qwi, orig_size;	// Quad Word Integer

	// t@Cǂɕۑ邩
	attach_flag = 0;
	printf("\n");
	print_progress_text(0, "Appending recovery record");
	time_last = GetTickCount();
	get_base_dir(recovery_file, save_path);
	if (compare_directory(base_dir, save_path) != 0){	// 悪قȂꏊȂRs[
		wcscpy(file_path, base_dir);
		wcscat(file_path, file_name);
		wcscat(save_path, file_name);
		if (CopyFile(file_path, save_path, FALSE) == 0){
			print_win32_err();
			printf_cp("cannot copy file, %s\n", save_path);
			return 1;
		}
		attach_flag = 0x02000000;	// Rs[
	} else {
		wcscpy(save_path, base_dir);
		wcscat(save_path, file_name);
		attach_flag = 0x01000000;	// \[XEt@Cꏊ̈
	}
	// ǂݎpĂ
	len = GetFileAttributes(save_path);
	if ((len & (FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_READONLY)) != 0){
		len &= ~(FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_READONLY);
		SetFileAttributes(save_path, len);
	}

	// \[XEt@CJ
	hFileWrite = CreateFile(save_path, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
	if (hFileWrite == INVALID_HANDLE_VALUE){
		print_win32_err();
		printf_cp("cannot open file, %s\n", save_path);
		return 1;
	}
	// ΉtH[}bgł邱ƂmF
	// 7z = 擪 6-byte  Signature Ŕ肷
	// zip = 擪 4-byte  local file header signature 
	//        22-byte  End of central directory record Ŕ肷
	if (!ReadFile(hFileWrite, buf, 28, &len, NULL) || (len != 28)){
		print_win32_err();
		goto error_end;
	}
	if (((unsigned int *)buf)[0] == 0x04034b50){
		// t@C̖ 22oCgǂݍ
		qwi.QuadPart = -22;
		if (!SetFilePointerEx(hFileWrite, qwi, NULL, FILE_END)){
			print_win32_err();
			goto error_end;
		}
		if (!ReadFile(hFileWrite, header, 22, &len, NULL) || (len != 22)){
			print_win32_err();
			goto error_end;
		}
		if (((unsigned int *)header)[0] == 0x06054b50){
			memcpy(&len, header + 12, 4);	// wb_[ɋL^ĂTCYmF
			memcpy(&rv, header + 16, 4);
			if (rv + len + 22 == total_file_size)
				attach_flag |= 0x10000000;
		}
	} else if ((((unsigned short *)buf)[0] == 0x7A37) && (((unsigned int *)(buf + 2))[0] == 0x1C27AFBC)){
		__int64 size1, size2;
		memcpy(&size1, buf + 12, 8);	// wb_[ɋL^ĂTCYmF
		memcpy(&size2, buf + 20, 8);
		if (32 + size1 + size2 == total_file_size)
			attach_flag |= 0x20000000;
	}
	if ((attach_flag & 0x30000000) == 0){
		printf("\ninvalid archive format\n");
		goto error_end;
	}

	// t@C̈ʒu𖖔ɂ
	qwi.QuadPart = 0;
	if (!SetFilePointerEx(hFileWrite, qwi, &orig_size, FILE_END)){
		print_win32_err();
		goto error_end;
	}
	//printf("qwi.QuadPart = %I64d\n", orig_size.QuadPart);

	if (no_index == 0){	// CfbNXEt@CJ
		hFileRead = CreateFile(recovery_file, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
		if (hFileRead == INVALID_HANDLE_VALUE){
			print_win32_err();
			printf_cp("cannot open file, %s\n", recovery_file);
			goto error_end;
		}

		// ɒǉĂ
		do {
			if (!ReadFile(hFileRead, buf, IO_SIZE, &len, NULL)){
				print_win32_err();
				goto error_end;
			}
			if (len == 0)
				break;
			attach_flag |= len;
			if (!WriteFile(hFileWrite, buf, len, &rv, NULL)){
				print_win32_err();
				goto error_end;
			}
		} while (len > 0);
		CloseHandle(hFileRead);
		hFileRead = NULL;
	}

	// JoEt@CJ
	for (num = 0; num < recovery_num; num++){
		// oߕ\
		if (GetTickCount() - time_last >= UPDATE_TIME){
			if (print_progress(((num + 1) * 1000) / recovery_num)){
				goto error_end;
			}
			time_last = GetTickCount();
		}

		// t@C̈ʒu擪ɂ
		qwi.QuadPart = 0;
		if (!SetFilePointerEx(rcv_hFile[num], qwi, NULL, FILE_BEGIN)){
			print_win32_err();
			goto error_end;
		}

		// ɒǉĂ
		do {
			if (!ReadFile(rcv_hFile[num], buf, IO_SIZE, &len, NULL)){
				print_win32_err();
				goto error_end;
			}
			if (len == 0)
				break;
			attach_flag |= len;
			if (!WriteFile(hFileWrite, buf, len, &rv, NULL)){
				print_win32_err();
				goto error_end;
			}
		} while (len > 0);
	}

	// ZIP ɂȂ疖Ƀwb_[ǉ
	if (attach_flag & 0x10000000){
		if (!WriteFile(hFileWrite, header, 22, &rv, NULL) || (rv != 22)){
			print_win32_err();
			goto error_end;
		}
	}

	print_progress_done();	// sčs̐擪ɖ߂Ă
	CloseHandle(hFileWrite);
	return 0;

error_end:	// G[
	if (attach_flag & 0x02000000){	// Rs[t@C폜
		if (hFileWrite != NULL){
			CloseHandle(hFileWrite);
			hFileWrite = NULL;
		}
		DeleteFile(save_path);
	} else if ((attach_flag & 0x01FFFFFF) > 0x01000000){	// ǉoCg菜
		if (SetFilePointerEx(hFileWrite, orig_size, NULL, FILE_BEGIN) != 0)
			SetEndOfFile(hFileWrite);
	}
	if (hFileWrite != NULL)
		CloseHandle(hFileWrite);
	if (hFileRead != NULL)
		CloseHandle(hFileRead);
	return 1;
}

