// md5_crc.c
// Copyright : 2019-05-02 Yutaka Sawada
// License : GPL

#ifndef _UNICODE
#define _UNICODE
#endif
#ifndef UNICODE
#define UNICODE
#endif
#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0501
#endif

#include <process.h>
#include <stdio.h>

#include <windows.h>

#include "common2.h"
#include "crc.h"
#include "phmd5.h"
#include "md5_crc.h"


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

// oCgz MD5 nbVl߂
void data_md5(
	unsigned char *data_in,	// nbVl߂oCgz
	unsigned int data_len,	// ̓oCg
	unsigned char *hash)	// nbVl 16oCg
{
	PHMD5 ctx;

	Phmd5Begin(&ctx);
	Phmd5Process(&ctx, data_in, data_len);
	Phmd5End(&ctx);
	memcpy(hash, ctx.hash, 16);
}

// ubN MD5 nbVl߂
void data_md5_block(
	unsigned char *data_in,	// nbVl߂oCgz
	unsigned int data_len,	// ̓oCg
	unsigned char *hash)	// nbVl 16oCg
{
	PHMD5 ctx;

	Phmd5Begin(&ctx);
	Phmd5Process(&ctx, data_in, data_len);
	if (data_len < block_size)	// ubNETCY܂ 0Ŗ߂ČvZ
		Phmd5ProcessZero(&ctx, block_size - data_len);
	Phmd5End(&ctx);
	memcpy(hash, ctx.hash, 16);
}

// t@C MD5-16k nbVl߂
int file_md5_16(
	wchar_t *file_path,		// nbVl߂t@C
	unsigned char *hash)	// t@C̐擪 16KB ̃nbVl (16oCg)
{
	unsigned char buf[16384];
	unsigned int len;
	PHMD5 hash_ctx;
	HANDLE hFile;

	// \[XEt@CJ
	hFile = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
		return 1;

	// 擪 16KB ̃nbVlvZ
	Phmd5Begin(&hash_ctx);	// 
	if (!ReadFile(hFile, buf, 16384, &len, NULL)){
		CloseHandle(hFile);
		return 2;
	}
	Phmd5Process(&hash_ctx, buf, len);	// MD5 vZ
	Phmd5End(&hash_ctx);	// ŏI
	memcpy(hash, hash_ctx.hash, 16);

	CloseHandle(hFile);
	return 0;
}

// t@C̎w蕔 MD5 nbVl߂
int file_md5(
	HANDLE hFileRead,		// nbVl߂t@C̃nh
	__int64 offset,
	unsigned int data_len,	// ̓oCg
	unsigned char *hash)	// nbVl (16oCg)
{
	unsigned char buf[IO_SIZE];
	unsigned int rv, len;
	PHMD5 hash_ctx;

	// t@C̊Jnʒu offsetoCgڂɂ
	if (!SetFilePointerEx(hFileRead, *((PLARGE_INTEGER)&offset), NULL, FILE_BEGIN))
		return 1;

	Phmd5Begin(&hash_ctx);	// 

	while (data_len > 0){
		len = IO_SIZE;
		if (data_len < IO_SIZE)
			len = data_len;
		if (!ReadFile(hFileRead, buf, len, &rv, NULL) || (len != rv))
			return 2;
		data_len -= len;

		Phmd5Process(&hash_ctx, buf, len);	// MD5 vZ
	}
	Phmd5End(&hash_ctx);	// ŏI
	memcpy(hash, hash_ctx.hash, 16);

	return 0;
}

/*------------------  ȉ MD5  CRC-32 ̑g֐  ------------------*/

// eS 0 ̃f[^ MD5 nbVl CRC-32 ߂
void data_md5_crc32_zero(
	unsigned char *hash)	// nbVl (16 + 4oCg, MD5 + CRC-32)
{
	PHMD5 ctx;

	Phmd5Begin(&ctx);
	Phmd5ProcessZero(&ctx, block_size);
	Phmd5End(&ctx);
	memcpy(hash, ctx.hash, 16);
	memcpy(hash + 16, &window_mask, 4);	// CRC-32  window_mask Ɠ
}

// 2GB ̃t@C̊Jnʒuȍ~̃nbVlvZ
unsigned int file_crc_part(HANDLE hFile)
{
	unsigned char buf[4096];
	unsigned int len, crc = 0xFFFFFFFF;

	// ܂œǂݍ
	do {
		if (!ReadFile(hFile, buf, 4096, &len, NULL) || (len == 0))
			break;
		crc = crc_update(crc, buf, len);
	} while (len > 0);

	return crc ^ 0xFFFFFFFF;
}

// t@C offset oCgڂubNETCY MD5  CRC-32 ߂
int file_md5_crc32_block(
	HANDLE hFileRead,			// MD5  CRC ߂t@C̃nh
	__int64 offset,
	unsigned int avail_size,	// ̓oCg
	unsigned char *hash)		// nbVl (16 + 4oCg, MD5 + CRC-32)
{
	unsigned char buf[IO_SIZE];
	unsigned int rv, len, left, crc;
	PHMD5 hash_ctx;

	// t@C̊Jnʒu offsetoCgڂɂ
	if (!SetFilePointerEx(hFileRead, *((PLARGE_INTEGER)&offset), NULL, FILE_BEGIN))
		return 1;

	crc = 0xFFFFFFFF;	// 
	Phmd5Begin(&hash_ctx);

	left = avail_size;	// ۂɓǂݍރTCY
	while (left > 0){
		len = IO_SIZE;
		if (left < IO_SIZE)
			len = left;
		if (!ReadFile(hFileRead, buf, len, &rv, NULL)){
			print_win32_err();	// G[ʒm
			return 2;
		} else if (len != rv){
			return 3;
		}
		left -= len;

		crc = crc_update(crc, buf, len);	// CRC-32 vZ
		Phmd5Process(&hash_ctx, buf, len);	// MD5 vZ
	}

	if (avail_size < block_size){	// ubNETCY܂ 0Ŗ߂ČvZ
		crc = crc_update_zero(crc, block_size - avail_size);
		Phmd5ProcessZero(&hash_ctx, block_size - avail_size);
	}

	crc ^= 0xFFFFFFFF;	// ŏI
	memcpy(hash + 16, &crc, 4);
	Phmd5End(&hash_ctx);
	memcpy(hash, hash_ctx.hash, 16);

	return 0;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

//#define TIMER // p

#ifdef TIMER
//#undef IO_SIZE
//#define IO_SIZE	131072	// 256 KB ܂ő₵ĂӖAalignwʂȂ
static unsigned int time_start, time1_start, time2_start;
static unsigned int time_total = 0, time2_total = 0, time3_total = 0;
static unsigned int time5_total = 0, time6_total = 0;
#endif

typedef struct {
	unsigned char * volatile buf;
	volatile int size;
	HANDLE hFile;
	HANDLE run;
	HANDLE end;
} HASH_TH;

// ǂݍݗp̃TuEXbh
static DWORD WINAPI thread_read(LPVOID lpParameter)
{
	int rv;
	HASH_TH *th;
	PHMD5 *hash_ctx;

	th = (HASH_TH *)lpParameter;
	hash_ctx = (PHMD5 *)(th->buf);
	_mm_sfence();
	SetEvent(th->end);	// ݒIʒm

	WaitForSingleObject(th->run, INFINITE);	// Jn̍}҂
	while (th->size > 0){
#ifdef TIMER
time2_start = GetTickCount();
#endif
		// t@Cǂݍ
		if (!ReadFile(th->hFile, th->buf, th->size, &rv, NULL)){
			print_win32_err();	// G[ʒm
			th->size = -1;
		} else if (th->size != rv){
			th->size = -1;
		} else if (cpu_num > 1){
			Phmd5Process(hash_ctx, th->buf, th->size);	// MD5 vZ
		}
#ifdef TIMER
time2_total += GetTickCount() - time2_start;
#endif
		_mm_sfence();	// [ւ݂̏
		SetEvent(th->end);	// ǂݍݏIʒm
		WaitForSingleObject(th->run, INFINITE);	// Jn̍}҂
	}

	// I
	CloseHandle(th->run);
	CloseHandle(th->end);
	return 0;
}

// nbVlvZp̃TuEXbh
static DWORD WINAPI thread_hash(LPVOID lpParameter)
{
	HASH_TH *th;
	PHMD5 *hash_ctx;

	th = (HASH_TH *)lpParameter;
	hash_ctx = (PHMD5 *)(th->buf);
	_mm_sfence();
	SetEvent(th->end);	// ݒIʒm

	WaitForSingleObject(th->run, INFINITE);	// Jn̍}҂
	while (th->size > 0){
		Phmd5Process(hash_ctx, th->buf, th->size);	// MD5 vZ
		_mm_sfence();
		SetEvent(th->end);	// vZIʒm
		WaitForSingleObject(th->run, INFINITE);	// Jn̍}҂
	}

	// I
	CloseHandle(th->run);
	CloseHandle(th->end);
	return 0;
}

// t@C̃nbVlƊeXCX̃`FbNT𓯎ɌvZ
int file_hash_crc(
	wchar_t *file_path,			// nbVl߂t@C
	__int64 file_left,
	unsigned char *hash,		// nbVl (16oCg)
	unsigned char *sum,			// `FbNT (MD5 + CRC-32) ̔z
	unsigned int *time_last,	// OɌoߕ\
	__int64 *prog_now)			// oߕ\ł݈̌ʒu
{
	unsigned char *buf, buf1[IO_SIZE * 2];
	unsigned int err = 0, len, off, crc, block_left = 0, read_size;
	PHMD5 hash_ctx, block_ctx;
	HANDLE hFile, hSub;
	HASH_TH th[1];
#ifdef TIMER
time1_start = GetTickCount();
#endif

	hSub = NULL;
	buf = buf1 + IO_SIZE;

	// \[XEt@CJ
	hFile = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (hFile == INVALID_HANDLE_VALUE){
		print_win32_err();
		return 1;
	}

	// t@CEANZXpXbh̏
	th->size = 0;
	th->run = CreateEvent(NULL, FALSE, FALSE, NULL);	// Auto Reset ɂ
	if (th->run == NULL){
		print_win32_err();
		printf("error, sub-thread\n");
		err = 1;
		goto error_end;
	}
	th->end = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (th->end == NULL){
		print_win32_err();
		CloseHandle(th->run);
		printf("error, sub-thread\n");
		err = 1;
		goto error_end;
	}
	th->buf = (unsigned char *)&hash_ctx;
	th->hFile = hFile;
	_mm_sfence();	// [ւ݂̏ĂXbhN
	hSub = (HANDLE)_beginthreadex(NULL, 65536, thread_read, (LPVOID)th, 0, NULL);
	if (hSub == NULL){
		print_win32_err();
		CloseHandle(th->run);
		CloseHandle(th->end);
		printf("error, sub-thread\n");
		err = 1;
		goto error_end;
	}

	// ŏ̕ǂݍ
	read_size = IO_SIZE;
	if (file_left < IO_SIZE)
		read_size = (unsigned int)file_left;
	if (!ReadFile(hFile, buf1, read_size, &len, NULL)){
		print_win32_err();
		err = 1;
		goto error_end;
	} else if (read_size != len){
		err = 1;
		goto error_end;
	}
	Phmd5Begin(&hash_ctx);
	if (cpu_num > 1)
		Phmd5Process(&hash_ctx, buf1, read_size);	// MD5 vZ

	while (file_left > 0){
		len = read_size;
		file_left -= read_size;
		(*prog_now) += read_size;

		// O̓ǂݍ݂Î҂
		WaitForSingleObject(th->end, INFINITE);
		if (th->size < 0){	// O̓ǂݍ݂sĂ
			err = 1;
			goto error_end;
		}
		// ̕ǂݍ݊JnĂ
		if (file_left > 0){
			read_size = IO_SIZE;
			if (file_left < IO_SIZE)
				read_size = (unsigned int)file_left;
			ResetEvent(th->end);	// ZbgĂ
			th->buf = buf;
			th->size = read_size;
			_mm_sfence();	// [ւ݂̏ĂXbhĊJ
			SetEvent(th->run);	// TuEXbhɓǂݍ݂Jn
		}
		// obt@[ւ
		if (buf == buf1){
			buf = buf1 + IO_SIZE;
		} else {
			buf = buf1;
		}

#ifdef TIMER
time_start = GetTickCount();
#endif
		if (cpu_num == 1)
			Phmd5Process(&hash_ctx, buf, len);	// MD5 vZ
		off = 0;	// `FbNTvZ
		if (block_left > 0){	// O񑫂Ȃǉ
			//printf("file_left = %I64d, block_left = %d\n", file_left, block_left);
			if (block_left <= len){
				crc = crc_update(crc, buf, block_left) ^ 0xFFFFFFFF;	// CRC-32 vZ
				Phmd5Process(&block_ctx, buf, block_left);	// MD5 vZ
				Phmd5End(&block_ctx);	// ŏI
				memcpy(sum, block_ctx.hash, 16);
				memcpy(sum + 16, &crc, 4);
				sum += 20;
				off += block_left;
				block_left = 0;
			} else {
				crc = crc_update(crc, buf, len);	// CRC-32 vZ
				Phmd5Process(&block_ctx, buf, len);	// MD5 vZ
				off = len;
				block_left -= len;
			}
		}
		for (; off < len; off += block_size){
			Phmd5Begin(&block_ctx);
			if (off + block_size <= len){
				crc = crc_update(0xFFFFFFFF, buf + off, block_size) ^ 0xFFFFFFFF;	// CRC-32 vZ
				Phmd5Process(&block_ctx, buf + off, block_size);	// MD5 vZ
				Phmd5End(&block_ctx);	// ŏI
				memcpy(sum, block_ctx.hash, 16);
				memcpy(sum + 16, &crc, 4);
				sum += 20;
			} else {	// XCXr܂łȂ
				crc = crc_update(0xFFFFFFFF, buf + off, len - off);	// CRC-32 vZ
				Phmd5Process(&block_ctx, buf + off, len - off);	// MD5 vZ
				block_left = block_size - len + off;
			}
		}
#ifdef TIMER
time3_total += GetTickCount() - time_start;
#endif

		// oߕ\
		if (GetTickCount() - (*time_last) >= UPDATE_TIME){
			if (print_progress((int)(((*prog_now) * 1000) / total_file_size))){
				err = 2;
				goto error_end;
			}
			(*time_last) = GetTickCount();
		}
	}

	// ŏIubN[Ȃ
	if (block_left > 0){	// c 0 ŃpfBO (PAR2 dľ_)
		crc = crc_update_zero(crc, block_left) ^ 0xFFFFFFFF;	// CRC-32 vZ
		Phmd5ProcessZero(&block_ctx, block_left);
		Phmd5End(&block_ctx);
		memcpy(sum, block_ctx.hash, 16);
		memcpy(sum + 16, &crc, 4);
	}
	Phmd5End(&hash_ctx);	// ŏI
	memcpy(hash, hash_ctx.hash, 16);

error_end:
	CloseHandle(hFile);
	if (hSub){
		WaitForSingleObject(th->end, INFINITE);	// 쒆Ȃ҂
		th->size = 0;	// Iw
		_mm_sfence();
		SetEvent(th->run);
		WaitForSingleObject(hSub, INFINITE);
		CloseHandle(hSub);
	}

#ifdef TIMER
time_total += GetTickCount() - time1_start;
if (*prog_now == total_file_size){
	printf("\nsub   %d.%03d sec\n", time2_total / 1000, time2_total % 1000);
	printf("main  %d.%03d sec\n", time3_total / 1000, time3_total % 1000);
	if (time_total > 0){
		time_start = (int)((total_file_size * 125) / ((__int64)time_total * 131072));
	} else {
		time_start = 0;
	}
	printf("total %d.%03d sec, %d MB/s\n", time_total / 1000, time_total % 1000, time_start);
}
#endif
	return err;
}

// t@C̃nbVlׂ (SẴXCX̃`FbNT)
// 0`=jĂ邪ڂ܂œ, -1=MD5-16kقȂ, -2=LZ, -3=(Sǉ)
int file_hash_check(
	int num,				// file_ctx ɂt@Cԍ
	wchar_t *file_name,		// \t@C
	HANDLE hFile,			// t@C̃nh
	int prog_min,			// o̒l𒴂獇v\ (INT_MAX Ȃ\Ȃ)
	file_ctx_r *files,		// e\[XEt@C̏
	source_ctx_r *s_blk)	// e\[XEubN̏ (NULL ȂrȂ)
{
	unsigned char *buf, buf1[IO_SIZE * 2];
	int find_next, comp_num = 0;
	unsigned int len, off, crc, block_left = 0, read_size;
	unsigned int time_last;
	__int64 file_size, file_left;
	PHMD5 hash_ctx, block_ctx;
	HANDLE hSub;
	HASH_TH th[1];
#ifdef TIMER
time1_start = GetTickCount();
#endif

	prog_last = -1;	// ؒ̃t@C𖈉\
	time_last = GetTickCount();
	hSub = NULL;
	buf = buf1 + IO_SIZE;
	file_size = files[num].size;	// {̃t@CETCY
	find_next = files[num].b_off;	// 擪ubN̔ԍ

	// t@C̐擪ɖ߂
//	file_left = 0;
//	if (!SetFilePointerEx(hFile, *((PLARGE_INTEGER)&file_left), NULL, FILE_BEGIN))
//		return -1;

	// ܂͐擪 16KB ̃nbVlr
	len = 16384;
	if (file_size < 16384){
		len = (unsigned int)file_size;
		file_left = 0;
	} else {
		file_left = file_size - 16384;	// {̃t@CETCY܂łȂ
	}
#ifdef TIMER
time_start = GetTickCount();
#endif
	if (!ReadFile(hFile, buf, len, &off, NULL)){
		print_win32_err();	// G[ʒm
		return -1;
	} else if (len != off){
		return -1;
	}
#ifdef TIMER
time2_total += GetTickCount() - time_start;
#endif

	Phmd5Begin(&hash_ctx);	// 
	Phmd5Process(&hash_ctx, buf, len);	// MD5 vZ
	Phmd5End(&hash_ctx);	// ŏI
	if (file_left == 0){	// t@C 16KB ȉ
		if (s_blk != NULL){	// SẴXCXr
			for (off = 0; off < len; off += block_size){
				Phmd5Begin(&block_ctx);
				if (off + block_size <= len){
					crc = crc_update(0, buf + off, block_size);
					Phmd5Process(&block_ctx, buf + off, block_size);
				} else {	// ̔[ȃXCXȂ
					crc = crc_update(0, buf + off, len - off);
					Phmd5Process(&block_ctx, buf + off, len - off);
					block_left = block_size - len + off;	// ubNETCY܂ 0Ŗ߂ČvZ
					crc = crc_update_zero(crc, block_left);
					Phmd5ProcessZero(&block_ctx, block_left);
				}
				Phmd5End(&block_ctx);
				if ((crc != s_blk[find_next].crc) || (memcmp(block_ctx.hash, s_blk[find_next].hash, 16) != 0))
					return comp_num;	// XCX CRC-32  MD5 قȂ
				comp_num++;
				find_next++;
			}
		}
		if (memcmp(hash_ctx.hash, files[num].hash + 16, 16) != 0)	// 16k MD5 قȂ
			return -1;
		return -3;	// 
	}
	// 16KB 傫 16k MD5 قȂꍇ́AXCXPʂ̌Ȃ
	if (memcmp(hash_ctx.hash, files[num].hash + 16, 16) != 0)
		return -1;

	// 16k MD5 vc̕vZ
	Phmd5Begin(&hash_ctx);	// 
	Phmd5Process(&hash_ctx, buf, 16384);	// MD5 vZ
	if (s_blk != NULL){
		for (off = 0; off < 16384; off += block_size){
			Phmd5Begin(&block_ctx);
			if (off + block_size <= 16384){
				crc = crc_update(0, buf + off, block_size);
				Phmd5Process(&block_ctx, buf + off, block_size);
				Phmd5End(&block_ctx);
				if ((crc != s_blk[find_next].crc) || (memcmp(block_ctx.hash, s_blk[find_next].hash, 16) != 0))
					return comp_num;	// XCX CRC-32  MD5 قȂ
				comp_num++;
				find_next++;
			} else {	// XCXr܂łȂ
				crc = crc_update(0, buf + off, 16384 - off);
				Phmd5Process(&block_ctx, buf + off, 16384 - off);
				block_left = block_size - 16384 + off;
			}
		}
	}

	// t@CEANZXpXbh̏
	th->size = 0;
	th->run = CreateEvent(NULL, FALSE, FALSE, NULL);	// Auto Reset ɂ
	if (th->run == NULL)
		return comp_num;
	th->end = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (th->end == NULL){
		CloseHandle(th->run);
		return comp_num;
	}
	th->buf = (unsigned char *)&hash_ctx;
	th->hFile = hFile;
	_mm_sfence();	// [ւ݂̏ĂXbhN
	hSub = (HANDLE)_beginthreadex(NULL, 65536, thread_read, (LPVOID)th, 0, NULL);
	if (hSub == NULL){
		CloseHandle(th->run);
		CloseHandle(th->end);
		return comp_num;
	}

	// ŏ̕ǂݍ
	read_size = IO_SIZE;
	if (file_left < IO_SIZE)
		read_size = (unsigned int)file_left;
#ifdef TIMER
time_start = GetTickCount();
#endif
	if (!ReadFile(hFile, buf1, read_size, &len, NULL)){
		print_win32_err();	// G[ʒm
		goto error_end;
	} else if (read_size != len){
		goto error_end;
	}
#ifdef TIMER
time2_total += GetTickCount() - time_start;
#endif
	if (cpu_num > 1)
		Phmd5Process(&hash_ctx, buf1, read_size);	// MD5 vZ

	while (file_left > 0){
		len = read_size;	// ǂݍ񂾃TCY
		file_left -= read_size;

		// O̓ǂݍ݂Î҂
		WaitForSingleObject(th->end, INFINITE);
		if (th->size < 0)	// O̓ǂݍ݂sĂ
			goto error_end;
		// ̕ǂݍ݊JnĂ
		if (file_left > 0){
			ResetEvent(th->end);	// ZbgĂ
			read_size = IO_SIZE;
			if (file_left < IO_SIZE)
				read_size = (unsigned int)file_left;
			th->buf = buf;
			th->size = read_size;
			_mm_sfence();	// [ւ݂̏ĂXbhĊJ
			SetEvent(th->run);	// TuEXbhɓǂݍ݂Jn
		}
		// obt@[ւ
		if (buf == buf1){
			buf = buf1 + IO_SIZE;
		} else {
			buf = buf1;
		}

#ifdef TIMER
time_start = GetTickCount();
#endif
		if (cpu_num == 1)
			Phmd5Process(&hash_ctx, buf, len);	// MD5 vZ
		if (s_blk != NULL){
			off = 0;
			if (block_left > 0){	// O񑫂Ȃǉ
				if (block_left <= len){
					crc = crc_update(crc, buf, block_left);
					Phmd5Process(&block_ctx, buf, block_left);
					Phmd5End(&block_ctx);
					if ((crc != s_blk[find_next].crc) || (memcmp(block_ctx.hash, s_blk[find_next].hash, 16) != 0))
						goto error_end;	// XCX CRC-32  MD5 قȂ
					comp_num++;
					find_next++;
					off = block_left;
					block_left = 0;
				} else {
					crc = crc_update(crc, buf, len);
					Phmd5Process(&block_ctx, buf, len);
					off = len;
					block_left -= len;
				}
			}
			for (; off < len; off += block_size){
				Phmd5Begin(&block_ctx);
				if (off + block_size <= len){
					// s_blk[].crc ͏lƍŏI 0xFFFFFFFF 菜Ă
					crc = crc_update(0, buf + off, block_size);
					Phmd5Process(&block_ctx, buf + off, block_size);
					Phmd5End(&block_ctx);
					if ((crc != s_blk[find_next].crc) || (memcmp(block_ctx.hash, s_blk[find_next].hash, 16) != 0))
						goto error_end;	// XCX CRC-32  MD5 قȂ
					comp_num++;
					find_next++;
				} else {	// XCXr܂łȂ
					crc = crc_update(0, buf + off, len - off);
					Phmd5Process(&block_ctx, buf + off, len - off);
					block_left = block_size - len + off;
				}
			}
		}
#ifdef TIMER
time3_total += GetTickCount() - time_start;
#endif

		// oߕ\
		if (GetTickCount() - time_last >= UPDATE_TIME){
			if (comp_num <= prog_min){	// ŒlȉȂ猟oXCX̍v͕\Ȃ
				len = -1;
			} else {
				len = first_num + comp_num - prog_min;
			}
			if (print_progress_file((int)(((file_size - file_left) * 1000) / file_size), len, file_name)){
				comp_num = -2;
				goto error_end;
			}
			time_last = GetTickCount();
		}
	}
	Phmd5End(&hash_ctx);	// ŏI

	if ((s_blk != NULL) && (block_left > 0)){	// ̔[ȃXCXȂ
		crc = crc_update_zero(crc, block_left);
		Phmd5ProcessZero(&block_ctx, block_left);
		Phmd5End(&block_ctx);
		if ((crc != s_blk[find_next].crc) || (memcmp(block_ctx.hash, s_blk[find_next].hash, 16) != 0))
			goto error_end;	// XCX CRC-32  MD5 قȂ
		comp_num++;
	}

	if (memcmp(hash_ctx.hash, files[num].hash, 16) == 0)	// t@CS MD5 
		comp_num = -3;

error_end:
	if (hSub){
		WaitForSingleObject(th->end, INFINITE);	// 쒆Ȃ҂
		th->size = 0;	// Iw
		_mm_sfence();
		SetEvent(th->run);
		WaitForSingleObject(hSub, INFINITE);
		CloseHandle(hSub);
	}

#ifdef TIMER
time_total += GetTickCount() - time1_start;
	printf("\nsub   %d.%03d sec\n", time2_total / 1000, time2_total % 1000);
	printf("main  %d.%03d sec\n", time3_total / 1000, time3_total % 1000);
	if (time_total > 0){
		time_start = (int)((file_size * 125) / ((__int64)time_total * 131072));
	} else {
		time_start = 0;
	}
	printf("total %d.%03d sec, %d MB/s\n", time_total / 1000, time_total % 1000, time_start);
#endif
	return comp_num;
}

// LbVŃt@C̃nbVlׂ
// 0`=jĂ邪ڂ܂œ, -1=MD5-16kقȂ, -2=LZ, -3=(Sǉ), -4=݂Ȃ
int file_hash_direct(
	int num,				// file_ctx ɂt@Cԍ
	wchar_t *file_path,		// nbVl߂t@C
	wchar_t *file_name,		// \t@C
	file_ctx_r *files,		// e\[XEt@C̏
	source_ctx_r *s_blk)	// e\[XEubN̏ (NULL ȂrȂ)
{
	unsigned char *buf, *buf1 = NULL;
	int find_next, comp_num = 0;
	unsigned int len, off, crc, block_left = 0, read_size;
	unsigned int time_last;
	__int64 file_size, file_left, file_off;
	PHMD5 hash_ctx, block_ctx;
	HANDLE hFile, hSub;
	OVERLAPPED ol;
	HASH_TH th[1];
#ifdef TIMER
time1_start = GetTickCount();
#endif

	prog_last = -1;	// ؒ̃t@C𖈉\
	time_last = GetTickCount();
	memset(&ol, 0, sizeof(OVERLAPPED));
	hSub = NULL;
	file_size = files[num].size;	// {̃t@CETCY
	find_next = files[num].b_off;	// 擪ubN̔ԍ

	// \[XEt@CJ
	hFile = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
						FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED, NULL);
	if (hFile == INVALID_HANDLE_VALUE){
		off = GetLastError();
		if (off == 2)	// file not found (2)
			return -4;
		if ((off == 32) ||		// ERROR_SHARING_VIOLATION
				(off == 33)){	// ERROR_LOCK_VIOLATION
			Sleep(300);	// ҂ĂĒ킷
			hFile = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
						FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED, NULL);
		} else {
			//printf("cannot set FILE_FLAG_NO_BUFFERING %d\n", GetLastError());
			// fBXNELbV𖳌ɂłȂꍇ
			hFile = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
						FILE_FLAG_OVERLAPPED | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
		}
		if (hFile == INVALID_HANDLE_VALUE)
			return -4;
	}

	// obt@[̋EZN^[ETCY̔{ɂ
	// IO_SIZE  16KB ȏ 4KB ̔{ɂ邱
	buf1 = _aligned_malloc(IO_SIZE * 2, 4096);
	if (buf1 == NULL){
		comp_num = -1;
		goto error_end;
	}
	buf = buf1 + IO_SIZE;

	// 񓯊t@CEANZX̏
	ol.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (ol.hEvent == NULL){
		comp_num = -1;
		goto error_end;
	}

	// ܂͐擪 16KB ̃nbVlr
	ol.Offset = 0;
	ol.OffsetHigh = 0;
	file_off = 16384;
	if (file_size < 16384){
		len = (unsigned int)file_size;
		read_size = (len + 4095) & ~4095;	// 4KB ̔{ɂ
		file_left = 0;
	} else {
		len = 16384;
		read_size = 16384;
		file_left = file_size - 16384;	// {̃t@CETCY܂łȂ
	}
	off = ReadFile(hFile, buf, read_size, NULL, &ol);
	if ((off == 0) && (GetLastError() != ERROR_IO_PENDING)){
		comp_num = -1;
		goto error_end;
	}
	WaitForSingleObject(ol.hEvent, INFINITE);

	Phmd5Begin(&hash_ctx);	// 
	Phmd5Process(&hash_ctx, buf, len);	// MD5 vZ
	Phmd5End(&hash_ctx);	// ŏI
	if (file_left == 0){	// t@C 16KB ȉ
		if (s_blk != NULL){	// SẴXCXr
			for (off = 0; off < len; off += block_size){
				Phmd5Begin(&block_ctx);
				if (off + block_size <= len){
					crc = crc_update(0, buf + off, block_size);
					Phmd5Process(&block_ctx, buf + off, block_size);
				} else {	// ̔[ȃXCXȂ
					crc = crc_update(0, buf + off, len - off);
					Phmd5Process(&block_ctx, buf + off, len - off);
					block_left = block_size - len + off;	// ubNETCY܂ 0Ŗ߂ČvZ
					crc = crc_update_zero(crc, block_left);
					Phmd5ProcessZero(&block_ctx, block_left);
				}
				Phmd5End(&block_ctx);
				if ((crc != s_blk[find_next].crc) || (memcmp(block_ctx.hash, s_blk[find_next].hash, 16) != 0))
					goto error_end;	// XCX CRC-32  MD5 قȂ
				comp_num++;
				find_next++;
			}
		}
		if (memcmp(hash_ctx.hash, files[num].hash + 16, 16) != 0){	// 16k MD5 قȂ
			comp_num = -1;
			goto error_end;
		}
		comp_num = -3;	// 
		goto error_end;
	}
	// 16KB 傫 16k MD5 قȂꍇ́AXCXPʂ̌Ȃ
	if (memcmp(hash_ctx.hash, files[num].hash + 16, 16) != 0){
		comp_num = -1;
		goto error_end;
	}

	// 16k MD5 vc̕vZ
	Phmd5Begin(&hash_ctx);	// 
	Phmd5Process(&hash_ctx, buf, 16384);	// MD5 vZ
	if (s_blk != NULL){
		for (off = 0; off < 16384; off += block_size){
			Phmd5Begin(&block_ctx);
			if (off + block_size <= 16384){
				crc = crc_update(0, buf + off, block_size);
				Phmd5Process(&block_ctx, buf + off, block_size);
				Phmd5End(&block_ctx);
				if ((crc != s_blk[find_next].crc) || (memcmp(block_ctx.hash, s_blk[find_next].hash, 16) != 0))
					goto error_end;	// XCX CRC-32  MD5 قȂ
				comp_num++;
				find_next++;
			} else {	// XCXr܂łȂ
				crc = crc_update(0, buf + off, 16384 - off);
				Phmd5Process(&block_ctx, buf + off, 16384 - off);
				block_left = block_size - 16384 + off;
			}
		}
	}

	// ̕ǂݍ
	ol.Offset = (unsigned int)file_off;
	ol.OffsetHigh = (unsigned int)(file_off >> 32);
	file_off += IO_SIZE;
	read_size = IO_SIZE;
	if (file_left < IO_SIZE){
		read_size = (unsigned int)file_left;
		read_size = (read_size + 4095) & ~4095;	// 4KB ̔{ɂ
	}
	off = ReadFile(hFile, buf1, read_size, NULL, &ol);
	if ((off == 0) && (GetLastError() != ERROR_IO_PENDING))
		goto error_end;

	if (cpu_num > 1){	// nbVlvZpXbh̏
		th->run = CreateEvent(NULL, FALSE, FALSE, NULL);	// Auto Reset ɂ
		if (th->run == NULL)
			goto error_end;
		th->end = CreateEvent(NULL, TRUE, FALSE, NULL);
		if (th->end == NULL){
			CloseHandle(th->run);
			goto error_end;
		}
		th->buf = (unsigned char *)&hash_ctx;
		_mm_sfence();	// [ւ݂̏ĂXbhN
		hSub = (HANDLE)_beginthreadex(NULL, 65536, thread_hash, (LPVOID)th, 0, NULL);
		if (hSub == NULL){
			CloseHandle(th->run);
			CloseHandle(th->end);
			goto error_end;
		}
		// IO ؂Ȃ悤ɁATuEXbh̗Dx𗎂Ƃ
		SetThreadPriority(hSub, THREAD_PRIORITY_BELOW_NORMAL);
	}

	while (file_left > 0){
		len = IO_SIZE;	// ǂݍ񂾃TCY
		if (file_left < IO_SIZE)
			len = (unsigned int)file_left;
		file_left -= len;

		// O̓ǂݍ݂Î҂
		if (hSub)
			WaitForSingleObject(th->end, INFINITE);
		WaitForSingleObject(ol.hEvent, INFINITE);

		// ̕ǂݍ݊JnĂ
		if (file_left > 0){
			ol.Offset = (unsigned int)file_off;
			ol.OffsetHigh = (unsigned int)(file_off >> 32);
			file_off += IO_SIZE;
			read_size = IO_SIZE;
			if (file_left < IO_SIZE){
				read_size = (unsigned int)file_left;
				read_size = (read_size + 4095) & ~4095;	// 4KB ̔{ɂ
			}
			off = ReadFile(hFile, buf, read_size, NULL, &ol);
			if ((off == 0) && (GetLastError() != ERROR_IO_PENDING))
				goto error_end;
		}
		// obt@[ւ
		if (buf == buf1){
			buf = buf1 + IO_SIZE;
		} else {
			buf = buf1;
		}

#ifdef TIMER
time_start = GetTickCount();
#endif
		if (hSub){	// TuEXbhŌvZJn
			ResetEvent(th->end);	// ZbgĂ
			th->buf = buf;
			th->size = len;
			_mm_sfence();	// [ւ݂̏ĂXbhĊJ
			SetEvent(th->run);
		} else {
			Phmd5Process(&hash_ctx, buf, len);	// MD5 vZ
		}
		if (s_blk != NULL){
			off = 0;
			if (block_left > 0){	// O񑫂Ȃǉ
				if (block_left <= len){
					crc = crc_update(crc, buf, block_left);
					Phmd5Process(&block_ctx, buf, block_left);
					Phmd5End(&block_ctx);
					if ((crc != s_blk[find_next].crc) || (memcmp(block_ctx.hash, s_blk[find_next].hash, 16) != 0))
						goto error_end;	// XCX CRC-32  MD5 قȂ
					comp_num++;
					find_next++;
					off = block_left;
					block_left = 0;
				} else {
					crc = crc_update(crc, buf, len);
					Phmd5Process(&block_ctx, buf, len);
					off = len;
					block_left -= len;
				}
			}
			for (; off < len; off += block_size){
				Phmd5Begin(&block_ctx);
				if (off + block_size <= len){
					crc = crc_update(0, buf + off, block_size);
					Phmd5Process(&block_ctx, buf + off, block_size);
					Phmd5End(&block_ctx);
					if ((crc != s_blk[find_next].crc) || (memcmp(block_ctx.hash, s_blk[find_next].hash, 16) != 0))
						goto error_end;	// XCX CRC-32  MD5 قȂ
					comp_num++;
					find_next++;
				} else {	// XCXr܂łȂ
					crc = crc_update(0, buf + off, len - off);
					Phmd5Process(&block_ctx, buf + off, len - off);
					block_left = block_size - len + off;
				}
			}
		}
#ifdef TIMER
time6_total += GetTickCount() - time_start;
#endif

		// oߕ\
		if (GetTickCount() - time_last >= UPDATE_TIME){
			if (print_progress_file((int)(((file_size - file_left) * 1000) / file_size), -1, file_name)){
				comp_num = -2;
				goto error_end;
			}
			time_last = GetTickCount();
		}
	}

	if ((s_blk != NULL) && (block_left > 0)){	// ̔[ȃXCXȂ
		crc = crc_update_zero(crc, block_left);
		Phmd5ProcessZero(&block_ctx, block_left);
		Phmd5End(&block_ctx);
		if ((crc != s_blk[find_next].crc) || (memcmp(block_ctx.hash, s_blk[find_next].hash, 16) != 0))
			goto error_end;	// XCX CRC-32  MD5 قȂ
		comp_num++;
	}

	if (hSub)	// TuEXbȟvZI҂
		WaitForSingleObject(th->end, INFINITE);
	Phmd5End(&hash_ctx);	// ŏI
	if (memcmp(hash_ctx.hash, files[num].hash, 16) == 0)	// t@CS MD5 
		comp_num = -3;

error_end:
	CancelIo(hFile);	// 񓯊 IO 
	CloseHandle(hFile);
	if (ol.hEvent)
		CloseHandle(ol.hEvent);
	if (hSub){
		WaitForSingleObject(th->end, INFINITE);	// 쒆Ȃ҂
		th->size = 0;	// Iw
		_mm_sfence();
		SetEvent(th->run);
		WaitForSingleObject(hSub, INFINITE);
		CloseHandle(hSub);
	}
	if (buf1)
		_aligned_free(buf1);

#ifdef TIMER
time5_total += GetTickCount() - time1_start;
	printf("\nmain  %d.%03d sec\n", time6_total / 1000, time6_total % 1000);
	printf("total %d.%03d sec\n", time5_total / 1000, time5_total % 1000);
#endif
	return comp_num;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

// CRC-32 g 1oCg̃o[XgEG[
// -1=G[ł, 0=G[, 1=G[ł
// S 0 ̃f[^̔͑ōs
int correct_error(
	unsigned char *data_in,		// nbVl߂oCgz
	unsigned int data_len,		// ̓oCg
	unsigned char *hash,		// ubNETCY̖{̃nbVl (16oCg, MD5)
	unsigned int crc,			// ̓oCg̖{ CRC-32
	unsigned int *error_off,	// G[Jnʒu
	unsigned char *error_mag)	// G[eAXORŃG[
{
	unsigned char data_hash[16];
	unsigned int data_crc, pos;

	// ܂G[݂邩𒲂ׂ
	data_crc = crc_update(0, data_in, data_len);
	if (data_crc == crc){
		// MD5 ̓ubNETCYvZ
		data_md5_block(data_in, data_len, data_hash);
		if (memcmp(data_hash, hash, 16) == 0)
			return 0;	// G[
		return -1;	// CRC-32 ŃG[ołȂꍇ͒łȂ
	}
	if (data_len > 0x100000)
		return -1;	// 1MB ȏȂG[Ȃ
	// CRC-32  1oCg݂ƁAsm̓G[ʒuƂ 1 / 2^24 炢H
	// f[^ETCY 1MB (1^20) ƁAłȂĂ 1/16 ̊m MD5 ɂ錟ZKv

	// G[ʒuƒxvZ
	crc ^= data_crc;	// CRC-32 ̍ɂ
	for (pos = 1; pos <= data_len; pos++){
		// 1 oCgkȂŏ̃G[ʒuT
		crc = reverse_table[(crc >> 24)] ^ (crc << 8);

		// ͈͓̃G[
		if ((crc & 0xFFFFFF00) == 0){
			// ɃG[Ă݂
			*error_mag = (unsigned char)crc;
			data_in[data_len - pos] ^= *error_mag;	// G[ 1oCg

			// łĂ邩mF
			data_md5_block(data_in, data_len, data_hash);
			if (memcmp(data_hash, hash, 16) == 0){
				//printf("position = %d byte from end, magnitude = 0x%08X\n", pos - 1, crc);
				*error_off = data_len - pos;
				return 1;	// G[ł
			}
			// sȂf[^ɖ߂Ă
			//printf("MD5 found errors\n");
			data_in[data_len - pos] ^= *error_mag;
		}
	}

	return -1;	// ł
}

// CRC-32 g 1oCg̃o[XgEG[
// -1=G[ł, 0=G[, 1=G[ł
// CRC-32 ͊ɌvZς݁AG[rς
int correct_error2(
	unsigned char *data_in,		// nbVl߂oCgz
	unsigned int data_len,		// ̓oCg
	unsigned int data_crc,		// ̓oCg CRC-32
	unsigned char *hash,		// ubNETCY̖{̃nbVl (16oCg, MD5)
	unsigned int crc,			// ̓oCg̖{ CRC-32
	unsigned int *error_off,	// G[Jnʒu
	unsigned char *error_mag)	// G[eAXORŃG[
{
	unsigned char data_hash[16];
	unsigned int pos;

	if (data_len > 0x100000)
		return -1;	// 1MB ȏȂG[Ȃ

	// G[ʒuƒxvZ
	crc ^= data_crc;	// CRC-32 ̍ɂ
	for (pos = 1; pos <= data_len; pos++){
		// 1 oCgkȂŏ̃G[ʒuT
		crc = reverse_table[(crc >> 24)] ^ (crc << 8);

		// ͈͓̃G[
		if ((crc & 0xFFFFFF00) == 0){
			// ɃG[Ă݂
			*error_mag = (unsigned char)crc;
			data_in[data_len - pos] ^= *error_mag;	// G[ 1oCg

			// łĂ邩mF
			data_md5_block(data_in, data_len, data_hash);
			if (memcmp(data_hash, hash, 16) == 0){
				//printf("position = %d byte from end, magnitude = 0x%08X\n", pos - 1, crc);
				*error_off = data_len - pos;
				return 1;	// G[ł
			}
			// sȂf[^ɖ߂Ă
			//printf("MD5 found errors\n");
			data_in[data_len - pos] ^= *error_mag;
		}
	}

	return -1;	// ł
}

/*
// MD5  CRC32  8oCg̃f[^tZ邽߂̊֐
void data_reverse8(
	unsigned char *data_out,	// sȓ̓f[^ (8oCg)
	unsigned char *hash,		// nbVl (16oCg)
	unsigned char *crc)			// CRC (4oCg)
{
	unsigned int i, rv, crc1, crc2, data[16], state[4], time;
	// crc = 0xffffffff ^ data[0`3];	// lƍŏ̃f[^ XOR 
	// crc = table[crc & 0xff] ^ (crc >> 8);
	// crc = table[crc & 0xff] ^ (crc >> 8);
	// crc = table[crc & 0xff] ^ (crc >> 8);
	// crc = table[crc & 0xff] ^ (crc >> 8);
	// crc ^= data[4`7];	// ʂƎ̃f[^ XOR 
	// crc = table[crc & 0xff] ^ (crc >> 8);
	// crc = table[crc & 0xff] ^ (crc >> 8);
	// crc = table[crc & 0xff] ^ (crc >> 8);
	// crc = table[crc & 0xff] ^ (crc >> 8);
	// crc ^= 0xffffffff; ŏI

time = GetTickCount();

	memcpy(&crc2, crc, 4);
	crc2 ^= 0xffffffff;	// ŏI
	rv = CrcReverse4(crc2);	// 2ڂCRCvZl = crc ^ data[4`7]
	// 1ڂCRCvŽʂ킩Ȃ̂ŁA̓f[^킩Ȃ (XORl킩Ă)
	// ŁA̓f[^ωȂ1ڂCRCtZāAMD5ŊmF
	for (i = 0; i < 16; i++)
		data[i] = 0;
	data[2] = 0x00000080;	// f[^I[ 1bit
	data[14] = 0x00000040;	// f[^ETCY 8byte = 64bit
	do {
		crc1 = rv ^ data[1];
		data[0] = CrcReverse4(crc1) ^ 0xffffffff;	// tZď
		if ((data[1] & 0x00ffffff) == 0)
			printf("%08x%08x\n", data[0], data[1]);
		// ̃f[^ōĂ邩MD5Œׂ
		state[0] = 0x67452301;
		state[1] = 0xefcdab89;
		state[2] = 0x98badcfe;
		state[3] = 0x10325476;
		MD5Transform(data, state);

		if (memcmp(hash, state, 16) == 0)
			break;
		data[1]++;
	} while (data[1] != 0);

	memcpy(data_out, (unsigned char *)data, 8);
// Pentium 3 800MHz ƊSɖ50AeLXgȂ20炢?
time = (GetTickCount() - time);
printf("\n%3u.%03u sec\n", time / 1000, time % 1000);
}
*/

