// par2_cmd.c
// Copyright : 2019-04-27 Yutaka Sawada
// License : GPL

#ifndef _UNICODE
#define _UNICODE
#endif
#ifndef UNICODE
#define UNICODE
#endif
#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0501
#endif

#include <stdio.h>

#include <windows.h>

#include "common2.h"
#include "par2.h"
#include "ini.h"
#include "lib_opencl.h"
#include "version.h"


static void print_help(void)
{
	printf(
"Usage\n"
"t(rial)  [options] <par file> [input files]\n"
"c(reate) [options] <par file> [input files]\n"
"  available: f,fu,fo,fa,fe,ss,sn,sr,sm,rr,rn,rp,rs,rd,rf,ri,\n"
"\t     lr,lp,ls,lc,m,vs,vd,c,d,in,up,uo\n"
"v(erify) [f,fu,   m,vl,vs,vd,d,uo        ] <par file> [external files]\n"
"r(epair) [f,fu,lc,m,vl,vs,vd,d,uo,b,br,bi] <par file> [external files]\n"
"l(ist)   [uo,h] <par file>\n"
"\nOption\n"
" /f    : Use file-list instead of filename\n"
" /fu   : Use file-list which is encoded with UTF-8\n"
" /fo   : Search file only for wildcard\n"
" /fa\"*\": Adding file at search with wildcard\n"
" /fe\"*\": Excluding file at search with wildcard\n"
" /ss<n>: Slice size\n"
" /sn<n>: Number of source blocks\n"
" /sr<n>: Rate of source block count and size\n"
" /sm<n>: Slice size becomes a multiple of this value\n"
" /rr<n>: Rate of redundancy (%%)\n"
" /rn<n>: Number of recovery blocks\n"
" /rp<n>: Number of possible recovered files\n"
" /rs<n>: Starting recovery block number\n"
" /rd<n>: How to distribute recovery blocks to recovery files\n"
" /rf<n>: Number of recovery files\n"
" /ri   : Use file index to name recovery files\n"
" /lr<n>: Limit number of recovery blocks in a recovery file\n"
" /lp<n>: Limit repetition of packets in a recovery file\n"
" /ls<n>: Limit size of splited files\n"
" /lc<n>: Limit CPU feature\n"
" /m<n> : Memory usage\n"
" /vl<n>: Verification level\n"
" /vs<n>: Skip verification by recent result\n"
" /vd\"*\": Set directory of recent result\n"
" /c\"*\" : Set comment\n"
" /d\"*\" : Set directory of input files\n"
" /in   : Do not create index file\n"
" /up   : Write Unicode Filename packet for non-ASCII filename\n"
" /uo   : Console output is encoded with UTF-8\n"
" /b    : Backup existing files at repair\n"
" /br   : Send existing files into recycle bin at repair\n"
" /bi   : Replace files even when repair was failed\n"
" /h    : List hash value of input files\n"
	);
}

// ̕\
static void print_environment(void)
{
	MEMORYSTATUSEX statex;

	// u\\?\v͏ɕtĂ̂ŕ\ۂɂ͖
	printf_cp("Base Directory\t: \"%s\"\n", base_dir);
	printf_cp("Recovery File\t: \"%s\"\n", recovery_file);

	printf("CPU thread\t: %d / %d\n", cpu_num & 0xFFFF, cpu_num >> 24);
	cpu_num &= 0xFFFF;	// pRAɂĂ
	printf("CPU cache\t: %d KB per set\n", (cpu_flag & 0xFFFF8000) >> 10);
#ifndef _WIN64	// 32-bit ł MMX, SSE2, SSSE3 ̂ǂꂩ\
	printf("CPU extra\t:");
	if (cpu_flag & 1){
		if (cpu_flag & 256){
			printf(" SSSE3(old)");
		} else {
			printf(" SSSE3");
		}
	} else if (cpu_flag & 128){
		printf(" SSE2");
	} else {
		printf(" MMX");
	}
#else	// 64-bit ł SSE2, SSSE3 \
	printf("CPU extra\t: x64");
	if (cpu_flag & 1){
		if (cpu_flag & 256){
			printf(" SSSE3(old)");
		} else {
			printf(" SSSE3");
		}
	} else if (cpu_flag & 128){
		printf(" SSE2");
	}
#endif
	if (cpu_flag & 8)
		printf(" CLMUL");
	if (cpu_flag & 16)
		printf(" AVX2");
	printf("\nMemory usage\t: ");
	if (memory_use == 0){
		printf("Auto");
	} else {
		printf("%d/8", memory_use);
	}
	statex.dwLength = sizeof(statex);
	if (GlobalMemoryStatusEx(&statex))
		printf(" (%I64d MB available)", statex.ullAvailPhys >> 20);
	printf("\n\n");
}

// i[t@CEXgǂݍŃobt@[ɏ
static int read_list(
	wchar_t *list_path,		// Xg̃pX
	unsigned int code_page)	// CP_OEMCP  CP_UTF8
{
	char buf[MAX_LEN * 3];
	wchar_t file_name[MAX_LEN], file_path[MAX_LEN];
	int len;
	__int64 file_size;
	FILE *fp;
	WIN32_FILE_ATTRIBUTE_DATA AttrData;

	list_len = 0;
	list_max = ALLOC_LEN;
	list_buf = (wchar_t *)malloc(list_max * 2);
	if (list_buf == NULL){
		printf("malloc, %d\n", list_max * 2);
		return 1;
	}

	// ǂݍރt@CJ
	fp = _wfopen(list_path, L"rb");
	if (fp == NULL){
		printf_cp("cannot open file-list, %s\n", list_path);
		return 1;
	}

	// sǂݍ
	wcscpy(file_path, base_dir);
	while (fgets(buf, MAX_LEN * 3, fp)){
		if (ferror(fp))
			break;
		buf[MAX_LEN * 3 - 1] = 0;
		// ɉs΍폜
		for (len = 0; len < MAX_LEN * 3; len++){
			if (buf[len] == 0)
				break;
			if ((buf[len] == '\n') || (buf[len] == '\r')){
				buf[len] = 0;
				break;
			}
		}
		if (buf[0] == 0)
			continue;	// sȂ玟̍s

		// ǂݍ񂾓ejR[hɕϊÃfBNgL̕󂯂Ă
		if (!MultiByteToWideChar(code_page, 0, buf, -1, file_name, MAX_LEN - 1)){
			fclose(fp);
			printf("MultiByteToWideChar, %s\n", buf);
			return 1;
		}

		// t@CfBNgȉɑ݂邱ƂmF
		len = copy_path_prefix(file_path, MAX_LEN - ADD_LEN, file_name, base_dir);	// ΃pXɂĂr
		if (len == 0){
			fclose(fp);
			printf_cp("filename is invalid, %s\n", file_name);
			return 1;
		}
		if ((len <= base_len) || (_wcsnicmp(base_dir, file_path, base_len) != 0)){	// fBNgOȂ
			fclose(fp);
			printf_cp("out of base-directory, %s\n", file_path);
			return 1;
		}
		if (!GetFileAttributesEx(file_path, GetFileExInfoStandard, &AttrData)){
			print_win32_err();
			fclose(fp);
			printf_cp("input file is not found, %s\n", file_path);
			return 1;
		}

		// t@CdȂ悤ɂ
		wcscpy(file_name, file_path + base_len);
		if (search_file_path(list_buf, list_len, file_name))
			continue;

		if (AttrData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY){	// tH_Ȃ
			len = (int)wcslen(file_name);
			if (file_name[len - 1] != '\\'){
				file_name[len] = '\\';	// tH_̖͕Ku\vɂ
				file_name[len + 1] = 0;
			}
			file_size = 0;
			split_size = 0;
		} else {	// t@CȂ
			file_size = ((__int64)AttrData.nFileSizeHigh << 32) | (__int64)AttrData.nFileSizeLow;
			total_file_size += file_size;
			if (file_size > 0)
				entity_num++;
			// TuEfBNg܂̓tH_܂ޏꍇ̓t@C𖳌ɂ
			if ((split_size != 0) && (wcschr(file_name, '\\') != NULL))
				split_size = 0;
		}

		// XgɃRs[
		if (add_file_path(file_name)){
			printf("add_file_path\n");
			return 1;
		}
		file_num++;
	}
	fclose(fp);
/*
fp = fopen("list_buf.txt", "wb");
len = 0;
while (len < list_len){
	fwprintf(fp, L"%s\n", list_buf + len);
	len += (wcslen(list_buf + len) + 1);
}
fclose(fp);
*/
	return 0;
}

// t@Căt@CEXgɒǉ
static int search_files(
	wchar_t *search_path,	// t@C̃tEpXA* ? 
	int dir_len,			// fBNg̒
	int file_only,			// 0x20 = t@Ĉ݂ɂ
	int single_file)		// -1 = *?ŌwA0` = PƎw
{
	int len, dir_len2, old_num;
	__int64 file_size;
	HANDLE hFind;
	WIN32_FIND_DATA FindData;

	if (list_buf == NULL){
		list_len = 0;
		list_max = ALLOC_LEN;
		list_buf = (wchar_t *)malloc(list_max * 2);
		if (list_buf == NULL){
			printf("malloc, %d\n", list_max * 2);
			return 1;
		}
	}

	// Ɂu\vtătH_w肵ĂȂȂ
	len = (int)wcslen(search_path);
	if (search_path[len - 1] == '\\'){
		unsigned int rv;
		rv = GetFileAttributes(search_path);
		if ((rv != INVALID_FILE_ATTRIBUTES) && (rv & FILE_ATTRIBUTE_DIRECTORY)){	// ̃tH_݂Ȃ
			if (!search_file_path(list_buf, list_len, search_path + base_len)){	// tH_dȂ悤ɂ
				// ̃tH_ǉ
				if (add_file_path(search_path + base_len)){
					printf("add_file_path\n");
					return 1;
				}
				file_num++;
			}
		}
		return 0;
	}

	// 
	hFind = FindFirstFile(search_path, &FindData);
	if (hFind == INVALID_HANDLE_VALUE)	// Ȃ
		return 0;
	do {
		if ((wcscmp(FindData.cFileName, L".") == 0) || (wcscmp(FindData.cFileName, L"..") == 0))
			continue;	// ẽpX͖
		if ((single_file < 0) && (FindData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN))
			continue;	// ͉BtĂt@C𖳎

		len = (int)wcslen(FindData.cFileName);	// t@C̕
		if (dir_len + len >= MAX_LEN - ADD_LEN - 2){	// Ɂu\*vtčČ̂
			FindClose(hFind);
			printf("filename is too long\n");
			return 1;
		}
		// ݂̃fBNgɌt@CA
		wcscpy(search_path + dir_len, FindData.cFileName);

		// tH_Ȃ
		if (FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY){
			if (file_only == 0){
/*
				if (list2_buf){	// r̃tH_rďOȂ
					int off = 0;
					while (off < list2_len){
						if (list2_buf[off++] == '-'){	// deny
							if (PathMatchWild(search_path + base_len, list2_buf + off) != 0){
								printf_cp("deny : '%s'", search_path + base_len);
								printf_cp(" : '%s'\n", list2_buf + off);
								off = -1;
								break;
							}
						}
						off += wcslen(list2_buf + off) + 1;
					}
					if (off < 0)
						continue;	// OΏۂȂ疳
				}
*/

				// tH_̖́u\vɂ
				wcscat(search_path, L"\\");
				if (!search_file_path(list_buf, list_len, search_path + base_len)){	// tH_dȂ悤ɂ
					split_size = 0;
					old_num = file_num;	// ̃tH_̒Ƀt@C̂𒲂ׂ

					// ̃tH_̒gXɌ
					dir_len2 = (int)wcslen(search_path);
					search_path[dir_len2    ] = '*';	// Ɂu*vǉ
					search_path[dir_len2 + 1] = 0;
					if (search_files(search_path, dir_len2, file_only, single_file)){
						FindClose(hFind);
						printf("cannot search inner folder\n");
						return 1;
					}
					if (old_num == file_num){	// ̃tH_܂߂
						search_path[dir_len2] = 0;	// pɒǉu*v菜

						if (list2_buf){	// OtH_w肳ĂȂ
							if (exclude_path(search_path + base_len) != 0)
								continue;	// OΏۂȂ疳
						}

						// XgɃRs[
						if (add_file_path(search_path + base_len)){
							printf("add_file_path\n");
							return 1;
						}
						file_num++;
					}
				}
			}
		} else {	// t@CȂ
			if ((single_file < 0) && (list2_buf)){	// Ot@Cw肳ĂȂ
				if (exclude_path(search_path + base_len) != 0)
					continue;	// OΏۂȂ疳
			}

			if (!search_file_path(list_buf, list_len, search_path + base_len)){	// t@CdȂ悤ɂ
				// XgɃRs[
				if (add_file_path(search_path + base_len)){
					printf("add_file_path\n");
					return 1;
				}
				file_num++;
				file_size = ((__int64)FindData.nFileSizeHigh << 32) | (__int64)FindData.nFileSizeLow;
				total_file_size += file_size;
				if (file_size > 0)
					entity_num++;
				// TuEfBNg܂̓tH_܂ޏꍇ̓t@C𖳌ɂ
				if ((split_size != 0) && (wcschr(search_path + base_len, '\\') != NULL))
					split_size = 0;
			}
		}
	} while (FindNextFile(hFind, &FindData)); // ̃t@C
	FindClose(hFind);

	return 0;
}

// Ot@C̃XgǂݍŃobt@[ɏ
static int read_external_list(
	wchar_t *list_path,		// Xg̃pX
	unsigned int code_page)	// CP_OEMCP  CP_UTF8
{
	char buf[MAX_LEN * 3];
	wchar_t file_name[MAX_LEN], file_path[MAX_LEN], *tmp_p;
	int len, rv;
	FILE *fp;

	// ǂݍރt@CJ
	fp = _wfopen(list_path, L"rb");
	if (fp == NULL){
		printf_cp("cannot open file-list, %s\n", list_path);
		return 1;
	}

	// sǂݍ
	wcscpy(file_path, base_dir);
	while (fgets(buf, MAX_LEN * 3, fp)){
		if (ferror(fp))
			break;
		buf[MAX_LEN * 3 - 1] = 0;
		// ɉs΍폜
		for (len = 0; len < MAX_LEN * 3; len++){
			if (buf[len] == 0)
				break;
			if ((buf[len] == '\n') || (buf[len] == '\r')){
				buf[len] = 0;
				break;
			}
		}
		if (buf[0] == 0)
			continue;	// sȂ玟̍s

		// ǂݍ񂾓ejR[hɕϊÃfBNgL̕󂯂Ă
		if (!MultiByteToWideChar(code_page, 0, buf, -1, file_name, MAX_LEN - 1)){
			fclose(fp);
			printf("MultiByteToWideChar, %s\n", buf);
			return 1;
		}

		// t@C݂邱ƂmF
		len = copy_path_prefix(file_path, MAX_LEN - ADD_LEN, file_name, NULL);
		if (len == 0){
			fclose(fp);
			printf_cp("filename is invalid, %s\n", file_name);
			return 1;
		}
		//printf_cp("external file, %s\n", file_path);
		rv = GetFileAttributes(file_path);
		if ((rv == INVALID_FILE_ATTRIBUTES) || (rv & FILE_ATTRIBUTE_DIRECTORY)){
			fclose(fp);
			printf_cp("external file is not found, %s\n", file_path);
			return 1;
		}
		if (!search_file_path(list2_buf, list2_len, file_path)){	// t@CdȂ悤ɂ
			if (list2_len + len >= list2_max){ // ̈悪ȂȂȂg
				list2_max += ALLOC_LEN;
				tmp_p = (wchar_t *)realloc(list2_buf, list2_max * 2);
				if (tmp_p == NULL){
					fclose(fp);
					printf("realloc, %d\n", list2_max);
					return 1;
				} else {
					list2_buf = tmp_p;
				}
			}
			// ̃t@Cǉ
			//printf_cp("add external file, %s\n", file_path);
			wcscpy(list2_buf + list2_len, file_path);
			list2_len += len + 1;
		}
	}
	fclose(fp);

	return 0;
}

// Õt@Căt@CEXgɒǉ
static int search_external_files(
	wchar_t *search_path,	// t@C̃tEpXA* ? 
	int dir_len,			// fBNg̒
	int single_file)		// -1 = *?ŌwA0` = PƎw
{
	wchar_t *tmp_p;
	int len;
	HANDLE hFind;
	WIN32_FIND_DATA FindData;

	// 
	hFind = FindFirstFile(search_path, &FindData);
	if (hFind == INVALID_HANDLE_VALUE)	// Ȃ
		return 0;
	do {
		if (FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			continue;	// tH_͖
		if ((single_file < 0) && (FindData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN))
			continue;	// ͉BtĂt@C𖳎
		len = dir_len + (int)wcslen(FindData.cFileName);
		if (len >= MAX_LEN)
			continue;	// t@C͖

		// ݂̃fBNgɌt@CA
		wcscpy(search_path + dir_len, FindData.cFileName);
		if (!search_file_path(list2_buf, list2_len, search_path)){	// t@CdȂ悤ɂ
			if (list2_len + len >= list2_max){ // ̈悪ȂȂȂg
				list2_max += ALLOC_LEN;
				tmp_p = (wchar_t *)realloc(list2_buf, list2_max * 2);
				if (tmp_p == NULL){
					printf("realloc, %d\n", list2_max);
					return 1;
				} else {
					list2_buf = tmp_p;
				}
			}
			// ̃t@Cǉ
			//printf_cp("add external file, %s\n", search_path);
			wcscpy(list2_buf + list2_len, search_path);
			list2_len += len + 1;
		}
	} while (FindNextFile(hFind, &FindData)); // ̃t@C
	FindClose(hFind);

	return 0;
}

// JoEt@Cƃ\[XEt@COɂȂȂ悤ɂ
static int check_recovery_match(int switch_p)	// CfbNXEt@CȂ
{
	wchar_t recovery_base[MAX_LEN], file_ext[EXT_LEN], *tmp_p;
	int num, len, list_off = 0, ext_len, recovery_len;

	// fBNgOɃJoEt@C݂ȂȂ
	if (_wcsnicmp(recovery_file, base_dir, base_len) != 0)
		return 0;

	// JoEt@C̊gqɂ͎w肳ꂽ̂g
	ext_len = 0;
	wcscpy(recovery_base, recovery_file);
	tmp_p = offset_file_name(recovery_base);
	tmp_p = wcsrchr(tmp_p, '.');
	if (tmp_p != NULL){
		if (wcslen(tmp_p) < EXT_LEN){
			wcscpy(file_ext, tmp_p);	// gqL^Ă
			ext_len = (int)wcslen(file_ext);
			*tmp_p = 0;	// gq菜
		}
	}
	recovery_len = (int)wcslen(recovery_base + base_len);

	// t@CƂɔr
	for (num = 0; num < file_num; num++){
		tmp_p = list_buf + list_off;
		while (list_buf[list_off] != 0)
			list_off++;
		list_off++;

		// t@CƂ̑Ov𒲂ׂ
		if (_wcsnicmp(recovery_base + base_len, tmp_p, recovery_len) == 0){	// JoEt@CƊ
			len = (int)wcslen(tmp_p);
			// gqׂ
			if ((ext_len != 0) && (len > ext_len)){
				if (_wcsnicmp(file_ext, tmp_p + (len - ext_len), ext_len) == 0){
					if (recovery_len + ext_len == len){
						// CfbNXEt@CƔr
						if (switch_p == 0){
							printf_cp("filename is invalid, %s\n", tmp_p);
							return 1;
						}
					} else if (len - recovery_len - ext_len >= EXT_LEN){
						if (_wcsnicmp(tmp_p + recovery_len, L".vol", 4) == 0){
							printf_cp("filename is invalid, %s\n", tmp_p);
							return 1;
						}
					}
				}
			}
		}
	}
	return 0;
}

#define MAX_ADJUST 1024

// œKȃubNETCY𒲂ׂ
static int check_block_size(unsigned int adjust_size, int alloc_method, int limit_num)
{
	wchar_t file_path[MAX_LEN];
	int list_off = 0, i, target, loop_count;
	int lower_num, upper_num, new_num, sbc_rate;
	int source_max_num, source_min_num, source_num_best;
	unsigned int base_size, max_size, min_size, new_size;
	unsigned int block_max_size, block_min_size, block_size_best;
	__int64 *file_size, pad_size, pad_size_best;
	WIN32_FILE_ATTRIBUTE_DATA AttrData;

	base_size = 4;
	if (adjust_size != 0)
		base_size = adjust_size;

	// et@C̃TCY擾
	file_size = (__int64 *)malloc(sizeof(__int64) * file_num);
	if (file_size == NULL){
		printf("malloc, %d\n", sizeof(__int64) * file_num);
		return 1;
	}
	wcscpy(file_path, base_dir);
	block_max_size = 0;	// t@CƃubNETCY̍őlȂ
	for (i = 0; i < file_num; i++){
		wcscpy(file_path + base_len, list_buf + list_off);
		if (!GetFileAttributesEx(file_path, GetFileExInfoStandard, &AttrData)){
			print_win32_err();
			printf_cp("GetFileAttributesEx, %s\n", list_buf + list_off);
			free(file_size);
			return 1;
		}
		if (AttrData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY){	// tH_Ȃ
			file_size[i] = 0;
		} else {
			file_size[i] = ((__int64)AttrData.nFileSizeHigh << 32) | (unsigned __int64)AttrData.nFileSizeLow;
			if (file_size[i] > MAX_BLOCK_SIZE){
				block_max_size = MAX_BLOCK_SIZE;
			} else if (file_size[i] > (__int64)block_max_size){
				block_max_size = (unsigned int)(file_size[i]);
			}
		}
		while (list_buf[list_off] != 0)
			list_off++;
		list_off++;
	}

	// ubNETCY̍ől = ubN̍ŏl
	// őt@C̃TCYPʂ̔{ɂTCŶ͂
	if (base_size == 4){	// ؂グŒPʂ̔{ɂ
		block_max_size = (block_max_size + 3) & 0xFFFFFFFC;
	} else {
		block_max_size = ((block_max_size + base_size - 1) / base_size) * base_size;
	}
	source_min_num = 0;
	for (i = 0; i < file_num; i++){
		if (file_size[i] > 0)
			source_min_num += (int)((file_size[i] + (__int64)block_max_size - 1) / (__int64)block_max_size);
	}
	// ubNETCY̍ŏl = ubN̍ől
	if ((source_min_num >= MAX_SOURCE_NUM) || (block_max_size <= base_size)){	// t@CA܂͏
		source_max_num = source_min_num;
		block_min_size = block_max_size;
	} else {
		pad_size = (total_file_size + MAX_SOURCE_NUM - 1) / MAX_SOURCE_NUM;	// łɂ
		if (pad_size > (__int64)block_max_size - base_size)
			pad_size = (__int64)block_max_size - base_size;
		block_min_size = (unsigned int)pad_size;
		if (base_size == 4){	// ؂グŒPʂ̔{ɂ
			block_min_size = (block_min_size + 3) & 0xFFFFFFFC;
		} else {
			block_min_size = ((block_min_size + base_size - 1) / base_size) * base_size;
		}
		source_max_num = 0;
		for (i = 0; i < file_num; i++){
			if (file_size[i] > 0)
				source_max_num += (int)((file_size[i] + (__int64)block_min_size - 1) / (__int64)block_min_size);
		}
	}
	if (source_max_num > MAX_SOURCE_NUM){	// őubN𒴂AłȂTCYT
		i = source_max_num / source_min_num;	// Ɖ{炢Ⴄ
		new_size = (block_min_size * i + block_max_size) / (1 + i);	// TCYƍőubNETCY̊
		if (base_size == 4){	// ؂̂ĂŒPʂ̔{ɂ
			new_size &= 0xFFFFFFFC;
		} else {
			new_size -= new_size % base_size;
		}
		//printf("over = %d (%d blocks), rate = %d, new_size = %d\n", block_min_size, source_max_num, i, new_size);
		min_size = block_min_size;
		block_min_size = block_max_size;	// ŏubNETCY̍őlɂ
		source_max_num = source_min_num;	// ̃ubN
		while (block_min_size - min_size > base_size){
			new_num = 0;
			for (i = 0; i < file_num; i++){
				if (file_size[i] > 0)
					new_num += (int)((file_size[i] + (__int64)new_size - 1) / (__int64)new_size);
			}
			if (new_num > MAX_SOURCE_NUM){	// w肵ubNzAubNETCY傫
				min_size = new_size;	// zTCYŏlɂ
			} else {	// ubNETCY
				source_max_num = new_num;
				block_min_size = new_size;	// zȂŏubNETCY̍őlɂ
			}
			new_size = (min_size + block_min_size) / 2;	// ŏlƍől̒
			if (base_size == 4){	// ؂̂ĂŒPʂ̔{ɂ
				new_size &= 0xFFFFFFFC;
			} else {
				new_size -= new_size % base_size;
			}
			if (new_size <= min_size)
				new_size = min_size + base_size;
			//printf("over size = %d, ok = %d (%d blocks)\n", min_size, block_min_size, source_max_num);
		}
	}
	//printf("max = %d (%d blocks), min = %d (%d blocks)\n", block_max_size, source_min_num, block_min_size, source_max_num);

	// ubNETCY̏l߂
	max_size = block_max_size;
	min_size = block_min_size;
	lower_num = source_min_num;
	upper_num = source_max_num;
	if (alloc_method > 0){	// ubNETCYw肳ꂽ
		target = alloc_method;
		if (target >= (int)block_max_size){	// ubNETCY͈͓̔ɂ
			target = block_max_size;
			block_size = target;
		} else if (target <= (int)block_min_size){
			target = block_min_size;
			block_size = target;
		} else {	// Pʂ̔{Ŏw肳ꂽTCYɋ߂ɂ
			block_size = ((target + (base_size / 2)) / base_size) * base_size;
		}
		source_num = 0;
		for (i = 0; i < file_num; i++){
			if (file_size[i] > 0)	// ubNvZ
				source_num += (int)((file_size[i] + block_size - 1) / (__int64)block_size);
		}
		if ((limit_num > 0) && (source_num > limit_num)){	// ubN̐z
			alloc_method = 0;	// ubNw肵Ȃ
			max_size = block_max_size;
			min_size = block_min_size;
			lower_num = source_min_num;
			upper_num = source_max_num;
		}

	} else if (alloc_method < 0){	// (ubNETCY * ? = ubN)w肳ꂽ
		target = -alloc_method;	// ʗpɕɂĂ̂𐳂ɖ߂
		// ubNETCY̏l͒Pɍvt@CETCYvZ
		// SBC=root(F*rate/10000)=root(F*rate)/100
		pad_size = total_file_size * (__int64)target;	// rate/10000  rootȂ /100 ɂȂ
		new_num = sqrt64(pad_size);
		new_num = (new_num + 50) / 100;	// _ȉ͎ľܓ
		if (new_num > source_max_num){
			new_num = source_max_num;
		} else if (new_num < source_min_num){
			new_num = source_min_num;
		}
		if (new_num < 2)
			new_num = 2;	// Œł2ubNɂ
		pad_size = new_num;
		pad_size = (total_file_size + pad_size - 1) / pad_size;
		if (pad_size >= (__int64)block_max_size){
			new_size = block_max_size;
		} else {
			new_size = (unsigned int)pad_size;
			if (new_size <= block_min_size){
				new_size = block_min_size;
			} else {
				new_size = ((new_size + (base_size / 2)) / base_size) * base_size;	// ľܓĒPʂ̔{ɂ
			}
		}
		// ڕW͈͓̊ɂ
		sbc_rate = ((10000 * source_min_num) + (block_max_size / 2)) / block_max_size;
		if (target <= sbc_rate){
			if (sbc_rate < 10){
				target = 10;	// 0.1% ɂ͂Ȃ
			} else {
				target = sbc_rate;
			}
		} else {
			sbc_rate = ((10000 * source_max_num) + (block_min_size / 2)) / block_min_size;
			if (target > sbc_rate)
				target = sbc_rate;
		}
		//printf("target rate = %d.%02d%%, size = %d, max = %d, min = %d\n", target /100, target %100, new_size, max_size, min_size);
		// ڕWɋ߂Â
		while (max_size - min_size > base_size){
			new_num = 0;
			for (i = 0; i < file_num; i++){
				if (file_size[i] > 0)
					new_num += (int)((file_size[i] + new_size - 1) / (__int64)new_size);
			}

			sbc_rate = ((10000 * new_num) + (new_size / 2)) / new_size;
			//printf(" new_num = %d, new_size = %d, rate = %d.%02d%%\n", new_num, new_size, sbc_rate /100, sbc_rate %100);
			if (sbc_rate > target){	// w肵zAubNETCY傫
				upper_num = new_num;
				min_size = new_size;	// zTCYŏlɂ
				new_size = (min_size + max_size) / 2;	// ŏlƍől̒
				if (base_size == 4){	// ؂グŒPʂ̔{ɂ
					new_size = (new_size + 3) & 0xFFFFFFFC;
				} else {
					new_size = ((new_size + base_size - 1) / base_size) * base_size;
				}
				if (new_size >= max_size)
					new_size = max_size - base_size;
			} else {	// ubNETCY
				lower_num = new_num;
				max_size = new_size;	// zȂTCYőlɂ
				new_size = (min_size + max_size) / 2;	// ŏlƍől̒
				if (base_size == 4){	// ؂̂ĂŒPʂ̔{ɂ
					new_size &= 0xFFFFFFFC;
				} else {
					new_size -= new_size % base_size;
				}
				if (new_size <= min_size)
					new_size = min_size + base_size;
			}
			//printf(" lower = %d, max = %d, upper = %d, min = %d\n", lower_num, max_size, upper_num, min_size);
			if (lower_num == upper_num)
				break;
		}
		if (lower_num == upper_num){	// ubNȂ珬ubNETCY̕
			block_size = min_size;
			source_num = upper_num;
		} else {
			sbc_rate = ((10000 * lower_num) + (max_size / 2)) / max_size;
			i = ((10000 * upper_num) + (min_size / 2)) / min_size;
			// ߂ɂ
			if ((target - sbc_rate) < (i - target)){
				block_size = max_size;
				source_num = lower_num;
			} else {
				block_size = min_size;
				source_num = upper_num;
			}
		}
		if ((limit_num > 0) && (source_num > limit_num)){	// ubN̐z
			if (lower_num > limit_num){
				alloc_method = 0;	// ubNw肵Ȃ
				max_size = block_max_size;
				min_size = block_min_size;
				lower_num = source_min_num;
				upper_num = source_max_num;
			} else {	// Ȃg
				block_size = max_size;
				source_num = lower_num;
			}
		}
	}
	if (alloc_method == 0){	// ubNw肳ꂽA܂̓ubN̐z
		target = limit_num;
		if (target > source_max_num){	// ڕWubN͈͓ɂ
			target = source_max_num;
		} else if (target < source_min_num){
			target = source_min_num;
		}
		// ubNETCY̏l͒Pɍvt@CETCYڕWubNŊ
		pad_size = target;
		pad_size = (total_file_size + pad_size - 1) / pad_size;
		if (pad_size >= (__int64)block_max_size){
			new_size = block_max_size;
		} else {
			new_size = (unsigned int)pad_size;
			if (new_size <= block_min_size){
				new_size = block_min_size;
			} else {
				new_size = ((new_size + (base_size / 2)) / base_size) * base_size;	// ľܓĒPʂ̔{ɂ
			}
		}
		//printf("target num = %d, size = %d, max = %d, min = %d\n", target, new_size, max_size, min_size);
		// ڕWɋ߂Â
		while (max_size - min_size > base_size){
			new_num = 0;
			for (i = 0; i < file_num; i++){
				if (file_size[i] > 0)
					new_num += (int)((file_size[i] + new_size - 1) / (__int64)new_size);
			}
			//printf(" new_num = %d, new_size = %d\n", new_num, new_size);

			if (new_num > target){	// w肵ubNzAubNETCY傫
				upper_num = new_num;
				min_size = new_size;	// zTCYŏlɂ
				new_size = (min_size + max_size) / 2;	// ŏlƍől̒
				if (base_size == 4){	// ؂グŒPʂ̔{ɂ
					new_size = (new_size + 3) & 0xFFFFFFFC;
				} else {
					new_size = ((new_size + base_size - 1) / base_size) * base_size;
				}
				if (new_size >= max_size)
					new_size = max_size - base_size;
			} else {	// ubNETCY
				lower_num = new_num;
				max_size = new_size;	// zȂTCYőlɂ
				new_size = (min_size + max_size) / 2;	// ŏlƍől̒
				if (base_size == 4){	// ؂̂ĂŒPʂ̔{ɂ
					new_size &= 0xFFFFFFFC;
				} else {
					new_size -= new_size % base_size;
				}
				if (new_size <= min_size)
					new_size = min_size + base_size;
			}
			//printf(" lower = %d, max = %d, upper = %d, min = %d\n", lower_num, max_size, upper_num, min_size);
		}
		// ubN߂ɂ
		if ((target - lower_num) < (upper_num - target)){
			block_size = max_size;
			source_num = lower_num;
		} else {
			block_size = min_size;
			source_num = upper_num;
		}
	}

	// Ƃ肠œKlɐݒ肵Ă
	block_size_best = block_size;
	source_num_best = source_num;
	pad_size = ((__int64)block_size * (__int64)source_num) - total_file_size;
	pad_size_best = pad_size;
/*	sbc_rate = ((10000 * source_num) + (block_size / 2)) / block_size;
	i = (int)((1000 * total_file_size) / (total_file_size + pad_size));
	printf("start: size= %d, count= %d, sbc= %d.%02d%%, pad= %I64d, rate= %d.%d%%\n",
		block_size, source_num, sbc_rate /100, sbc_rate %100, pad_size, i /10, i %10);
*/
	if (pad_size > 0){	// ߕӂŌǂȂ鏊T
		if (alloc_method == 0){	// ubNw肳ꂽ
			// ubNETCY傫Ă݂
			loop_count = 0;
			new_size = block_size + base_size;
			while ((loop_count < MAX_ADJUST * 2) && (new_size <= block_max_size)){
				loop_count++;
				new_num = 0;
				for (i = 0; i < file_num; i++){
					if (file_size[i] > 0)	// ubNvZ
						new_num += (int)((file_size[i] + new_size - 1) / (__int64)new_size);
				}
				if (new_num < target - (target / 8))
					break;	// ڕW痣ꂷƂ -12%
				pad_size = ((__int64)new_size * (__int64)new_num) - total_file_size;
				if (pad_size < pad_size_best){
					block_size_best = new_size;
					source_num_best = new_num;
					pad_size_best = pad_size;
					if (pad_size == 0)
						break;
				}
				//i = (int)((1000 * total_file_size) / (total_file_size + pad_size));
				//printf("+ : size = %d, count = %d, pad = %I64d, rate = %d.%d%%\n",
				//	new_size, new_num, pad_size, i / 10, i % 10);
				new_size += base_size;
			}
			//printf("+ : %d trials\n", loop_count);
			// ubNETCYĂ݂
			loop_count = 0;
			new_size = block_size - base_size;
			while ((loop_count < MAX_ADJUST) && (new_size >= block_min_size)){
				loop_count++;
				new_num = 0;
				for (i = 0; i < file_num; i++){
					if (file_size[i] > 0)	// ubNvZ
						new_num += (int)((file_size[i] + new_size - 1) / (__int64)new_size);
				}
				if (new_num > target + (target / 16))
					break;	// ڕW痣ꂷƂ +6%
				pad_size = ((__int64)new_size * (__int64)new_num) - total_file_size;
				if (pad_size < pad_size_best){
					block_size_best = new_size;
					source_num_best = new_num;
					pad_size_best = pad_size;
					if (pad_size == 0)
						break;
				}
				//i = (int)((1000 * total_file_size) / (total_file_size + pad_size));
				//printf("- : size = %d, count = %d, pad = %I64d, rate = %d.%d%%\n",
				//	new_size, new_num, pad_size, i / 10, i % 10);
				new_size -= base_size;
			}
			//printf("- : %d trials\n", loop_count);

		} else if (alloc_method < 0){	// (ubNETCY * ? = ubN)w肳ꂽ
			// ubNETCY傫Ă݂
			loop_count = 0;
			new_size = block_size + base_size;
			while ((loop_count < MAX_ADJUST * 2) && (new_size <= block_max_size)){
				loop_count++;
				new_num = 0;
				for (i = 0; i < file_num; i++){
					if (file_size[i] > 0)	// ubNvZ
						new_num += (int)((file_size[i] + new_size - 1) / (__int64)new_size);
				}
				sbc_rate = ((10000 * new_num) + (new_size / 2)) / new_size;	// ľܓ
				if (sbc_rate < target - 9)
					break;	// ڕW痣ꂷƂ -0.09%
				pad_size = ((__int64)new_size * (__int64)new_num) - total_file_size;
				if (pad_size < pad_size_best){
					block_size_best = new_size;
					source_num_best = new_num;
					pad_size_best = pad_size;
					if (pad_size == 0)
						break;
				}
				//i = (int)((1000 * total_file_size) / (total_file_size + pad_size));
				//printf("+ : size= %d, count= %d, sbc= %d.%02d%%, pad= %I64d, rate= %d.%d%%\n",
				//	new_size, source_num, sbc_rate /100, sbc_rate %100, pad_size, i /10, i %10);
				new_size += base_size;
			}
			//printf("+ : %d trials\n", loop_count);
			// ubNETCYĂ݂
			loop_count = 0;
			new_size = block_size - base_size;
			while ((loop_count < MAX_ADJUST) && (new_size >= block_min_size)){
				loop_count++;
				new_num = 0;
				for (i = 0; i < file_num; i++){
					if (file_size[i] > 0)	// ubNvZ
						new_num += (int)((file_size[i] + new_size - 1) / (__int64)new_size);
				}
				if ((limit_num > 0) && (new_num > limit_num + (limit_num / 16)))
					break;	// ubNzƂ +6%
				sbc_rate = ((10000 * new_num) + (new_size / 2)) / new_size;	// ľܓ
				if (sbc_rate > target + 9)
					break;	// ڕW痣ꂷƂ +0.09%
				pad_size = ((__int64)new_size * (__int64)new_num) - total_file_size;
				if (pad_size < pad_size_best){
					block_size_best = new_size;
					source_num_best = new_num;
					pad_size_best = pad_size;
					if (pad_size == 0)
						break;
				}
				//i = (int)((1000 * total_file_size) / (total_file_size + pad_size));
				//printf("- : size= %d, count= %d, sbc= %d.%02d%%, pad= %I64d, rate= %d.%d%%\n",
				//	new_size, new_num, sbc_rate /100, sbc_rate %100, pad_size, i /10, i %10);
				new_size -= base_size;
			}
			//printf("- : %d trials\n", loop_count);

		} else if (adjust_size > 0){	// ubNETCYƕύXPʂ̗w肳ꂽA
			// ubNETCY傫Ă݂
			loop_count = 0;
			new_size = block_size + base_size;
			while ((loop_count < MAX_ADJUST * 2) && (new_size <= block_max_size)){
				loop_count++;
				if ((int)new_size > target + (target / 2))
					break;	// ڕW痣ꂷƂ +50%
				new_num = 0;
				for (i = 0; i < file_num; i++){
					if (file_size[i] > 0)	// ubNvZ
						new_num += (int)((file_size[i] + new_size - 1) / (__int64)new_size);
				}
				pad_size = ((__int64)new_size * (__int64)new_num) - total_file_size;
				if (pad_size < pad_size_best){
					block_size_best = new_size;
					source_num_best = new_num;
					pad_size_best = pad_size;
					if (pad_size == 0)
						break;
				}
				//i = (int)((1000 * total_file_size) / (total_file_size + pad_size));
				//printf("+ : size = %d, count = %d, pad = %I64d, rate = %d.%d%%\n",
				//	new_size, new_num, pad_size, i / 10, i % 10);
				new_size += base_size;
			}
			//printf("+ : %d trials\n", loop_count);
			// ubNETCYĂ݂
			loop_count = 0;
			new_size = block_size - base_size;
			while ((loop_count < MAX_ADJUST) && (new_size >= block_min_size)){
				loop_count++;
				if ((int)new_size < target - (target / 4))
					break;	// ڕW痣ꂷƂ -25%
				new_num = 0;
				for (i = 0; i < file_num; i++){
					if (file_size[i] > 0)	// ubNvZ
						new_num += (int)((file_size[i] + new_size - 1) / (__int64)new_size);
				}
				if ((limit_num > 0) && (new_num > limit_num + (limit_num / 16)))
					break;	// ubNzƂ +6%
				pad_size = ((__int64)new_size * (__int64)new_num) - total_file_size;
				if (pad_size < pad_size_best){
					block_size_best = new_size;
					source_num_best = new_num;
					pad_size_best = pad_size;
					if (pad_size == 0)
						break;
				}
				//i = (int)((1000 * total_file_size) / (total_file_size + pad_size));
				//printf("- : size = %d, count = %d, pad = %I64d, rate = %d.%d%%\n",
				//	new_size, new_num, pad_size, i / 10, i % 10);
				new_size -= base_size;
			}
			//printf("- : %d trials\n", loop_count);
		}
	}

	// œKlg
/*	sbc_rate = ((10000 * source_num_best) + (block_size_best / 2)) / block_size_best;
	i = (int)((1000 * total_file_size) / (total_file_size + pad_size_best));
	printf("best : size= %d, count= %d, sbc= %d.%02d%%, pad= %I64d, rate= %d.%d%%\n\n",
		block_size_best, source_num_best, sbc_rate /100, sbc_rate %100, pad_size_best, i /10, i %10);
*/
	block_size = block_size_best;
	source_num = source_num_best;
	if (recovery_limit < 0){	// \[XEt@C̍őXCXvZ
		recovery_limit = -1;	// 
		for (i = 0; i < file_num; i++){
			if (file_size[i] > 0){	// XCXvZ
				upper_num = (int)((file_size[i] + block_size - 1) / (__int64)block_size);
				if (upper_num + recovery_limit > 0)
					recovery_limit = -upper_num;
			}
		}
	}

	free(file_size);
	return 0;
}

// t@CAKvȃpeBEubNvZ
static int calc_required_parity(int possible_count)
{
	wchar_t file_path[MAX_LEN];
	int list_off = 0, i, block_need, extra_rate, id, block_count = 0;
	int *file_block;
	__int64 file_size;
	WIN32_FILE_ATTRIBUTE_DATA AttrData;

	if (possible_count <= 0)
		return 0;
	extra_rate = possible_count % 100;
	possible_count /= 100;
	//printf("possible_count = %d, extra_rate = %d\n", possible_count, extra_rate);
	if (possible_count >= entity_num)
		return source_num;

	// et@C̃\[XEubN擾
	file_block = (int *)malloc(sizeof(int) * file_num);
	if (file_block == NULL){
		printf("malloc, %d\n", sizeof(int) * file_num);
		return -1;
	}
	wcscpy(file_path, base_dir);
	for (i = 0; i < file_num; i++){
		wcscpy(file_path + base_len, list_buf + list_off);
		if (!GetFileAttributesEx(file_path, GetFileExInfoStandard, &AttrData)){
			print_win32_err();
			printf_cp("GetFileAttributesEx, %s\n", list_buf + list_off);
			free(file_block);
			return -1;
		}
		if (AttrData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY){	// tH_Ȃ
			file_block[i] = 0;
		} else {
			file_size = ((__int64)AttrData.nFileSizeHigh << 32) | (unsigned __int64)AttrData.nFileSizeLow;
			file_block[i] = (int)((file_size + block_size) / block_size);
			if (block_count < file_block[i]){
				block_count = file_block[i];
				id = i;
			}
		}
		//printf("%d: %d blocks\n", i, file_block[i]);
		while (list_buf[list_off] != 0)
			list_off++;
		list_off++;
	}
	//printf("max block = %d, file = %d\n", block_count, id);
	block_need = block_count;

	if (possible_count < 1){
		block_need = (block_need * extra_rate + 99) / 100;
	} else {
		while (possible_count > 1){
			// ubNő傾t@C菜
			file_block[id] = 0;
			block_count = 0;

			// ̎ɑ傫ȃt@CT
			for (i = 0; i < file_num; i++){
				if (block_count < file_block[i]){
					block_count = file_block[i];
					id = i;
				}
			}
			if (block_count == 0)
				break;
			block_need += block_count;
			//printf("max block = %d, file = %d, sum = %d\n", block_count, id, block_need);
			possible_count--;
		}
		block_need = (block_need * (100 + extra_rate) + 99) / 100;
		if (block_need > source_num)
			block_need = source_num;
	}

	free(file_block);
	return block_need;
}

wmain(int argc, wchar_t *argv[])
{
	wchar_t uni_buf[MAX_LEN], *tmp_p;
	int i, j, k;
	unsigned int switch_set = 0;
/*
in= switch_set & 0x00000001
f = switch_set & 0x00000002
fu= switch_set & 0x00000004
up= switch_set & 0x00000008
h = switch_set & 0x00000010
fo= switch_set & 0x00000020
vl= switch_set & 0x000000C0
lp= switch_set & 0x00000700
rd= switch_set & 0x00003000
ri= switch_set & 0x00004000
*/
	printf("Parchive 2.0 client version %x.%x.%x.%x by Yutaka Sawada\n\n",
		CLIENT_VERSION >> 12, (CLIENT_VERSION >> 8) & 0xF,
		(CLIENT_VERSION >> 4) & 0xF, CLIENT_VERSION & 0xF);
	if (argc < 3){
		printf("Self-Test: ");
		if (par2_checksum(uni_buf) == 0){
			printf("Success");
		} else {
			printf("Incomplete");
		}
		printf("\n\n");
		print_help();
		return 0;
	}

	// 
	recovery_file[0] = 0;
	base_dir[0] = 0;
	ini_path[0] = 0;
	uni_buf[0] = 0;
	file_num = 0;
	entity_num = 0;
	recovery_num = 0;
	source_num = 0;
	parity_num = 0;
	memory_use = 0;
	recovery_limit = 0;
	first_num = 0;
	switch_b = 0;
	block_size = 0;
	split_size = 0;
	total_file_size = 0;
	list_buf = NULL;
	recv_buf = NULL;
	list2_buf = NULL;
	cp_output = GetConsoleOutputCP();
	check_cpu();	// CPU 

	// R}h
	switch (argv[1][0]){
	case 'c':	// create
	case 't':	// trial
	case 'v':	// verify
	case 'r':	// repair
	case 'l':	// list
		break;
	default:
		print_help();
		return 0;
	}

	// IvVƃJoEt@C̎w
	for (i = 2; i < argc; i++){
		tmp_p = argv[i];
		// IvV
		if (((tmp_p[0] == '/') || (tmp_p[0] == '-')) && (tmp_p[0] == argv[2][0])){
			tmp_p++;	// 擪̎ʕƂ΂
			// 쐬̃IvV
			if (wcscmp(tmp_p, L"f") == 0){
				switch_set |= 0x02;
			} else if (wcscmp(tmp_p, L"fu") == 0){
				switch_set |= 0x06;
			} else if (wcscmp(tmp_p, L"fo") == 0){
				switch_set |= 0x20;
			} else if (wcscmp(tmp_p, L"ri") == 0){
				switch_set |= 0x4000;
			} else if (wcscmp(tmp_p, L"in") == 0){
				switch_set |= 0x01;
			} else if (wcscmp(tmp_p, L"up") == 0){
				switch_set |= 0x08;
			// ̃IvV
			} else if (wcscmp(tmp_p, L"b") == 0){
				switch_b |= 1;
			} else if (wcscmp(tmp_p, L"br") == 0){
				switch_b |= 2;
			} else if (wcscmp(tmp_p, L"bi") == 0){
				switch_b |= 4;
			} else if (wcscmp(tmp_p, L"h") == 0){
				switch_set |= 0x10;
			// ʂ̃IvV
			} else if (wcscmp(tmp_p, L"uo") == 0){
				cp_output = CP_UTF8;

			// 쐬̃IvV (l)
			} else if (wcsncmp(tmp_p, L"ss", 2) == 0){
				block_size = 0;
				j = 2;
				while ((j < 2 + 10) && (tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
					block_size = (block_size * 10) + (tmp_p[j] - '0');
					j++;
				}
				// ̎_ł͂܂Pʂ킩Ȃ̂ŋߎȂ
				if (block_size > MAX_BLOCK_SIZE)
					block_size = MAX_BLOCK_SIZE;
				switch_b &= 0x003FFFFF;	// ̎w
			} else if (wcsncmp(tmp_p, L"sn", 2) == 0){
				source_num = 0;
				j = 2;
				while ((j < 2 + 6) && (tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
					source_num = (source_num * 10) + (tmp_p[j] - '0');
					j++;
				}
				if (source_num > MAX_SOURCE_NUM)
					source_num = MAX_SOURCE_NUM;
			} else if (wcsncmp(tmp_p, L"sr", 2) == 0){
				k = 0;
				j = 2;
				while ((j < 2 + 5) && (tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
					k = (k * 10) + (tmp_p[j] - '0');
					j++;
				}
				if (k > 1023)
					k = 1023;
				switch_b = (switch_b & 0x003FFFFF) | (k << 22);
				block_size = 0;	// ubNETCY̎w
			} else if (wcsncmp(tmp_p, L"sm", 2) == 0){
				k = 0;
				j = 2;
				while ((j < 2 + 8) && (tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
					k = (k * 10) + (tmp_p[j] - '0');
					j++;
				}
				if (k > 4194300)
					k = 4194300;
				k &= 0x003FFFFC;	// 4̔{ɂ
				switch_b = (switch_b & 0xFFC00003) | k;
			} else if (wcsncmp(tmp_p, L"lp", 2) == 0){
				j = 4;	// lp  lp4 Ɠɂ
				if ((tmp_p[2] >= '0') && (tmp_p[2] <= '7'))
					j = tmp_p[2] - '0';
				if ((switch_set & 0x0700) == 0)
					switch_set |= (j << 8);
			} else if (wcsncmp(tmp_p, L"rn", 2) == 0){
				parity_num = 0;
				j = 2;
				while ((j < 2 + 6) && (tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
					parity_num = (parity_num * 10) + (tmp_p[j] - '0');
					j++;
				}
				if (parity_num > MAX_PARITY_NUM)
					parity_num = MAX_PARITY_NUM;
			} else if (wcsncmp(tmp_p, L"rr", 2) == 0){
				parity_num = 0;
				j = 2;
				while ((j < 2 + 5) && (tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
					parity_num = (parity_num * 10) + (tmp_p[j] - '0');
					j++;
				}
				parity_num *= 100;
				if ((tmp_p[j] == '.') || (tmp_p[j] == ',')){	// _ȂA_ȉʂ܂œǂݎ
					j++;
					if ((tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
						parity_num += (tmp_p[j] - '0') * 10;
						j++;
						if ((tmp_p[j] >= '0') && (tmp_p[j] <= '9'))
							parity_num += tmp_p[j] - '0';
					}
				}
				if (parity_num > 100000)
					parity_num = 100000;
				parity_num *= -1;
			} else if (wcsncmp(tmp_p, L"rp", 2) == 0){
				parity_num = 0;
				j = 2;
				while ((j < 2 + 4) && (tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
					parity_num = (parity_num * 10) + (tmp_p[j] - '0');
					j++;
				}
				if (parity_num > 999)
					parity_num = 999;
				parity_num *= 100;
				if ((tmp_p[j] == '.') || (tmp_p[j] == ',')){	// _ȂA_ȉʂ܂œǂݎ
					j++;
					if ((tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
						parity_num += (tmp_p[j] - '0') * 10;
						j++;
						if ((tmp_p[j] >= '0') && (tmp_p[j] <= '9'))
							parity_num += tmp_p[j] - '0';
					}
				}
				parity_num = parity_num * -1 - 200000;
			} else if (wcsncmp(tmp_p, L"rd", 2) == 0){
				j = 0;	// rd  rd0 Ɠɂ
				if ((tmp_p[2] >= '0') && (tmp_p[2] <= '3'))	// 0`3 ͈̔
					j = tmp_p[2] - '0';
				if ((switch_set & 0x3000) == 0)
					switch_set |= (j << 12);
			} else if (wcsncmp(tmp_p, L"rf", 2) == 0){
				recovery_num = 0;
				j = 2;
				while ((j < 2 + 6) && (tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
					recovery_num = (recovery_num * 10) + (tmp_p[j] - '0');
					j++;
				}
				if (recovery_num > MAX_PARITY_NUM)
					recovery_num = MAX_PARITY_NUM;
			} else if (wcsncmp(tmp_p, L"rs", 2) == 0){
				first_num = 0;
				j = 2;
				while ((j < 2 + 6) && (tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
					first_num = (first_num * 10) + (tmp_p[j] - '0');
					j++;
				}
				if (first_num > MAX_PARITY_NUM - 1)
					first_num = MAX_PARITY_NUM - 1;
			} else if (wcsncmp(tmp_p, L"lr", 2) == 0){
				recovery_limit = 0;
				j = 2;
				while ((j < 2 + 6) && (tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
					recovery_limit = (recovery_limit * 10) + (tmp_p[j] - '0');
					j++;
				}
				if (recovery_limit <= 0)	// -lr0 Ȃ\[XEt@C̍őubNƓɂ
					recovery_limit = -1;
				if (recovery_limit > MAX_PARITY_NUM)
					recovery_limit = MAX_PARITY_NUM;
			} else if (wcsncmp(tmp_p, L"ls", 2) == 0){
				split_size = 0;
				j = 2;
				while ((j < 2 + 10) && (tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
					split_size = (split_size * 10) + (tmp_p[j] - '0');
					j++;
				}
				if (split_size > 0x80000000)
					split_size = 0x80000000;	// TCY 2GB܂
			// ̃IvV (l)
			} else if (wcsncmp(tmp_p, L"vl", 2) == 0){
				j = 0;	// vl  vl0 Ɠɂ
				if ((tmp_p[2] >= '0') && (tmp_p[2] <= '3'))	// 0`3 ͈̔
					j = tmp_p[2] - '0';
				if ((switch_set & 0xC0) == 0)
					switch_set |= (j << 6);
			// ʂ̃IvV (l)
			} else if (wcsncmp(tmp_p, L"lc", 2) == 0){
				k = 0;
				j = 2;
				while ((j < 2 + 5) && (tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
					k = (k * 10) + (tmp_p[j] - '0');
					j++;
				}
				if (k & 32){	// GPU g
					OpenCL_method = 1;	// Faster GPU
				} else if (k & 64){
					OpenCL_method = -1;	// Slower GPU
				}
				if (k & 16)	// SSSE3 gȂ
					cpu_flag &= 0xFFFFFFFE;
				if (k & 128)	// CLMUL gȂASSSE3 ̌ÂGR[_[g
					cpu_flag = (cpu_flag & 0xFFFFFFF7) | 0x100;
				if (k & 256)	// JIT(SSE2) gȂ
					cpu_flag &= 0xFFFFFF7F;
				if (k & 512)	// AVX2 gȂ
					cpu_flag &= 0xFFFFFFEF;
				if (k & 15){	// gpRAύX
					k &= 15;	// 1`15 ͈̔
					if (k == 12){	// RA 1/4 ɂ
						k = ((cpu_num & 0x00FF0000) >> 16) / 4;
					} else if (k == 13){	// RA̔ɂ
						k = ((cpu_num & 0x00FF0000) >> 16) / 2;
					} else if (k == 14){	// RA 3/4 ɂ
						k = (((cpu_num & 0x00FF0000) >> 16) * 3) / 4;
					} else if (k == 15){	// RAɂ
						k = (cpu_num & 0x00FF0000) >> 16;
					} else if (k > (cpu_num >> 24)){
						k = cpu_num >> 24;	// _RA𒴂Ȃ悤ɂ
					}
					if (k > MAX_CPU){
						k = MAX_CPU;
					} else if (k < 1){
						k = 1;
					}
					cpu_num = (cpu_num & 0xFFFF0000) | k;	// w肳ꂽRAʂɔzu
				}
			} else if (wcsncmp(tmp_p, L"m", 1) == 0){
				if ((tmp_p[1] >= '0') && (tmp_p[1] <= '7'))	// 0`7 ͈̔
					memory_use = tmp_p[1] - '0';
			} else if (wcsncmp(tmp_p, L"vs", 2) == 0){
				recent_data = 0;
				j = 2;
				while ((j < 2 + 2) && (tmp_p[j] >= '0') && (tmp_p[j] <= '9')){
					recent_data = (recent_data * 10) + (tmp_p[j] - '0');
					j++;
				}
				if ((recent_data == 8) || (recent_data > 15))
					recent_data = 0;

			// 쐬̃IvV ()
			} else if (wcsncmp(tmp_p, L"c", 1) == 0){
				tmp_p++;
				if (wcslen(tmp_p) >= COMMENT_LEN){
					printf("comment is too long\n");
					return 1;
				}
				wcscpy(uni_buf, tmp_p);
			} else if ((wcsncmp(tmp_p, L"fa", 2) == 0) || (wcsncmp(tmp_p, L"fe", 2) == 0)){
				tmp_p += 2;
				if ((argv[1][0] == 'c') || (argv[1][0] == 't')){
					j = (int)wcslen(tmp_p);
					if ((j == 0) || (list2_len + j + 1 > ALLOC_LEN - MAX_LEN))
						continue;
					if (list2_buf == NULL){
						list2_max = 0;	// allow list ̍ڐ
						list2_len = 0;
						list2_buf = (wchar_t *)malloc(ALLOC_LEN * 2);
					}
					if (list2_buf != NULL){
						if (tmp_p[-1] == 'a'){
							list2_max++;
							list2_buf[list2_len++] = '+';
						} else {
							list2_buf[list2_len++] = '-';
						}
						j = copy_wild(list2_buf + list2_len, tmp_p);
						list2_len += j;
					}
				}
			// ̃IvV ()
			} else if (wcsncmp(tmp_p, L"vd", 2) == 0){
				tmp_p += 2;
				j = copy_path_prefix(ini_path, MAX_LEN - INI_NAME_LEN - 1, tmp_p, NULL);
				if (j == 0){
					printf("save-directory is invalid\n");
					return 1;
				}
				if (ini_path[j - 1] != '\\'){	// u\vłȂΕtĂ
					ini_path[j    ] = '\\';
					ini_path[j + 1] = 0;
				}
				//printf("Save Directory : %S\n", ini_path);
				j = GetFileAttributes(ini_path);
				if ((j == INVALID_FILE_ATTRIBUTES) || !(j & FILE_ATTRIBUTE_DIRECTORY)){
					printf("save-directory is invalid\n");
					return 1;
				}
			// ʂ̃IvV ()
			} else if (wcsncmp(tmp_p, L"d", 1) == 0){
				tmp_p++;
				j = copy_path_prefix(base_dir, MAX_LEN - 2, tmp_p, NULL);	// ɒǉ镪̗]TĂ
				if (j == 0){
					printf("base-directory is invalid\n");
					return 1;
				}
				if (base_dir[j - 1] != '\\'){	// u\vłȂΕtĂ
					base_dir[j    ] = '\\';
					base_dir[j + 1] = 0;
				}
				j = GetFileAttributes(base_dir);
				if ((j == INVALID_FILE_ATTRIBUTES) || !(j & FILE_ATTRIBUTE_DIRECTORY)){
					printf("base-directory is invalid\n");
					return 1;
				}

			} else {	// Ή̃IvV
				printf_cp("invalid option, %s\n", tmp_p - 1);
				return 1;
			}
			continue;
		}
		// IvVȊOȂ烊JoEt@C
		j = copy_path_prefix(recovery_file, MAX_LEN - 20, tmp_p, NULL);	// őŁu.vol32768+32768.par2vǉ
		if (j == 0){
			printf("PAR filename is invalid\n");
			return 1;
		}
		if ((argv[1][0] == 'c') || (argv[1][0] == 't')){	// 쐬Ȃ
			// w肳ꂽt@CK؂ׂ
			if (sanitize_filename(offset_file_name(recovery_file), NULL, 0) != 0){
				printf("PAR filename is invalid\n");
				return 1;
			}
			// gqu.par2vȊOȂW̊gqǉ
			j = (int)wcslen(recovery_file);	// 򉻂Ńt@CZk邩Ȃ
			if (_wcsicmp(recovery_file + (j - 5), L".par2"))
				wcscpy(recovery_file + j, L".par2");
		} else {	// CȂ
			j = GetFileAttributes(recovery_file);
			if ((j == INVALID_FILE_ATTRIBUTES) || (j & FILE_ATTRIBUTE_DIRECTORY)){
				wchar_t search_path[MAX_LEN], file_ext[EXT_LEN];
				int name_len, dir_len;
				HANDLE hFind;
				WIN32_FIND_DATA FindData;
				// JoEt@C̖O̊
				get_base_filename(recovery_file, search_path, file_ext);
				if (file_ext[0] == 0)
					wcscpy(file_ext, L".par2");	// gqȗĂȂW̊gqǉĂ݂
				// Ƀ{[ԍ܂܂Ȃt@C
				name_len = (int)wcslen(search_path);
				dir_len = (int)(offset_file_name(recovery_file) - recovery_file);
				wcscpy(search_path + name_len, file_ext);
				hFind = FindFirstFile(search_path, &FindData);
				if (hFind != INVALID_HANDLE_VALUE){
					file_ext[0] = 0;
					FindClose(hFind);
				}
				if (file_ext[0] != 0){	// {[ԍ܂ރt@C
					wcscpy(search_path + name_len, L".vol*");
					wcscat(search_path, file_ext);	// u`.par2vu`.vol*.par2vɂ
					hFind = FindFirstFile(search_path, &FindData);
					if (hFind != INVALID_HANDLE_VALUE){
						do {
							name_len = (int)wcslen(FindData.cFileName);
							if (dir_len + name_len < MAX_LEN){	// t@CȂ
								file_ext[0] = 0;
								break;	// t@CŖȂ
							}
						} while (FindNextFile(hFind, &FindData));	// ̃t@C
						FindClose(hFind);
					}
				}
				if (file_ext[0] != 0){	// ̑̃t@C
					wcscpy(search_path + name_len, L"*");
					wcscat(search_path, file_ext);	// u`.par2vu`*.par2vɂ
					hFind = FindFirstFile(search_path, &FindData);
					if (hFind != INVALID_HANDLE_VALUE){
						file_ext[0] = '.';	// gqɖ߂Ă
						do {
							//printf("file name = %S\n", FindData.cFileName);
							name_len = (int)wcslen(FindData.cFileName);
							if (dir_len + name_len < MAX_LEN){	// t@CȂ
								file_ext[0] = 0;
								break;	// t@CŖȂ
							}
						} while (FindNextFile(hFind, &FindData));	// ̃t@C
						FindClose(hFind);
					}
				}
				if (file_ext[0] != 0){
					printf("valid file is not found\n");
					return 1;
				}
				wcscpy(recovery_file + dir_len, FindData.cFileName);
			}
		}
		break;
	}

/*
// fobOp̔rȂ
if (list2_buf){
	list2_buf[list2_len] = 0;
	wcscpy(base_dir, list2_buf + 1);
} else {
	base_dir[0] = 0;
}
i = PathMatchWild(uni_buf, base_dir);
printf("text = '%S'\n", uni_buf);
printf("wild = '%S'\n", base_dir);
printf("result = %d\n", i);
return 0;
*/

	if (recovery_file[0] == 0){	// JoEt@Cw肳ĂȂȂ
		printf("PAR file is not specified\n");
		return 1;
	}

	// input file ̈ʒuw肳ĖāA
	// ŏ̃\[XEt@C΃pXŎw肳ĂȂAg
	if ((argv[1][0] == 'c') || (argv[1][0] == 't')){
		if ((base_dir[0] == 0) && ((switch_set & 0x02) == 0) && (i + 1 < argc)){
			tmp_p = argv[i + 1];
			if (is_full_path(tmp_p) != 0){	// ΃pXȂ
				wchar_t file_path[MAX_LEN];
				if (copy_path_prefix(file_path, MAX_LEN - 2, tmp_p, NULL) == 0){
					printf("base-directory is invalid\n");
					return 1;
				}
				get_base_dir(file_path, base_dir);	// ŏ̃\[XEt@C̈ʒuɂ
				if (len_without_prefix(base_dir) == 0)	// TuEfBNg
					base_dir[0] = 0;
			}
		}
	}
	if (base_dir[0] == 0)	// input file ̈ʒuw肳ĂȂȂ
		get_base_dir(recovery_file, base_dir);	// JoEt@C̈ʒuɂ
	base_len = (int)wcslen(base_dir);

	// ʃt@C̈ʒuw肳ĂȂȂ
	if ((recent_data != 0) && (ini_path[0] == 0)){
		// st@C̃fBNgɂ
		j = GetModuleFileName(NULL, ini_path, MAX_LEN);
		if ((j == 0) || (j >= MAX_LEN)){
			printf("GetModuleFileName\n");
			return 1;
		}
		while (j > 0){
			if (ini_path[j - 1] == '\\'){
				ini_path[j] = 0;
				break;
			}
			j--;
		}
		if (j >= MAX_LEN - INI_NAME_LEN - 1){
			printf("save-directory is invalid\n");
			return 1;
		}
	}

	// ̕\
	print_environment();

	switch (argv[1][0]){
	case 'c':
	case 't':
		// JoEt@C쐬Ȃ\[XEt@C̃Xg
		i++;
		if (i >= argc){	// t@Cw肪Ȃ
			wchar_t search_path[MAX_LEN];
			int dir_len;
			// JoEt@C̊gq菜̂\[XEt@CƌȂ
			wcscpy(search_path, recovery_file);
			tmp_p = offset_file_name(search_path);
			tmp_p = wcsrchr(tmp_p, '.');
			if (tmp_p != NULL)
				*tmp_p = 0;	// gq菜
			if (_wcsnicmp(base_dir, search_path, base_len) != 0){	// fBNgÕt@C͋ۂ
				printf_cp("out of base-directory, %s\n", search_path);
				return 1;
			}
			// ̃t@C݂邩m߂
			dir_len = (int)wcslen(search_path) - 2;	// t@Ću\v𖳎āAfBNg̒߂
			while (search_path[dir_len] != '\\')
				dir_len--;
			dir_len++;
			if (search_files(search_path, dir_len, 0x20, 0)){	// t@CT
				free(list_buf);
				return 1;
			}
			if (file_num != 1){	// t@Cm߂
				free(list_buf);
				printf_cp("input file is not found, %s\n", search_path + base_len);
				return 1;
			}
		} else if (switch_set & 0x06){	// t@CEXg̓ǂݍ
			j = (switch_set & 0x04) ? CP_UTF8 : CP_OEMCP;
			if (read_list(argv[i], j)){
				free(list_buf);
				return 1;
			}
		} else {	// ̓t@C̎w
			wchar_t search_path[MAX_LEN];
			int dir_len;

/*
			if (list2_buf){
				// fobOp̃Xg\
				printf("filtering list :\n", list2_buf);
				j = 0;
				while (j < list2_len){
					printf(" '%S'\n", list2_buf + j);
					j += wcslen(list2_buf + j) + 1;
				}
				printf("length : %d, allow number = %d\n\n", list2_len, list2_max);
			}
*/

			for (; i < argc; i++){
				// t@CfBNgȉɑ݂邱ƂmF
				tmp_p = argv[i];
				j = copy_path_prefix(search_path, MAX_LEN - ADD_LEN - 2, tmp_p, base_dir);	// ΃pXɂĂr
				if (j == 0){
					free(list_buf);
					printf_cp("filename is invalid, %s\n", tmp_p);
					return 1;
				}
				if ((j <= base_len) || (_wcsnicmp(base_dir, search_path, base_len) != 0)){	// fBNgOȂ
					free(list_buf);
					printf_cp("out of base-directory, %s\n", tmp_p);
					return 1;
				}
				//printf("%d = %S\n", i, argv[i]);
				//printf_cp("search = %s\n", search_path);
				// u*vu?vŌȂꍇAt@CȂ΃G[ɂ
				j = -1;
				if (wcspbrk(search_path + base_len, L"*?") == NULL)
					j = file_num;
				// t@C
				dir_len = (int)wcslen(search_path) - 2;	// t@Ću\v𖳎āAfBNg̒߂
				while (search_path[dir_len] != '\\')
					dir_len--;
				dir_len++;
				if (search_files(search_path, dir_len, switch_set & 0x20, j)){
					free(list_buf);
					return 1;
				}
				// t@Cm߂ (łɓo^ς݂Ȃ炢)
				if ((j != -1) && (j == file_num) &&
						(search_file_path(list_buf, list_len, search_path + base_len) == 0)){
					free(list_buf);
					printf_cp("input file is not found, %s\n", search_path + base_len);
					return 1;
				}
			}
		}
		if (list2_buf){	// Ot@CEXg
			free(list2_buf);
			list2_buf = NULL;
		}
/*{
FILE *fp;
fp = fopen("list_buf.txt", "wb");
fwrite(list_buf, 2, list_len, fp);
fclose(fp);
}*/
		if (file_num < 1){
			free(list_buf);
			printf("input file is not found\n");
			return 1;
		}
		if (file_num > MAX_SOURCE_NUM){
			free(list_buf);
			printf("too many input files %d\n", file_num);
			return 1;
		}
		if (check_recovery_match(switch_set & 0x01)){
			free(list_buf);
			return 1;
		}

		if (total_file_size == 0){	// t@CETCYS 0 ƃubN͂
			block_size = 0;
			source_num = 0;
			parity_num = 0;
		} else {	// ubNvZ
			// œKȃubNETCY𒲂ׂ
			i = 0;
			if (block_size != 0){	// ubNETCYw
				i = block_size;
			} else {	// ubNETCYƃubN̊w
				i = ((unsigned int)switch_b >> 22) * -10;	//  0.01% ݂ɂ
			}
			j = 0;	// Wł̓\[XEubN𐧌Ȃ
			if (source_num != 0){	// ubNw
				j = source_num;
			} else if (i == 0){	// SĂw肵ĂȂ
				i = -100;	//  1% ɂ
				j = 3000;	// 3000ubN܂łɂ
			}
			if (check_block_size(switch_b & 0x003FFFFC, i, j)){
				free(list_buf);
				return 1;
			}

			if (parity_num <= -200000){	// łt@CŃpeBEubN߂
				j = calc_required_parity(parity_num * -1 - 200000);
				if (j < 0){
					free(list_buf);
					return 1;
				}
				parity_num = j;
			} else if (parity_num < 0){	// 璷(%)ŃpeBEubN߂
				parity_num = -parity_num;	// %  100{ɂȂĂ邱Ƃɒ
				i = (int)(((unsigned int)parity_num * (unsigned int)source_num) / 10000);
				j = (i * 10000) / source_num;
				//printf("num0 = %d, rate = %d, target = %d \n", i, j, parity_num);
				if (j != parity_num){
					j = ((i + 1) * 10000) / source_num;
					//printf("num1 = %d, rate = %d, target = %d \n", i + 1, j, parity_num);
					if (j == parity_num)
						i++;	// 1₵ď璷v
				}
				if (i > MAX_PARITY_NUM)
					i = MAX_PARITY_NUM;
				if (i == 0)
					i = 1;
				parity_num = i;
			}
		}
		// TCY̓ubNETCY̔{ɂ
		if (split_size >= 4){
			if (split_size < block_size){
				split_size = block_size;
			} else {
				split_size -= split_size % block_size;
			}
		}
		// JoEt@C̐Ƃ̍őubNvZ
		if (parity_num > 0){
			if (recovery_limit == 0){	// lݒȂpeBEubNɂ
				recovery_limit = parity_num;
			} else if (recovery_limit < 0){	// \[XEt@C̍őubN𐧌lɂ
				recovery_limit = -recovery_limit;
				if (split_size >= 4){	// \[XEt@C𕪊ꍇ͂̕TCY܂
					j = split_size / block_size;	// t@C̃ubN
					if (recovery_limit > j)
						recovery_limit = j;
				}
			}
			// peBEubNJoEt@Cɕz@
			i = (switch_set & 0x3000) >> 12;
			if (i == 0){	// ʂU
				if (recovery_num == 0){	// ݒȂA\[XEt@CƏ璷vZ
					recovery_num = entity_num * parity_num / source_num;
					if (recovery_num < 1){
						recovery_num = 1;
					} else if (recovery_num > 10){
						recovery_num = 10;	// QuickPar Ɠ 10܂łɂ
					}
				}
				if (recovery_num > parity_num)
					recovery_num = parity_num;
				// Œt@CvZ
				j = (parity_num + recovery_limit - 1) / recovery_limit;
				if (recovery_num < j)
					recovery_num = j;
			} else if (i == 1){	// {XňقȂʂɂ
				if (recovery_num == 0){	// ݒȂA\[XEt@CƏ璷vZ
					recovery_num = entity_num * parity_num / source_num;
					if (recovery_num < 3){
						if (parity_num >= 7){
							recovery_num = 3;	// 咆 3ɂB
						} else if (parity_num >= 3){
							recovery_num = 2;	// 召 2ɂB
						} else {
							recovery_num = 1;
						}
					} else if (recovery_num > 10){
						recovery_num = 10;	// QuickPar Ɠ 10܂łɂ
					}
				}
				// l1̏ꍇ̃t@C = őt@C
				i = 1;	// recovery_num
				j = 1;	// total count
				k = 1;	// count in the file
				while (j < parity_num){
					k *= 2;
					if (k > recovery_limit){
						i += (parity_num - j + recovery_limit - 1) / recovery_limit;
						break;
					}
					j += k;
					i++;
				}
				//printf("recovery_num = %d, max = %d (%d blocks)\n", recovery_num, i, k);
				if (recovery_num > i)
					recovery_num = i;
				i = (1 << recovery_num) - 1;	// 
				k = (parity_num + i - 1) / i;	// {
				//printf("recovery_num = %d, split = %d, base = %d\n", recovery_num, i, k);
				if (recovery_limit < parity_num){
					if (recovery_limit * recovery_num < parity_num){	// t@CȂȂ瑝₷
						recovery_num = (parity_num + recovery_limit - 1) / recovery_limit;
						//printf("min = %d, * %d = %d\n", recovery_num, recovery_limit, recovery_limit * recovery_num);
					}
					i = parity_num;
					j = recovery_num;
					if (j > 16){	// E𒴂Ă镪͏Ȃ
						i -= recovery_limit * (j - 16);
						j = 16;
					}
					k = (1 << j) - 1;	// 
					k = (i + k - 1) / k;	// {
					//printf("rest = %d blocks on %d files, base = %d, last = %d, %d, limit = %d\n", i, j, k, k << (j - 2), i - ((1 << (j - 1)) - 1) * k, recovery_limit);
					while ((j >= 2) && (k < recovery_limit) && (((k << (j - 2)) > recovery_limit) || (i - ((1 << (j - 1)) - 1) * k > recovery_limit))){
						i -= recovery_limit;
						j--;
						k = (1 << j) - 1;	// 
						k = (i + k - 1) / k;	// {
						//printf("rest = %d blocks on %d files, base = %d, last = %d, %d\n", i, j, k, k << (j - 2), i - ((1 << (j - 1)) - 1) * k);
					}
				}
				recovery_limit = (recovery_limit & 0xFFFF) | (k << 16);
			} else {
				if (i == 2){	// 1,2,4,8,16 2̏搔ɂ
					recovery_num = 0;
					i = 1;	// exp_num
					j = 0;	// total count
					k = 0;	// count in the file
					while (j < parity_num){
						k = i;
						if (k >= recovery_limit){	// TCYĂȂ
							k = recovery_limit;
						} else {
							i *= 2;
						}
						j += k;
						recovery_num++;
					}
					recovery_limit = k;
				} else {	// 1,1,2,5,10,10,20,50 Ƃ decimal weights sizing scheme ɂ
					recovery_num = 0;
					i = 1;	// exp_num
					j = 0;	// total count
					k = 0;	// count in the file
					while (j < parity_num){
						k = i;
						if (k >= recovery_limit){	// TCYĂȂ
							k = recovery_limit;
						} else {
							switch (recovery_num % 4){
							case 1:
							case 3:
								i = i * 2;
								break;
							case 2:
								i = (i / 2) * 5;
								break;
							}
						}
						j += k;
						recovery_num++;
					}
					recovery_limit = k;
				}
			}
		} else {
			recovery_num = 0;
		}
		printf("Input File count\t: %d\n", file_num);
		printf("Input File total size\t: %I64d\n", total_file_size);
		printf("Input File Slice size\t: %u\n", block_size);
		printf("Input File Slice count\t: %d\n", source_num);
		printf("Recovery Slice count\t: %d\n", parity_num);
		if (first_num > 0)
			printf("Recovery Slice start\t: %d\n", first_num);
		if (source_num != 0){
			i = (10000 * parity_num) / source_num;
		} else {
			i = 0;
		}
		printf("Redundancy rate\t\t: %d.%02d%%\n", i / 100, i % 100);
		printf("Recovery File count\t: %d\n", recovery_num);	// Index File ͊̕܂܂Ȃ
		if (parity_num > 0){
			i = (switch_set & 0x3000) >> 12;
			printf("Slice distribution\t: %d, ", i);
			switch (i){
			case 0:
				printf("uniform (until %d)\n", recovery_limit);
				break;
			case 1:
				printf("variable (base %d until %d)\n", (unsigned int)recovery_limit >> 16, recovery_limit & 0xFFFF);
				break;
			case 2:
				printf("power of two (until %d)\n", recovery_limit);
				break;
			case 3:
				printf("decimal weights (until %d)\n", recovery_limit);
				break;
			}
			k = (switch_set & 0x0700) >> 8;
			printf("Packet Repetition limit\t: %d\n", k);
		}
		if (split_size == 1){	// ɂɃJoER[hǉł邩
			if ((file_num != 1) || (total_file_size < 54) || (total_file_size > 0x7FFFFFFF)){
				split_size = 0;	// zip's min = 100-byte, 7z's min = 74-byte, őł 2 GB 
			} else {	// gq𒲂ׂ
				j = (int)wcslen(list_buf);
				if ((_wcsicmp(list_buf + j - 4, L".zip") != 0) && (_wcsicmp(list_buf + j - 3, L".7z") != 0)){
					split_size = 0;
				} else {
					printf_cp("Append recovery record\t: \"%s\"\n", list_buf);
				}
			}
		} else if (split_size >= 4){	// TCY͎w肳ꂽ\
			if (total_file_size == 0){
				split_size = 0;
			} else {
				printf("Split size\t\t: %u\n", split_size);
			}
		}
		if (parity_num + first_num > MAX_PARITY_NUM){	// ubN̐
			free(list_buf);
			printf("too many recovery blocks %d\n", parity_num + first_num);
			return 1;
		}
		j = (switch_set & 0x01) | ((switch_set & 0x08) >> 2);
		if (argv[1][0] == 'c'){
			i = par2_create(uni_buf, (switch_set & 0x0700) >> 8, (switch_set & 0x7000) >> 12, j);
		} else {
			i = par2_trial(uni_buf, (switch_set & 0x0700) >> 8, (switch_set & 0x7000) >> 12, j);
		}
		break;
	case 'v':
	case 'r':
		// ECȂǑΏۃt@C̃Xg邩Ȃ
		i++;
		if (i < argc){	// Õt@Cw肳ĂȂ
			wchar_t search_path[MAX_LEN];
			int dir_len;
			list2_len = 0;
			list2_max = ALLOC_LEN;
			list2_buf = (wchar_t *)malloc(list2_max * 2);
			if (list2_buf == NULL){
				printf("malloc, %d\n", list2_max * 2);
				return 1;
			}

			if (switch_set & 0x06){	// t@CEXg̓ǂݍ
				j = (switch_set & 0x04) ? CP_UTF8 : CP_OEMCP;
				if (read_external_list(argv[i], j))
					list2_len = 0;
				i = argc;	// ȏǂݍ܂Ȃ
			}

			for (; i < argc; i++){
				j = copy_path_prefix(search_path, MAX_LEN, argv[i], NULL);
				if (j == 0){
					free(list2_buf);
					printf_cp("filename is invalid, %s\n", argv[i]);
					return 1;
				}
				//printf_cp("external file, %s\n", search_path);
				dir_len = (int)wcslen(search_path) - 2;	// t@Ću\v𖳎āAfBNg̒߂
				while (search_path[dir_len] != '\\')
					dir_len--;
				dir_len++;
				// u*vu?vŌȂꍇAt@CȂ΃G[ɂ
				j = -1;
				if (wcspbrk(search_path + dir_len, L"*?") == NULL)
					j = list2_len;
				if (search_external_files(search_path, dir_len, j)){
					free(list2_buf);
					return 1;
				}
				if ((j != -1) && (j == list2_len)){	// t@Cm߂
					free(list2_buf);
					printf_cp("external file is not found, %s\n", search_path);
					return 1;
				}
			}
/*{
FILE *fp;
fp = fopen("list2_buf.txt", "wb");
fwrite(list2_buf, 2, list2_len, fp);
fclose(fp);
}*/
			if (list2_len == 0){	// Lȃt@CȂꍇ
				free(list2_buf);
				list2_buf = NULL;
			}
		}
		switch_b &= 7;
		if (argv[1][0] == 'v'){
			i = par2_verify(uni_buf, (switch_set & 0xC0) >> 6);
		} else {
			i = par2_repair(uni_buf, (switch_set & 0xC0) >> 6);
		}
		if (list2_buf)
			free(list2_buf);
		break;
	case 'l':
		i = par2_list(uni_buf, (switch_set & 0x10) >> 4);
		break;
	}

	if (list_buf);
		free(list_buf);
	//printf("ExitCode: 0x%02X\n", i);
	return i;
}

