// rs_decode.c
// Copyright : 2018-12-30 Yutaka Sawada
// License : GPL

#ifndef _UNICODE
#define _UNICODE
#endif
#ifndef UNICODE
#define UNICODE
#endif
#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0501
#endif

#include <malloc.h>
#include <process.h>
#include <stdio.h>

#include <windows.h>

#include "common2.h"
#include "crc.h"
#include "gf16.h"
#include "phmd5.h"
#include "lib_opencl.h"
#include "reedsolomon.h"
#include "rs_decode.h"


#ifdef TIMER
static unsigned int time_start, time_read = 0, time_write = 0, time_calc = 0;
static unsigned int read_count = 0, write_count = 0, skip_count = 0;
#endif

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
// 񓯊 IO

typedef struct {	// RS threading control struct
	unsigned short * volatile mat;	// s
	unsigned char * volatile buf;
	volatile unsigned int size;		// oCg
	volatile int count;
	volatile int off;
	volatile int now;
	HANDLE run;
	HANDLE end;
} RS_TH;

static DWORD WINAPI thread_decode2(LPVOID lpParameter)
{
	unsigned char *buf, *p_buf, *work_buf;
	unsigned short *factor;
	int i, j, th_id, part_num, chunk_num, max_num, left_num, left_max;
	unsigned int unit_size, len, off, chunk_size;
	HANDLE hRun, hEnd;
	RS_TH *th;
#ifdef TIMER
unsigned int loop_count2a = 0, loop_count2b = 0;
unsigned int time_start2, time_encode2a = 0, time_encode2b = 0;
#endif

	th = (RS_TH *)lpParameter;
	buf = th->buf;
	unit_size = th->size;
	part_num = th->count;
	chunk_size = th->off;
	th_id = th->now;	// Xbhԍ
	hRun = th->run;
	hEnd = th->end;
	_mm_sfence();
	SetEvent(hEnd);	// ݒ芮ʒm

	p_buf = buf + (size_t)unit_size * (source_num + cpu_num - 1);
	chunk_num = (unit_size + chunk_size - 1) / chunk_size;
	max_num = chunk_num * part_num;
	if (th_id > 0)
		work_buf = buf + (size_t)unit_size * (source_num + th_id - 1);

	WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	while (th->now < INT_MAX / 2){
#ifdef TIMER
time_start2 = GetTickCount();
#endif
		i = th->off;	// \[XEubNԍ
		factor = th->mat;
		len = chunk_size;
		if (th->buf == NULL){	// \[XEubNǂݍݒ
			// ubNƂɊ|ZĒǉĂ
			while ((j = InterlockedIncrement(&(th->now))) < max_num){	// j = ++th_now
				off = j / part_num;	// chunk ̔ԍ
				j = j % part_num;	// lost block ̔ԍ
				off *= chunk_size;
				if (off + len > unit_size)
					len = unit_size - off;	// Ō chunk TCYقȂ邩
				if (i == 0)	// ŏ̃ubNvZۂ
					memset(p_buf + ((size_t)unit_size * j + off), 0, len);	// ubN 0Ŗ߂
				galois_align_multiply(buf + ((size_t)unit_size * i + off), p_buf + ((size_t)unit_size * j + off), len, factor[source_num * j]);
#ifdef TIMER
loop_count2a++;
#endif
			}
#ifdef TIMER
time_encode2a += GetTickCount() - time_start2;
#endif
		} else {	// ubNݒ
			if (th_id == 0){
				work_buf = th->buf;
				if (i == 0)	// ŏvZꍇ̓ubN 0Ŗ߂
					memset(work_buf, 0, unit_size);
			} else {
				memset(work_buf, 0, unit_size);	// ubN 0Ŗ߂
			}
			left_num = source_num - i;	// cubN
			left_max = chunk_num * left_num;
			// \[XEubNƂɊ|ZĒǉĂ
			while ((j = InterlockedIncrement(&(th->now))) < left_max){	// j = ++th_now
				off = j / left_num;	// chunk ̔ԍ
				j = j % left_num;	// source ̔ԍ
				off *= chunk_size;
				if (off + len > unit_size)
					len = unit_size - off;	// Ō chunk TCYقȂ邩
				galois_align_multiply(buf + ((size_t)unit_size * (i + j) + off), work_buf + off, len, factor[i + j]);
#ifdef TIMER
loop_count2b++;
#endif
			}
#ifdef TIMER
time_encode2b += GetTickCount() - time_start2;
#endif
		}
		_mm_sfence();	// [ւ݂̏
		SetEvent(hEnd);	// vZIʒm
		WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	}
#ifdef TIMER
printf("sub-thread[%d] : total loop = %d\n", th_id, loop_count2a + loop_count2b);
if (time_encode2a > 0){
	i = (int)((__int64)loop_count2a * chunk_size * 125 / ((__int64)time_encode2a * 131072));
} else {
	i = 0;
}
if (loop_count2a > 0)
	printf(" 1st decode %d.%03d sec, %d loop, %d MB/s\n", time_encode2a / 1000, time_encode2a % 1000, loop_count2a, i);
if (time_encode2b > 0){
	i = (int)((__int64)loop_count2b * chunk_size * 125 / ((__int64)time_encode2b * 131072));
} else {
	i = 0;
}
printf(" 2nd decode %d.%03d sec, %d loop, %d MB/s\n", time_encode2b / 1000, time_encode2b % 1000, loop_count2b, i);
#endif

	// I
	CloseHandle(hRun);
	CloseHandle(hEnd);
	return 0;
}

static DWORD WINAPI thread_decode3(LPVOID lpParameter)
{
	unsigned char *buf, *p_buf, *work_buf;
	unsigned short *factor;
	int i, j, th_id, block_lost, chunk_num, max_num, left_num, left_max;
	unsigned int unit_size, len, off, chunk_size;
	HANDLE hRun, hEnd;
	RS_TH *th;
#ifdef TIMER
unsigned int loop_count2a = 0, loop_count2b = 0;
unsigned int time_start2, time_encode2a = 0, time_encode2b = 0;
#endif

	th = (RS_TH *)lpParameter;
	buf = th->buf;
	block_lost = th->size;
	left_num = th->count;
	chunk_size = th->off;
	th_id = th->now;	// Xbhԍ
	hRun = th->run;
	hEnd = th->end;
	_mm_sfence();
	SetEvent(hEnd);	// ݒ芮ʒm

	unit_size = (block_size + HASH_SIZE + (sse_unit - 1)) & ~(sse_unit - 1);
	p_buf = buf + (size_t)unit_size * (left_num + cpu_num - 1);
	chunk_num = (unit_size + chunk_size - 1) / chunk_size;
	max_num = chunk_num * block_lost;
	if (th_id > 0)
		work_buf = buf + (size_t)unit_size * (left_num + th_id - 1);

	WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	while (th->now < INT_MAX / 2){
#ifdef TIMER
time_start2 = GetTickCount();
#endif
		i = th->off;	// \[XEubNԍ
		factor = th->mat;
		len = chunk_size;
		if (th->buf == NULL){	// \[XEubNǂݍݒ
			left_num = th->count + i;
			// ubNƂɊ|ZĒǉĂ
			while ((j = InterlockedIncrement(&(th->now))) < max_num){	// j = ++th_now
				off = j / block_lost;	// chunk ̔ԍ
				j = j % block_lost;		// lost block ̔ԍ
				off *= chunk_size;
				if (off + len > unit_size)
					len = unit_size - off;	// Ō chunk TCYقȂ邩
				if (left_num == 0)
					memset(p_buf + ((size_t)unit_size * j + off), 0, len);	// ubN 0Ŗ߂
				galois_align_multiply(buf + ((size_t)unit_size * i + off), p_buf + ((size_t)unit_size * j + off), len, factor[source_num * j]);
#ifdef TIMER
loop_count2a++;
#endif
			}
#ifdef TIMER
time_encode2a += GetTickCount() - time_start2;
#endif
		} else {	// ubNݒ
			if (th_id == 0){
				work_buf = th->buf;
			} else {
				memset(work_buf, 0, unit_size);	// ubN 0Ŗ߂
			}
			left_num = th->count - i;	// cubN
			left_max = chunk_num * left_num;
			// \[XEubNƂɊ|ZĒǉĂ
			while ((j = InterlockedIncrement(&(th->now))) < left_max){	// j = ++th_now
				off = j / left_num;	// chunk ̔ԍ
				j = j % left_num;	// source ̔ԍ
				off *= chunk_size;
				if (off + len > unit_size)
					len = unit_size - off;	// Ō chunk TCYقȂ邩
				galois_align_multiply(buf + ((size_t)unit_size * (i + j) + off), work_buf + off, len, factor[i + j]);
#ifdef TIMER
loop_count2b++;
#endif
			}
#ifdef TIMER
time_encode2b += GetTickCount() - time_start2;
#endif
		}
		_mm_sfence();	// [ւ݂̏
		SetEvent(hEnd);	// vZIʒm
		WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	}
#ifdef TIMER
printf("sub-thread[%d] : total loop = %d\n", th_id, loop_count2a + loop_count2b);
if (time_encode2a > 0){
	i = (int)((__int64)loop_count2a * chunk_size * 125 / ((__int64)time_encode2a * 131072));
} else {
	i = 0;
}
if (loop_count2a > 0)
	printf(" 1st encode %d.%03d sec, %d loop, %d MB/s\n", time_encode2a / 1000, time_encode2a % 1000, loop_count2a, i);
if (time_encode2b > 0){
	i = (int)((__int64)loop_count2b * chunk_size * 125 / ((__int64)time_encode2b * 131072));
} else {
	i = 0;
}
printf(" 2nd encode %d.%03d sec, %d loop, %d MB/s\n", time_encode2b / 1000, time_encode2b % 1000, loop_count2b, i);
#endif

	// I
	CloseHandle(hRun);
	CloseHandle(hEnd);
	return 0;
}

// ubNƂɌvZ邽߂̃Xbh
static DWORD WINAPI thread_decode_each(LPVOID lpParameter)
{
	unsigned char *s_buf, *p_buf, *work_buf;
	unsigned short *factor, *factor2;
	int i, j, th_id, block_lost, src_start, src_num, *last_id, max_num;
	unsigned int unit_size, len, off, chunk_size;
	HANDLE hRun, hEnd;
	RS_TH *th;
#ifdef TIMER
unsigned int loop_count2a = 0, loop_count2b = 0;
unsigned int time_start2, time_encode2a = 0, time_encode2b = 0;
#endif

	th = (RS_TH *)lpParameter;
	last_id = (int *)(th->mat);
	p_buf = th->buf;
	unit_size = th->size;
	chunk_size = th->off;
	block_lost = th->count;
	th_id = th->now;	// Xbhԍ
	hRun = th->run;
	hEnd = th->end;
	_mm_sfence();
	SetEvent(hEnd);	// ݒ芮ʒm

	last_id += th_id;	// ŌɌvZubNԍL^Ă
	max_num = ((unit_size + chunk_size - 1) / chunk_size) * block_lost;

	WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	while (th->now < INT_MAX / 2){
#ifdef TIMER
time_start2 = GetTickCount();
#endif
		s_buf = th->buf;
		factor = th->mat;

		if (th->size == 0xFFFFFFFF){	// \[XEubNǂݍݒ
			src_start = th->off;	// \[XEubNԍ
			len = chunk_size;
			// peBEubNƂɊ|ZĒǉĂ
			while ((j = InterlockedIncrement(&(th->now))) < max_num){	// j = ++th_now
				off = j / block_lost;	// chunk ̔ԍ
				j = j % block_lost;		// lost block ̔ԍ
				off *= chunk_size;
				if (off + len > unit_size)
					len = unit_size - off;	// Ō chunk TCYقȂ邩
				if (src_start == 0)	// ŏ̃ubNvZۂ
					memset(p_buf + ((size_t)unit_size * j + off), 0, len);	// ubN 0Ŗ߂
				galois_align_multiply(s_buf + off, p_buf + ((size_t)unit_size * j + off), len, factor[source_num * j]);
#ifdef TIMER
loop_count2a++;
#endif
			}
#ifdef TIMER
time_encode2a += GetTickCount() - time_start2;
#endif
		} else {
			// XbhƂɍ쐬peBEubNς
			src_num = th->count;
			while ((j = InterlockedIncrement(&(th->now))) < block_lost){	// j = ++th_now
				work_buf = p_buf + (size_t)unit_size * j;
				factor2 = factor + source_num * j;

				// chunk ɕČvZ
				len = chunk_size;
				off = 0;
				while (off < unit_size){
					// \[XEubNƂɃpeBǉĂ
					for (i = 0; i < src_num; i++)
						galois_align_multiply(s_buf + ((size_t)unit_size * i + off), work_buf, len, factor2[i]);

					work_buf += len;
					off += len;
					if (off + len > unit_size)
						len = unit_size - off;
				}

				InterlockedExchange(last_id, j);	// vZIubN̔ԍ
#ifdef TIMER
loop_count2b += src_num;
#endif
			}
#ifdef TIMER
time_encode2b += GetTickCount() - time_start2;
#endif
		}
		_mm_sfence();	// [ւ݂̏
		SetEvent(hEnd);	// vZIʒm
		WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	}
#ifdef TIMER
loop_count2a /= (unit_size + chunk_size - 1) / chunk_size;	// chunkŊăubNɂ
printf("sub-thread[%d] : total loop = %d\n", th_id, loop_count2a + loop_count2b);
if (time_encode2a > 0){
	i = (int)((__int64)loop_count2a * unit_size * 125 / ((__int64)time_encode2a * 131072));
} else {
	i = 0;
}
if (loop_count2a > 0)
	printf(" 1st decode %d.%03d sec, %d loop, %d MB/s\n", time_encode2a / 1000, time_encode2a % 1000, loop_count2a, i);
if (time_encode2b > 0){
	i = (int)((__int64)loop_count2b * unit_size * 125 / ((__int64)time_encode2b * 131072));
} else {
	i = 0;
}
printf(" 2nd decode %d.%03d sec, %d loop, %d MB/s\n", time_encode2b / 1000, time_encode2b % 1000, loop_count2b, i);
#endif

	// I
	CloseHandle(hRun);
	CloseHandle(hEnd);
	return 0;
}

// GPU Ή̃TuEXbh (Xbhԍ͍ŌɂȂ)
static DWORD WINAPI thread_decode_gpu(LPVOID lpParameter)
{
	unsigned char *s_buf, *p_buf;
	unsigned short *factor;
	int i, j, block_lost, src_num, *last_id;
	unsigned int unit_size;
	HANDLE hRun, hEnd;
	RS_TH *th;
#ifdef TIMER
unsigned int time_start2, time_encode2 = 0, loop_count2 = 0;
#endif

	th = (RS_TH *)lpParameter;
	last_id = (int *)(th->mat);
	last_id += th->now;	// ŌɌvZubNԍL^Ă
	p_buf = th->buf;
	unit_size = th->size;
	block_lost = th->count;
	hRun = th->run;
	hEnd = th->end;
	_mm_sfence();
	SetEvent(hEnd);	// ݒ芮ʒm

	WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	while (th->now < INT_MAX / 2){
#ifdef TIMER
time_start2 = GetTickCount();
#endif
		// GPU̓\[XEubNǂݍݒɌĂ΂Ȃ
		s_buf = th->buf;
		factor = th->mat;
		src_num = th->count;

		// ŏɃ\[XEubNVRAM֓]
		i = gpu_copy_blocks(s_buf, unit_size, src_num);
		if (i != 0){
			th->size = i;
			InterlockedExchange(&(th->now), INT_MAX / 2);	// TuEXbȟvZ𒆒f
		}

		// XbhƂɍ쐬peBEubNς
		while ((j = InterlockedIncrement(&(th->now))) < block_lost){	// j = ++th_now
			// {͋ts񂩂畔IɃRs[
			i = gpu_multiply_blocks(src_num, factor + source_num * j, p_buf + (size_t)unit_size * j, unit_size);
			if (i != 0){
				th->size = i;
				break;
			}

			InterlockedExchange(last_id, j);	// vZIubN̔ԍ
#ifdef TIMER
loop_count2 += src_num;
#endif
		}
#ifdef TIMER
time_encode2 += GetTickCount() - time_start2;
#endif
		// ŌVRAM
		th->size = gpu_finish();

		_mm_sfence();	// [ւ݂̏
		SetEvent(hEnd);	// vZIʒm
		WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	}
#ifdef TIMER
printf("gpu-thread    : total loop = %d\n", loop_count2);
if (time_encode2 > 0){
	i = (int)((__int64)loop_count2 * unit_size * 125 / ((__int64)time_encode2 * 131072));
} else {
	i = 0;
}
printf(" 2nd encode %d.%03d sec, %d loop, %d MB/s\n", time_encode2 / 1000, time_encode2 % 1000, loop_count2, i);
#endif

	// I
	CloseHandle(hRun);
	CloseHandle(hEnd);
	return 0;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int decode_method1(	// \[XEubN̏ꍇ
	wchar_t *file_path,
	HANDLE *rcv_hFile,		// JoEt@C̃nh
	file_ctx_r *files,		// \[XEt@C̏
	source_ctx_r *s_blk,	// \[XEubN̏
	parity_ctx_r *p_blk)	// peBEubN̏
{
	unsigned char *buf = NULL, *work_buf, *hash;
	int err = 0, id;
	unsigned int io_size, unit_size, len, block_off;
	unsigned int time_last, prog_num = 0, prog_base;
	__int64 file_off;
	HANDLE hFile = NULL;

	// ƃobt@[mۂ
	len = 0;
	io_size = get_io_size(2, &len, 1, sse_unit);
	//io_size = (((io_size + 2) / 3 + HASH_SIZE + (sse_unit - 1)) & ~(sse_unit - 1)) - HASH_SIZE;	// p
	unit_size = io_size + HASH_SIZE;	// `FbNT̕₷
	len = 2 * unit_size + HASH_SIZE;
	buf = _aligned_malloc(len, sse_unit);
	if (buf == NULL){
		printf("malloc, %d\n", len);
		err = 1;
		goto error_end;
	}
	work_buf = buf + unit_size;
	hash = work_buf + unit_size;
	prog_base = (block_size + io_size - 1) / io_size;	// fЂ̌
#ifdef TIMER
	printf("\n read one block, and keep one recovering block\n");
	printf("buffer size = %d MB, io_size = %d, split = %d\n", len >> 20, io_size, (block_size + io_size - 1) / io_size);
	id = try_cache_blocking(unit_size);
	printf("cache: limit size = %d, chunk_size = %d, split = %d\n", cpu_flag & 0x7FFF8000, id, (unit_size + id - 1) / id);
#endif

	// Ɨp̃\[XEt@CJ
	wcscpy(file_path, base_dir);
	hFile = handle_temp_file(list_buf + files[s_blk[0].file].name, file_path);
	if (hFile == INVALID_HANDLE_VALUE){
		hFile = NULL;
		err = 1;
		goto error_end;
	}

	// Ԃ̃peBEubNg
	for (id = 0; id < parity_num; id++){
		if (p_blk[id].exist == 1)
			break;
	}
	//printf("parity_num = %d, id = %d\n", parity_num, id);

	// obt@[ETCYƂɃ\[XEubN𕜌
	print_progress_text(0, "Recovering slice");
	time_last = GetTickCount();
	block_off = 0;
	while (block_off < block_size){
#ifdef TIMER
time_start = GetTickCount();
#endif
		// peBEubNǂݍ
		len = block_size - block_off;
		if (len > io_size)
			len = io_size;
		file_off = p_blk[id].off + (__int64)block_off;
		if (file_read_data(rcv_hFile[p_blk[id].file], file_off, buf, len)){
			printf("file_read_data, recovery slice %d\n", id);
			err = 1;
			goto error_end;
		}
		if (len < io_size)
			memset(buf + len, 0, io_size - len);
		// peBEubÑ`FbNTvZ
		checksum16_altmap(buf, buf + io_size, io_size);
#ifdef TIMER
time_read += GetTickCount() - time_start;
#endif

#ifdef TIMER
time_start = GetTickCount();
#endif
		// ꂽ\[XEubN𕜌
		memset(work_buf, 0, unit_size);
		// factor Ŋƌɖ߂
		galois_align_multiply(buf, work_buf, unit_size, galois_divide(1, galois_power(2, id)));
#ifdef TIMER
time_calc += GetTickCount() - time_start;
#endif

		// oߕ\
		prog_num++;
		if (GetTickCount() - time_last >= UPDATE_TIME){
			if (print_progress((prog_num * 1000) / prog_base)){
				err = 2;
				goto error_end;
			}
			time_last = GetTickCount();
		}

#ifdef TIMER
time_start = GetTickCount();
#endif
		// ꂽ\[XEubÑ`FbNT؂
		checksum16_return(work_buf, hash, io_size);
		if (memcmp(work_buf + io_size, hash, HASH_SIZE) != 0){
			printf("checksum mismatch, recovered input slice %d\n", 0);
			err = 1;
			goto error_end;
		}
		// Ɨp̃\[XEt@Cɏ
		len = s_blk[0].size - block_off;
		if (len > io_size)
			len = io_size;
		if (file_write_data(hFile, (__int64)block_off, work_buf, len)){
			printf("file_write_data, input slice %d\n", 0);
			err = 1;
			goto error_end;
		}
#ifdef TIMER
time_write += GetTickCount() - time_start;
#endif

		block_off += io_size;
	}
	print_progress_done();	// ubN̒fЉɂĂ 100% ŊƂ͌Ȃ

#ifdef TIMER
printf("read   %d.%03d sec\n", time_read / 1000, time_read % 1000);
printf("write  %d.%03d sec\n", time_write / 1000, time_write % 1000);
printf("decode %d.%03d sec\n", time_calc / 1000, time_calc % 1000);
#endif

error_end:
	if (hFile)
		CloseHandle(hFile);
	if (buf)
		_aligned_free(buf);
	return err;
}

int decode_method2(	// \[XEf[^Sēǂݍޏꍇ
	wchar_t *file_path,
	int block_lost,			// ꂽ\[XEubN̐
	HANDLE *rcv_hFile,		// JoEt@C̃nh
	file_ctx_r *files,		// \[XEt@C̏
	source_ctx_r *s_blk,	// \[XEubN̏
	parity_ctx_r *p_blk,		// peBEubN̏
	unsigned short *mat)
{
	unsigned char *buf = NULL, *p_buf, *work_buf, *sub_buf, *hash;
	unsigned short *id;
	int err = 0, i, j, k, last_file, part_num, recv_now;
	unsigned int io_size, unit_size, len, block_off;
	unsigned int time_last;
	__int64 file_off, prog_num = 0, prog_base;
	HANDLE hFile = NULL;
	HANDLE hSub[MAX_CPU], hRun[MAX_CPU], hEnd[MAX_CPU];
	RS_TH th[1];

	memset(hSub, 0, sizeof(HANDLE) * (MAX_CPU));
	id = mat + (block_lost * source_num);	// Ԗڂ̏\[XEubNǂ̃peBőւ邩

	// ƃobt@[mۂ
	part_num = source_num >> PART_MAX_RATE;	// \[XEubNɑ΂銄ōőʂ߂
	if (part_num > block_lost)
		part_num = block_lost;
	i = source_num + (cpu_num - 1);	// \[XEubNƃTuEXbh̍Ɨ̈悪Kv
	io_size = get_io_size(i, &part_num, 1, sse_unit);
	//io_size = (((io_size + 2) / 3 + HASH_SIZE + (sse_unit - 1)) & ~(sse_unit - 1)) - HASH_SIZE;	// p
	unit_size = io_size + HASH_SIZE;	// `FbNT̕₷
	file_off = (i + part_num) * (size_t)unit_size + HASH_SIZE;
	buf = _aligned_malloc((size_t)file_off, sse_unit);
	if (buf == NULL){
		printf("malloc, %I64d\n", file_off);
		err = 1;
		goto error_end;
	}
	sub_buf = buf + (size_t)unit_size * source_num;	// TuEXbh̍Ɨ̈
	p_buf = sub_buf + (size_t)unit_size * (cpu_num - 1);	// ubN𕔕IɋL^̈
	hash = p_buf + (size_t)unit_size * part_num;
	prog_base = (block_size + io_size - 1) / io_size;
	prog_base *= source_num * block_lost;	// Ŝ̒fЂ̌
#ifdef TIMER
	printf("\n read all blocks, and keep some recovering blocks\n");
	printf("buffer size = %I64d MB, io_size = %d, split = %d\n", file_off >> 20, io_size, (block_size + io_size - 1) / io_size);
	len = try_cache_blocking(unit_size);
	printf("cache: limit size = %d, chunk_size = %d, split = %d\n", cpu_flag & 0x7FFF8000, len, (unit_size + len - 1) / len);
	printf("prog_base = %I64d, part_num = %d\n", prog_base, part_num);
#endif

	// }`EXbh̏
	th->buf = buf;
	th->size = unit_size;
	th->count = part_num;
	th->off = try_cache_blocking(unit_size);	// LbV̍œK݂
	for (j = 0; j < cpu_num; j++){	// TuEXbhƂ
		hRun[j] = CreateEvent(NULL, FALSE, FALSE, NULL);	// Auto Reset ɂ
		if (hRun[j] == NULL){
			print_win32_err();
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		hEnd[j] = CreateEvent(NULL, TRUE, FALSE, NULL);
		if (hEnd[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		// TuEXbhN
		th->run = hRun[j];
		th->end = hEnd[j];
		th->now = j;	// Xbhԍ
		_mm_sfence();	// [ւ݂̏ĂXbhN
		hSub[j] = (HANDLE)_beginthreadex(NULL, 65536, thread_decode2, (LPVOID)th, 0, NULL);
		if (hSub[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			CloseHandle(hEnd[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		if (j >= (cpu_num * 3) / 4)	// IO ؂Ȃ悤ɁATuEXbh̗Dx
			SetThreadPriority(hSub[j], THREAD_PRIORITY_BELOW_NORMAL);
		WaitForSingleObject(hEnd[j], INFINITE);	// ݒI̍}҂ (ZbgȂ)
	}

	// obt@[ETCYƂɃ\[XEubN𕜌
	print_progress_text(0, "Recovering slice");
	time_last = GetTickCount();
	wcscpy(file_path, base_dir);
	block_off = 0;
	while (block_off < block_size){
		th->buf = NULL;
		th->off = -1;	// ܂vZĖ

#ifdef TIMER
time_start = GetTickCount();
#endif
		last_file = -1;
		recv_now = 0;	// Ԗڂ̑փubN
		for (i = 0; i < source_num; i++){
			switch(s_blk[i].exist){
			case 0:		// obt@[ɃpeBEubN̓eǂݍ
				len = block_size - block_off;
				if (len > io_size)
					len = io_size;
				file_off = p_blk[id[recv_now]].off + (__int64)block_off;
				if (file_read_data(rcv_hFile[p_blk[id[recv_now]].file], file_off, buf + (size_t)unit_size * i, len)){
					printf("file_read_data, recovery slice %d\n", id[recv_now]);
					err = 1;
					goto error_end;
				}
				if (len < io_size)
					memset(buf + ((size_t)unit_size * i + len), 0, io_size - len);
				recv_now++;
				// peBEubÑ`FbNTvZ
				checksum16_altmap(buf + (size_t)unit_size * i, buf + ((size_t)unit_size * i + io_size), io_size);
#ifdef TIMER
read_count++;
#endif
				break;
			case 3:		// \[XEubN̓e͑S 0
				len = 0;
				memset(buf + (size_t)unit_size * i, 0, unit_size);
				break;
			default:	// obt@[Ƀ\[XEubN̓eǂݍ
				if (s_blk[i].file != last_file){	// ʂ̃t@CȂJ
					last_file = s_blk[i].file;
					if (hFile){
						CloseHandle(hFile);	// Õt@C
						hFile = NULL;
					}
					if ((files[last_file].state & 7) != 0){	// 蒼ƃt@Cǂݍ
						get_temp_name(list_buf + files[last_file].name, file_path + base_len);
					} else if ((files[last_file].state & 32) != 0){	// Osɂ͕ʖt@Cǂݍ
						wcscpy(file_path + base_len, list_buf + files[last_file].name2);
					} else {	// Sȃ\[XEt@Cǂݍ
						wcscpy(file_path + base_len, list_buf + files[last_file].name);
					}
					hFile = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
					if (hFile == INVALID_HANDLE_VALUE){
						print_win32_err();
						hFile = NULL;
						printf_cp("cannot open file, %s\n", file_path);
						err = 1;
						goto error_end;
					}
				}
				if (s_blk[i].size > block_off){
					len = s_blk[i].size - block_off;
					if (len > io_size)
						len = io_size;
					file_off = (i - files[last_file].b_off) * (__int64)block_size + (__int64)block_off;
					if (file_read_data(hFile, file_off, buf + (size_t)unit_size * i, len)){
						printf("file_read_data, input slice %d\n", i);
						err = 1;
						goto error_end;
					}
					if (len < io_size)
						memset(buf + ((size_t)unit_size * i + len), 0, io_size - len);
					// \[XEubÑ`FbNTvZ
					checksum16_altmap(buf + (size_t)unit_size * i, buf + ((size_t)unit_size * i + io_size), io_size);
#ifdef TIMER
read_count++;
#endif
				} else {
					len = 0;
					memset(buf + (size_t)unit_size * i, 0, unit_size);
				}
			}

			if ((len > 0) && (i + 1 < source_num)){	// Ō̃ubNȊOȂ
				// TuEXbh̓󋵂𒲂ׂ
				j = WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, 0);
				if ((j != WAIT_TIMEOUT) && (j != WAIT_FAILED)){	// vZłȂȂ
					// oߕ\
					prog_num += part_num;
					if (GetTickCount() - time_last >= UPDATE_TIME){
						if (print_progress((int)((prog_num * 1000) / prog_base))){
							err = 2;
							goto error_end;
						}
						time_last = GetTickCount();
					}
					// vZIubN̎vZJn
					th->off += 1;
					if (th->off > 0){	// obt@[ɓǂݍ񂾎vZ
						while ((s_blk[th->off].exist != 0) &&
								((s_blk[th->off].size <= block_off) || (s_blk[th->off].exist == 3))){
							prog_num += part_num;
							th->off += 1;
#ifdef TIMER
skip_count++;
#endif
						}
					}
					th->mat = mat + th->off;
					th->now = -1;	// l - 1
					_mm_sfence();	// [ւ݂̏ĂXbhĊJ
					for (j = 0; j < (cpu_num + 1) / 2; j++){	// ǂݍݒ̓Xbh炷
						ResetEvent(hEnd[j]);	// ZbgĂ
						SetEvent(hRun[j]);	// TuEXbhɌvZJn
					}
				}
			}
		}
		if (hFile){	// Ō̃\[XEt@C
			CloseHandle(hFile);
			hFile = NULL;
		}
#ifdef TIMER
time_read += GetTickCount() - time_start;
#endif

		WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, INFINITE);	// TuEXbȟvZI̍}҂
		th->off += 1;	// vZJn\[XEubN̔ԍ
		if (th->off > 0){	// vZsvȃ\[XEubN͂Ƃ΂
			while ((s_blk[th->off].exist != 0) &&
					((s_blk[th->off].size <= block_off) || (s_blk[th->off].exist == 3))){
				prog_num += part_num;
				th->off += 1;
#ifdef TIMER
skip_count++;
#endif
			}
		}
#ifdef TIMER
		j = (int)(((__int64)(th->off * part_num) * 1000) / (source_num * block_lost));
		printf("partial decode = %d (%d.%d%%), read = %d, skip = %d\n", th->off, j / 10, j % 10, read_count, skip_count);
#endif

		// ŏ̏ubNT
		recv_now = 0;	// Ԗڂ̏ubN
		for (i = 0; i < source_num; i++){
			if (s_blk[i].exist == 0){
				if (s_blk[i].size > block_off){
					if (recv_now >= part_num){	// IȃGR[h{ӏȍ~Ȃ
						work_buf = p_buf;
					} else {
						work_buf = p_buf + (size_t)unit_size * recv_now;
					}
					//printf("%d: s_blk[%d].size = %d, off = %d\n", recv_now, i, s_blk[i].size, th->off);
					// oߕ\
					prog_num += source_num - th->off;
					if (GetTickCount() - time_last >= UPDATE_TIME){
						if (print_progress((int)((prog_num * 1000) / prog_base))){
							err = 2;
							goto error_end;
						}
						time_last = GetTickCount();
					}
					// ŏ̃ubN̕JnĂ
					th->mat = mat + source_num * recv_now;
					th->buf = work_buf;
					th->now = -1;	// l - 1
					_mm_sfence();
					for (j = 0; j < cpu_num; j++){
						ResetEvent(hEnd[j]);	// ZbgĂ
						SetEvent(hRun[j]);	// TuEXbhɌvZJn
					}
					break;
				} else {
					//printf("%d: s_blk[%d] = empty, off = %d\n", recv_now, i, th->off);
					prog_num += source_num - th->off;
					recv_now++;	// KvĂԍ͐i߂
					if (recv_now == part_num)	// IȃGR[h{ӏȍ~Ȃ
						th->off = 0;	// ŏ̃\[XEubNvZ
				}
			}
		}

		// ꂽ\[XEubNƂɕ
		last_file = -1;
		while (recv_now < block_lost){
			recv_now++;	// ̏ubN̔ԍɂ
			// ubN̕Î҂Ċ
			WaitForMultipleObjects(cpu_num, hEnd, TRUE, INFINITE);	// TuEXbȟvZI̍}҂
			for (j = 1; j < cpu_num; j++){
				// ŌɃTuEXbh̃obt@[ XOR 
				galois_align_xor(sub_buf + (size_t)unit_size * (j - 1), work_buf, unit_size);
			}

			if (recv_now < block_lost){	// Ō̃ubNȊOȂ
				// ̏ubNT
				for (k = i + 1; k < source_num; k++){
					if (s_blk[k].exist == 0){
						if (s_blk[k].size > block_off)
							break;
						if (recv_now >= part_num)	// IȃGR[h{ӏȍ~Ȃ
							th->off = 0;	// ŏ̃\[XEubNvZ
						//printf("%d: s_blk[%d] = empty, off = %d\n", recv_now, k, th->off);
						prog_num += source_num - th->off;
						recv_now++;	// KvĂԍ͐i߂
					}
				}
				if (k < source_num){	// ubN݂Ȃ
					if (recv_now >= part_num){	// IȃGR[h{ӏȍ~Ȃ
						if (th->buf != p_buf){	// GR[hƏ݂Ń_uEobt@O
							th->buf = p_buf;
						} else {
							th->buf += unit_size;
						}
						th->off = 0;	// ŏ̃\[XEubNvZ
					} else {
						th->buf = p_buf + (size_t)unit_size * recv_now;
					}
					//printf("%d: s_blk[%d].size = %d, off = %d\n", recv_now, k, s_blk[k].size, th->off);
					// oߕ\
					prog_num += source_num - th->off;
					if (GetTickCount() - time_last >= UPDATE_TIME){
						if (print_progress((int)((prog_num * 1000) / prog_base))){
							err = 2;
							goto error_end;
						}
						time_last = GetTickCount();
					}
					// ̃ubN̕JnĂ
					th->mat = mat + source_num * recv_now;
					th->now = -1;	// l - 1
					_mm_sfence();	// [ւ݂̏ĂXbhĊJ
					for (j = 0; j < cpu_num; j++){
						ResetEvent(hEnd[j]);	// ZbgĂ
						SetEvent(hRun[j]);	// TuEXbhɌvZJn
					}
				}
			}

#ifdef TIMER
time_start = GetTickCount();
#endif
			// ꂽ\[XEubÑ`FbNT؂
			checksum16_return(work_buf, hash, io_size);
			if (memcmp(work_buf + io_size, hash, HASH_SIZE) != 0){
				printf("checksum mismatch, recovered input slice %d\n", i);
				err = 1;
				goto error_end;
			}
			// Ɨp̃\[XEt@Cɏ
			if (s_blk[i].file != last_file){	// ʂ̃t@CȂJ
				last_file = s_blk[i].file;
				if (hFile){
					CloseHandle(hFile);	// Õt@C
					hFile = NULL;
				}
				hFile = handle_temp_file(list_buf + files[last_file].name, file_path);
				if (hFile == INVALID_HANDLE_VALUE){
					hFile = NULL;
					err = 1;
					goto error_end;
				}
				//printf("file %d, open %S\n", last_file, file_path);
			}
			// \[XEt@Cł̃ubNfЂ̑傫ƈʒu
			len = s_blk[i].size - block_off;
			if (len > io_size)
				len = io_size;
			if (file_write_data(hFile, (i - files[last_file].b_off) * (__int64)block_size + block_off, work_buf, len)){
				printf("file_write_data, input slice %d\n", i);
				err = 1;
				goto error_end;
			}
#ifdef TIMER
time_write += GetTickCount() - time_start;
write_count++;
#endif

			if (recv_now >= part_num) {	// IȃGR[h{ӏȍ~Ȃ
				if (work_buf != p_buf){	// GR[hƏ݂Ń_uEobt@O
					work_buf = p_buf;
				} else {
					work_buf += unit_size;
				}
			} else {
				work_buf = p_buf + (size_t)unit_size * recv_now;
			}
			i = k;	// ̏ubN̔ԍ
		}

		block_off += io_size;
		// Ō̍ƃt@C
		CloseHandle(hFile);
		hFile = NULL;
	}
	print_progress_done();	// ubN̒fЉɂĂ 100% ŊƂ͌Ȃ

#ifdef TIMER
printf("read   %d.%03d sec\n", time_read / 1000, time_read % 1000);
j = ((block_size + io_size - 1) / io_size) * block_lost;
printf("write  %d.%03d sec, count = %d/%d\n", time_write / 1000, time_write % 1000, write_count, j);
#endif

error_end:
	InterlockedExchange(&(th->now), INT_MAX / 2);	// TuEXbȟvZ𒆒f
	for (j = 0; j < cpu_num; j++){
		if (hSub[j]){	// TuEXbhI
			SetEvent(hRun[j]);
			WaitForSingleObject(hSub[j], INFINITE);
			CloseHandle(hSub[j]);
		}
	}
	if (hFile)
		CloseHandle(hFile);
	if (buf)
		_aligned_free(buf);
	return err;
}

int decode_method3(	// ubNSĕێłꍇ
	wchar_t *file_path,
	int block_lost,			// ꂽ\[XEubN̐
	HANDLE *rcv_hFile,		// JoEt@C̃nh
	file_ctx_r *files,		// \[XEt@C̏
	source_ctx_r *s_blk,	// \[XEubN̏
	parity_ctx_r *p_blk,	// peBEubN̏
	unsigned short *mat)
{
	unsigned char *buf = NULL, *p_buf, *work_buf, *sub_buf, *hash;
	unsigned short *id;
	int err = 0, i, j, k, last_file, source_off, read_num, recv_now, parity_now;
	unsigned int unit_size, len;
	unsigned int time_last, prog_num = 0, prog_base;
	__int64 file_off;
	HANDLE hFile = NULL;
	HANDLE hSub[MAX_CPU], hRun[MAX_CPU], hEnd[MAX_CPU];
	RS_TH th[1];

	memset(hSub, 0, sizeof(HANDLE) * (MAX_CPU));
	id = mat + (block_lost * source_num);	// Ԗڂ̏\[XEubNǂ̃peBőւ邩
	unit_size = (block_size + HASH_SIZE + (sse_unit - 1)) & ~(sse_unit - 1);	// `FbNT̕₷

	// ƃobt@[mۂ
	read_num = read_block_num(block_lost, cpu_num - 1, 1, sse_unit);	// \[XEubNǂݍނ
	if (read_num == 0){
		//printf("cannot keep enough blocks, use another method\n");
		return -2;	// XCX𕪊ďȂƖ
	}
	//read_num = (read_num + 2) / 3 + 1;	// p
	len = read_num + block_lost + (cpu_num - 1);	// ubNƃTuEXbh̍Ɨ̈悪Kv
	file_off = len * (size_t)unit_size + HASH_SIZE;
	buf = _aligned_malloc((size_t)file_off, sse_unit);
	if (buf == NULL){
		printf("malloc, %I64d\n", file_off);
		err = 1;
		goto error_end;
	}
	sub_buf = buf + (size_t)unit_size * read_num;	// TuEXbh̍Ɨ̈
	p_buf = sub_buf + (size_t)unit_size * (cpu_num - 1);	// peBEubNL^̈
	hash = p_buf + (size_t)unit_size * block_lost;
	prog_base = source_num * block_lost;	// ubN̍v|Z
#ifdef TIMER
	printf("\n read some blocks, and keep all recovering blocks\n");
	printf("buffer size = %I64d MB, read_num = %d, round = %d\n", file_off >> 20, read_num, (source_num + read_num - 1) / read_num);
	len = try_cache_blocking(unit_size);	// LbV̍œK݂
	printf("cache: limit size = %d, chunk_size = %d, split = %d\n", cpu_flag & 0x7FFF8000, len, (unit_size + len - 1) / len);
	printf("prog_base = %d\n", prog_base);
#endif

	// }`EXbh̏
	th->buf = buf;
	th->size = block_lost;
	th->count = read_num;
	th->off = try_cache_blocking(unit_size);	// LbV̍œK݂
	for (j = 0; j < cpu_num; j++){	// TuEXbhƂ
		hRun[j] = CreateEvent(NULL, FALSE, FALSE, NULL);	// Auto Reset ɂ
		if (hRun[j] == NULL){
			print_win32_err();
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		hEnd[j] = CreateEvent(NULL, TRUE, FALSE, NULL);
		if (hEnd[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		// TuEXbhN
		th->run = hRun[j];
		th->end = hEnd[j];
		th->now = j;	// Xbhԍ
		_mm_sfence();	// [ւ݂̏ĂXbhN
		hSub[j] = (HANDLE)_beginthreadex(NULL, 65536, thread_decode3, (LPVOID)th, 0, NULL);
		if (hSub[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			CloseHandle(hEnd[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		if (j >= (cpu_num * 3) / 4)	// IO ؂Ȃ悤ɁATuEXbh̗Dx
			SetThreadPriority(hSub[j], THREAD_PRIORITY_BELOW_NORMAL);
		WaitForSingleObject(hEnd[j], INFINITE);	// ݒI̍}҂ (ZbgȂ)
	}

	// 񂩂ɕʂăubNǂݍŁA\[XEubN
	print_progress_text(0, "Recovering slice");
	time_last = GetTickCount();
	wcscpy(file_path, base_dir);
	parity_now = 0;	// Ԗڂ̑փubN
	source_off = 0;	// ǂݍ݊JnXCXԍ
	while (source_off < source_num){
		if (read_num > source_num - source_off)
			read_num = source_num - source_off;
		th->buf = NULL;
		th->count = source_off;
		th->off = -1;	// ܂vZĖ

#ifdef TIMER
time_start = GetTickCount();
#endif
		last_file = -1;
		for (i = 0; i < read_num; i++){	// XCXǂݍŃ[ɔzuĂ
			switch(s_blk[source_off + i].exist){
			case 0:		// obt@[ɃpeBEubN̓eǂݍ
				if (file_read_data(rcv_hFile[p_blk[id[parity_now]].file], p_blk[id[parity_now]].off, buf + (size_t)unit_size * i, block_size)){
					printf("file_read_data, recovery slice %d\n", id[parity_now]);
					err = 1;
					goto error_end;
				}
				parity_now++;
				// peBEubÑ`FbNTvZ
				checksum16_altmap(buf + (size_t)unit_size * i, buf + ((size_t)unit_size * i + unit_size - HASH_SIZE), unit_size - HASH_SIZE);
#ifdef TIMER
read_count++;
#endif
				break;
			case 3:		// \[XEubN̓e͑S 0
				memset(buf + (size_t)unit_size * i, 0, unit_size);
				break;
			default:	// obt@[Ƀ\[XEubN̓eǂݍ
				if (s_blk[source_off + i].file != last_file){	// ʂ̃t@CȂJ
					last_file = s_blk[source_off + i].file;
					if (hFile){
						CloseHandle(hFile);	// Õt@C
						hFile = NULL;
					}
					if ((files[last_file].state & 7) != 0){	// 蒼ƃt@Cǂݍ
						get_temp_name(list_buf + files[last_file].name, file_path + base_len);
					} else if ((files[last_file].state & 32) != 0){	// Osɂ͕ʖt@Cǂݍ
						wcscpy(file_path + base_len, list_buf + files[last_file].name2);
					} else {	// Sȃ\[XEt@Cǂݍ (ǉs)
						wcscpy(file_path + base_len, list_buf + files[last_file].name);
					}
					hFile = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
					if (hFile == INVALID_HANDLE_VALUE){
						print_win32_err();
						hFile = NULL;
						printf_cp("cannot open file, %s\n", file_path);
						err = 1;
						goto error_end;
					}
				}
				len = s_blk[source_off + i].size;
				file_off = (source_off + i - files[last_file].b_off) * (__int64)block_size;
				if (file_read_data(hFile, file_off, buf + (size_t)unit_size * i, len)){
					printf("file_read_data, input slice %d\n", source_off + i);
					err = 1;
					goto error_end;
				}
				if (len < block_size)
					memset(buf + ((size_t)unit_size * i + len), 0, block_size - len);
				// \[XEubÑ`FbNTvZ
				checksum16_altmap(buf + (size_t)unit_size * i, buf + ((size_t)unit_size * i + unit_size - HASH_SIZE), unit_size - HASH_SIZE);
#ifdef TIMER
read_count++;
#endif
			}

			if (i + 1 < read_num){	// Ō̃ubNȊOȂ
				// TuEXbh̓󋵂𒲂ׂ
				j = WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, 0);
				if ((j != WAIT_TIMEOUT) && (j != WAIT_FAILED)){	// vZłȂȂ
					// oߕ\
					prog_num += block_lost;
					if (GetTickCount() - time_last >= UPDATE_TIME){
						if (print_progress(((__int64)prog_num * 1000) / prog_base)){
							err = 2;
							goto error_end;
						}
						time_last = GetTickCount();
					}
					// vZIubN̎vZJn
					th->off += 1;
					th->mat = mat + (th->off + source_off);
					th->now = -1;	// l - 1
					_mm_sfence();	// [ւ݂̏ĂXbhĊJ
					for (j = 0; j < (cpu_num + 1) / 2; j++){	// ǂݍݒ̓Xbh炷
						ResetEvent(hEnd[j]);	// ZbgĂ
						SetEvent(hRun[j]);	// TuEXbhɌvZJn
					}
				}
			}
		}
		if (hFile){	// Ō̃\[XEt@C
			CloseHandle(hFile);
			hFile = NULL;
		}
#ifdef TIMER
time_read += GetTickCount() - time_start;
#endif

		WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, INFINITE);	// TuEXbȟvZI̍}҂
		th->off += 1;	// vZJn\[XEubN̔ԍ
		if (source_off + th->off == 0)	// G[ȊO th->off  0 ɂȂȂ
			memset(p_buf, 0, (size_t)unit_size * block_lost);
#ifdef TIMER
		j = (th->off * 1000) / read_num;
		printf("partial decode = %d (%d.%d%%), read = %d\n", th->off, j / 10, j % 10, read_count);
#endif

		// ŏ̏ubNT
		recv_now = 0;	// Ԗڂ̏ubN
		for (i = 0; i < source_num; i++){
			if (s_blk[i].exist == 0){
				// oߕ\
				prog_num += read_num - th->off;
				if (GetTickCount() - time_last >= UPDATE_TIME){
					if (print_progress(((__int64)prog_num * 1000) / prog_base)){
						err = 2;
						goto error_end;
					}
					time_last = GetTickCount();
				}
				// ŏ̃ubN̕JnĂ
				//printf("%d: s_blk[%d].size = %d\n", recv_now, i, s_blk[i].size);
				work_buf = p_buf;
				th->count = read_num;
				th->mat = mat + source_off;
				th->buf = work_buf;
				th->now = -1;	// l - 1
				_mm_sfence();
				for (j = 0; j < cpu_num; j++){
					ResetEvent(hEnd[j]);	// ZbgĂ
					SetEvent(hRun[j]);	// TuEXbhɌvZJn
				}
				break;
			}
		}

		// ꂽ\[XEubNƂɕ
		last_file = -1;
		while (recv_now < block_lost){
			recv_now++;	// ̏ubN̔ԍɂ
			// ubN̕Î҂Ċ
			WaitForMultipleObjects(cpu_num, hEnd, TRUE, INFINITE);	// TuEXbȟvZI̍}҂
			for (j = 1; j < cpu_num; j++){
				// ŌɃTuEXbh̃obt@[ XOR 
				galois_align_xor(sub_buf + (size_t)unit_size * (j - 1), work_buf, unit_size);
			}

			if (recv_now < block_lost){	// Ō̃ubNȊOȂ
				// ̏ubNT
				for (k = i + 1; k < source_num; k++){
					if (s_blk[k].exist == 0)
						break;	// ubN͕K݂͂
				}
				// oߕ\
				prog_num += read_num - th->off;
				if (GetTickCount() - time_last >= UPDATE_TIME){
					if (print_progress(((__int64)prog_num * 1000) / prog_base)){
						err = 2;
						goto error_end;
					}
					time_last = GetTickCount();
				}
				// ̃ubN̕JnĂ
				//printf("%d: s_blk[%d].size = %d\n", recv_now, k, s_blk[k].size);
				th->mat = mat + (source_num * recv_now + source_off);	// ǂݍ񂾃\[XEubN̈ʒu܂ł炷
				th->buf += unit_size;
				th->now = -1;	// l - 1
				_mm_sfence();	// [ւ݂̏ĂXbhĊJ
				for (j = 0; j < cpu_num; j++){
					ResetEvent(hEnd[j]);	// ZbgĂ
					SetEvent(hRun[j]);	// TuEXbhɌvZJn
				}
			}

			if (source_off + read_num >= source_num){	// Ō݂̏Ȃ
#ifdef TIMER
time_start = GetTickCount();
#endif
				// ꂽ\[XEubÑ`FbNT؂
				checksum16_return(work_buf, hash, unit_size - HASH_SIZE);
				if (memcmp(work_buf + unit_size - HASH_SIZE, hash, HASH_SIZE) != 0){
					printf("checksum mismatch, recovered input slice %d\n", i);
					err = 1;
					goto error_end;
				}
				// Ɨp̃\[XEt@Cɏ
				if (s_blk[i].file != last_file){	// ʂ̃t@CȂJ
					last_file = s_blk[i].file;
					if (hFile){
						CloseHandle(hFile);	// Õt@C
						hFile = NULL;
					}
					hFile = handle_temp_file(list_buf + files[last_file].name, file_path);
					if (hFile == INVALID_HANDLE_VALUE){
						hFile = NULL;
						err = 1;
						goto error_end;
					}
					//printf("file %d, open %S\n", last_file, file_path);
				}
				// \[XEt@Cł̃ubN̑傫ƈʒu
				if (file_write_data(hFile, (i - files[last_file].b_off) * (__int64)block_size, work_buf, s_blk[i].size)){
					printf("file_write_data, input slice %d\n", i);
					err = 1;
					goto error_end;
				}
#ifdef TIMER
time_write += GetTickCount() - time_start;
#endif
			}

			work_buf += unit_size;
			i = k;	// ̏ubN̔ԍ
		}

		source_off += read_num;
		// Ō̍ƃt@C
		CloseHandle(hFile);
		hFile = NULL;
	}
	print_progress_done();	// ubN̒fЉɂĂ 100% ŊƂ͌Ȃ

#ifdef TIMER
printf("read   %d.%03d sec\n", time_read / 1000, time_read % 1000);
printf("write  %d.%03d sec\n", time_write / 1000, time_write % 1000);
#endif

error_end:
	InterlockedExchange(&(th->now), INT_MAX / 2);	// TuEXbȟvZ𒆒f
	for (j = 0; j < cpu_num; j++){
		if (hSub[j]){	// TuEXbhI
			SetEvent(hRun[j]);
			WaitForSingleObject(hSub[j], INFINITE);
			CloseHandle(hSub[j]);
		}
	}
	if (hFile)
		CloseHandle(hFile);
	if (buf)
		_aligned_free(buf);
	return err;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int decode_method4(	// SẴubNfГIɕێꍇ (GPUΉ)
	wchar_t *file_path,
	int block_lost,			// ꂽ\[XEubN̐
	HANDLE *rcv_hFile,		// JoEt@C̃nh
	file_ctx_r *files,		// \[XEt@C̏
	source_ctx_r *s_blk,	// \[XEubN̏
	parity_ctx_r *p_blk,		// peBEubN̏
	unsigned short *mat)
{
	unsigned char *buf = NULL, *p_buf, *work_buf, *hash;
	unsigned short *id;
	int err = 0, i, j, last_file, recv_now;
	int write_id, last_id_min, last_id_copy[(MAX_CPU + 1) * 2];
	volatile int *last_id;	// TuEXbh̓󋵂ێ
	int cpu_num1, cover_max, cover_from, cover_num;
	unsigned int io_size, unit_size, len, block_off;
	unsigned int time_last;
	__int64 file_off, prog_num = 0, prog_base;
	HANDLE hFile = NULL;
	HANDLE hSub[MAX_CPU + 1], hRun[MAX_CPU + 1], hEnd[MAX_CPU + 1];
	RS_TH th[1];

	memset(hSub, 0, sizeof(HANDLE) * (MAX_CPU + 1));
	id = mat + (block_lost * source_num);	// Ԗڂ̏\[XEubNǂ̃peBőւ邩
	last_id = last_id_copy + MAX_CPU + 1;

	// ƃobt@[mۂiGPU ̍Ɨ̈Ƃ2̗]TĂj
	// part_num g킸ASẴubNێ鏊decode_method2ƈقȂ邱ƂɒӁI
	io_size = get_io_size(source_num + block_lost + 2, NULL, 1, MEM_UNIT);
	//io_size = (((io_size + 1) / 2 + HASH_SIZE + (MEM_UNIT - 1)) & ~(MEM_UNIT - 1)) - HASH_SIZE;	// 2̎p
	//io_size = (((io_size + 2) / 3 + HASH_SIZE + (MEM_UNIT - 1)) & ~(MEM_UNIT - 1)) - HASH_SIZE;	// 3̎p
	unit_size = io_size + HASH_SIZE;	// `FbNT̕₷
	file_off = (source_num + block_lost) * (size_t)unit_size + HASH_SIZE;
	buf = _aligned_malloc((size_t)file_off, 4096);	// GPU p̋E
	if (buf == NULL){
		printf("malloc, %I64d\n", file_off);
		err = 1;
		goto error_end;
	}
	p_buf = buf + (size_t)unit_size * source_num;	// ubNL^̈
	hash = p_buf + (size_t)unit_size * block_lost;
	prog_base = (block_size + io_size - 1) / io_size;
	prog_base *= source_num * block_lost;	// Ŝ̒fЂ̌
#ifdef TIMER
	printf("\n read all blocks, and keep all recovering blocks (GPU)\n");
	printf("buffer size = %I64d MB, io_size = %d, split = %d\n", file_off >> 20, io_size, (block_size + io_size - 1) / io_size);
#endif

	// OpenCL ̏
	cpu_num1 = cpu_num + 1;	// GPU Ǘp̃Xbhǉ
	cover_max = source_num;
	i = init_OpenCL(unit_size, &cover_max, &len);
	if (i != 0){
		if (i != 3)	// GPU Ȃꍇ̓G[\Ȃ
			printf("init_OpenCL, %d, %d\n", i & 0xFF, i >> 8);
		i = free_OpenCL();
		if (i != 0)
			printf("free_OpenCL, %d, %d", i & 0xFF, i >> 8);
		//cpu_num1--;	// sȂꍇ͂̍sRgAEg
		// GPU g킸ɌvZ𑱍sꍇ͈ȉRgAEg
		err = -2;	// CPU ̕ɐ؂ւ
		goto error_end;
	}
	if (len == 0)	// GPULbVgȂACPUƎɃLbV̍œK݂
		len = try_cache_blocking(unit_size);
#ifdef TIMER
	printf("cache: limit size = %d, chunk_size = %d, split = %d\n", cpu_flag & 0x7FFF8000, len, (unit_size + len - 1) / len);
	printf("prog_base = %I64d, unit_size = %d, method = %d, cover_max = %d\n", prog_base, unit_size, OpenCL_method, cover_max);
#endif

	// }`EXbh̏
	th->mat = (unsigned short *)last_id;
	th->buf = p_buf;
	th->size = unit_size;
	th->count = block_lost;
	th->off = len;	// chunk size
	for (j = 0; j < cpu_num1; j++){	// TuEXbhƂ
		hRun[j] = CreateEvent(NULL, FALSE, FALSE, NULL);	// Auto Reset ɂ
		if (hRun[j] == NULL){
			print_win32_err();
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		hEnd[j] = CreateEvent(NULL, TRUE, FALSE, NULL);
		if (hEnd[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		// TuEXbhN
		th->run = hRun[j];
		th->end = hEnd[j];
		th->now = j;	// Xbhԍ
		_mm_sfence();	// [ւ݂̏ĂXbhN
		if (j == cpu_num){	// Ō̃Xbh GPU Ǘpɂ
			hSub[j] = (HANDLE)_beginthreadex(NULL, 65536, thread_decode_gpu, (LPVOID)th, 0, NULL);
		} else {
			hSub[j] = (HANDLE)_beginthreadex(NULL, 65536, thread_decode_each, (LPVOID)th, 0, NULL);
		}
		if (hSub[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			CloseHandle(hEnd[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		// IO ؂Ȃ悤ɁATuEXbh̗DxiGPUXbh͏Oj
		if ((j >= (cpu_num * 3) / 4) && (j < cpu_num))
			SetThreadPriority(hSub[j], THREAD_PRIORITY_BELOW_NORMAL);
		WaitForSingleObject(hEnd[j], INFINITE);	// ݒI̍}҂ (ZbgȂ)
	}

	// ubNfЂǂݍŁAubNfЂ𕜌
	print_progress_text(0, "Recovering slice");
	time_last = GetTickCount();
	wcscpy(file_path, base_dir);
	block_off = 0;
	while (block_off < block_size){
		th->size = 0xFFFFFFFF;	// 1st decode
		th->off = -1;	// ܂vZĖ
		write_id = 0;

#ifdef TIMER
time_start = GetTickCount();
#endif
		last_file = -1;
		recv_now = 0;	// Ԗڂ̑փubN
		for (i = 0; i < source_num; i++){
			switch(s_blk[i].exist){
			case 0:		// obt@[ɃpeBEubN̓eǂݍ
				len = block_size - block_off;
				if (len > io_size)
					len = io_size;
				file_off = p_blk[id[recv_now]].off + (__int64)block_off;
				if (file_read_data(rcv_hFile[p_blk[id[recv_now]].file], file_off, buf + (size_t)unit_size * i, len)){
					printf("file_read_data, recovery slice %d\n", id[recv_now]);
					err = 1;
					goto error_end;
				}
				if (len < io_size)
					memset(buf + ((size_t)unit_size * i + len), 0, io_size - len);
				recv_now++;
				// peBEubÑ`FbNTvZ
				checksum16_altmap(buf + (size_t)unit_size * i, buf + ((size_t)unit_size * i + io_size), io_size);
#ifdef TIMER
read_count++;
#endif
				break;
			case 3:		// \[XEubN̓e͑S 0
				len = 0;
				memset(buf + (size_t)unit_size * i, 0, unit_size);
				break;
			default:	// obt@[Ƀ\[XEubN̓eǂݍ
				if (s_blk[i].file != last_file){	// ʂ̃t@CȂJ
					last_file = s_blk[i].file;
					if (hFile){
						CloseHandle(hFile);	// Õt@C
						hFile = NULL;
					}
					if ((files[last_file].state & 7) != 0){	// 蒼ƃt@Cǂݍ
						get_temp_name(list_buf + files[last_file].name, file_path + base_len);
					} else if ((files[last_file].state & 32) != 0){	// Osɂ͕ʖt@Cǂݍ
						wcscpy(file_path + base_len, list_buf + files[last_file].name2);
					} else {	// Sȃ\[XEt@Cǂݍ
						wcscpy(file_path + base_len, list_buf + files[last_file].name);
					}
					hFile = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
					if (hFile == INVALID_HANDLE_VALUE){
						print_win32_err();
						hFile = NULL;
						printf_cp("cannot open file, %s\n", file_path);
						err = 1;
						goto error_end;
					}
				}
				if (s_blk[i].size > block_off){
					len = s_blk[i].size - block_off;
					if (len > io_size)
						len = io_size;
					file_off = (i - files[last_file].b_off) * (__int64)block_size + (__int64)block_off;
					if (file_read_data(hFile, file_off, buf + (size_t)unit_size * i, len)){
						printf("file_read_data, input slice %d\n", i);
						err = 1;
						goto error_end;
					}
					if (len < io_size)
						memset(buf + ((size_t)unit_size * i + len), 0, io_size - len);
					// \[XEubÑ`FbNTvZ
					checksum16_altmap(buf + (size_t)unit_size * i, buf + ((size_t)unit_size * i + io_size), io_size);
#ifdef TIMER
read_count++;
#endif
				} else {
					len = 0;
					memset(buf + (size_t)unit_size * i, 0, unit_size);
				}
			}

			if ((len > 0) && (i + 1 < source_num)){	// Ō̃ubNȊOȂ
				// TuEXbh̓󋵂𒲂ׂ
				j = WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, 0);
				if ((j != WAIT_TIMEOUT) && (j != WAIT_FAILED)){	// vZłȂȂ
					// oߕ\
					prog_num += block_lost;
					if (GetTickCount() - time_last >= UPDATE_TIME){
						if (print_progress((int)((prog_num * 1000) / prog_base))){
							err = 2;
							goto error_end;
						}
						time_last = GetTickCount();
					}
					// vZIubN̎vZJn
					th->off += 1;
					if (th->off > 0){	// obt@[ɓǂݍ񂾎vZ
						while ((s_blk[th->off].exist != 0) &&
								((s_blk[th->off].size <= block_off) || (s_blk[th->off].exist == 3))){
							prog_num += block_lost;
							th->off += 1;
#ifdef TIMER
skip_count++;
#endif
						}
					}
					th->buf = buf + (size_t)unit_size * th->off;
					th->mat = mat + th->off;
					th->now = -1;	// l - 1
					_mm_sfence();	// [ւ݂̏ĂXbhĊJ
					for (j = 0; j < (cpu_num + 1) / 2; j++){
						ResetEvent(hEnd[j]);	// ZbgĂ
						SetEvent(hRun[j]);	// TuEXbhɌvZJn
					}
				}
			}
		}
		if (hFile){	// Ō̃\[XEt@C
			CloseHandle(hFile);
			hFile = NULL;
		}
#ifdef TIMER
time_read += GetTickCount() - time_start;
#endif

		WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, INFINITE);	// TuEXbȟvZI̍}҂
		th->size = 0;	// 2nd decode
		th->off += 1;	// vZJn\[XEubN̔ԍ
		if (th->off > 0){	// vZsvȃ\[XEubN͂Ƃ΂
			while ((s_blk[th->off].exist != 0) &&
					((s_blk[th->off].size <= block_off) || (s_blk[th->off].exist == 3))){
				prog_num += block_lost;
				th->off += 1;
#ifdef TIMER
skip_count++;
#endif
			}
		} else {	// G[ȊO th->off  0 ɂȂȂ
			memset(p_buf, 0, (size_t)unit_size * block_lost);
		}
#ifdef TIMER
		j = (th->off * 1000) / source_num;
		printf("partial decode = %d (%d.%d%%), read = %d, skip = %d\n", th->off, j / 10, j % 10, read_count, skip_count);
#endif
		recv_now = -1;	// ubN̖{̃\[Xԍ
		last_file = -1;

		// VRAM ̃TCYɉĕ
		cover_from = th->off;
		i = (source_num - cover_from + cover_max - 1) / cover_max;	// ɕď邩
		cover_num = (source_num - cover_from + i - 1) / i;	// xɏʂ𕽋ω
		while (cover_from < source_num){
			// \[XEubN邩
			if (cover_from + cover_num > source_num)
				cover_num = source_num - cover_from;
			//printf("cover_from = %d, cover_num = %d\n", cover_from, cover_num);

			// GPU  CPU XbhƂɏubNvZ
			th->buf = buf + (size_t)unit_size * cover_from;
			th->mat = mat + cover_from;
			th->off = cover_from;
			th->count = cover_num;
			for (j = 0; j < cpu_num1; j++)
				last_id[j] = 0;
			th->now = -1;	// l - 1
			_mm_sfence();	// [ւ݂̏ĂXbhĊJ
			for (j = 0; j < cpu_num1; j++){
				ResetEvent(hEnd[j]);	// ZbgĂ
				SetEvent(hRun[j]);	// TuEXbhɌvZJn
			}

			// Ō̃\[XEubNWȂAvZIubN
			if (cover_from + cover_num >= source_num){
				// TuEXbȟvZI̍} UPDATE_TIME / 2 ҂
				while (WaitForMultipleObjects(cpu_num1, hEnd, TRUE, UPDATE_TIME / 2) == WAIT_TIMEOUT){
					last_id_min = block_lost;	// ubNڂ܂ŌvZIĂ邩
					for (j = 0; j < cpu_num1; j++)	// CEXbḧ̗ɃRs[Ăr
						last_id_copy[j] = last_id[j];
					for (j = 0; j < cpu_num1; j++){
						if (last_id_min > last_id_copy[j])
							last_id_min = last_id_copy[j];
					}
					//printf("%d, %d, min = %d\n", last_id[0], last_id[1], last_id_min);

					//printf("\n write block from %d to %d\n", write_id, last_id_min - 1);
					// TuEXbĥǂꂩŏIubNɓB珑݂~߂
					while ((write_id < last_id_min) && (th->now < parity_num)){
#ifdef TIMER
time_start = GetTickCount();
#endif
						work_buf = p_buf + (size_t)unit_size * write_id;
						for (j = recv_now + 1; j < source_num; j++){	// Ԃ̃\[XEubN
							if (s_blk[j].exist == 0){
								recv_now = j;
								break;
							}
						}
						//printf(" lost block[%d] = source block[%d]\n", write_id, recv_now);

						// ꂽ\[XEubÑ`FbNT؂
						checksum16_return(work_buf, hash, io_size);
						if (memcmp(work_buf + io_size, hash, HASH_SIZE) != 0){
							printf("checksum mismatch, recovered input slice %d\n", recv_now);
							err = 1;
							goto error_end;
						}
						if (s_blk[recv_now].size <= block_off){	// ݕsv
							write_id++;
							continue;
						}
						// Ɨp̃\[XEt@Cɏ
						if (s_blk[recv_now].file != last_file){	// ʂ̃t@CȂJ
							last_file = s_blk[recv_now].file;
							if (hFile){
								CloseHandle(hFile);	// Õt@C
								hFile = NULL;
							}
							hFile = handle_temp_file(list_buf + files[last_file].name, file_path);
							if (hFile == INVALID_HANDLE_VALUE){
								hFile = NULL;
								err = 1;
								goto error_end;
							}
							//printf("file %d, open %S\n", last_file, file_path);
						}
						// \[XEt@Cł̃ubNfЂ̑傫ƈʒu
						len = s_blk[recv_now].size - block_off;
						if (len > io_size)
							len = io_size;
						if (file_write_data(hFile, (recv_now - files[last_file].b_off) * (__int64)block_size + block_off, work_buf, len)){
							printf("file_write_data, input slice %d\n", recv_now);
							err = 1;
							goto error_end;
						}
						write_id++;
#ifdef TIMER
write_count++;
time_write += GetTickCount() - time_start;
#endif
						if (GetTickCount() - time_last >= UPDATE_TIME){	// ݒoߕ\
							j = th->now - cpu_num1;
							if (j < 0)
								j = 0;
							if (print_progress((int)((((__int64)prog_num + cover_num * j) * 1000) / prog_base))){
								err = 2;
								goto error_end;
							}
							time_last = GetTickCount();
						}
					}
				}

			} else {	// rȂoߕ\ɂ
				// TuEXbȟvZI̍} UPDATE_TIME ҂
				while (WaitForMultipleObjects(cpu_num1, hEnd, TRUE, UPDATE_TIME) == WAIT_TIMEOUT){
					// th-now ōlȂ̂ŁAvZIĂ̂ th-now - cpu_num1 ƂȂ
					j = th->now - cpu_num1;
					if (j < 0)
						j = 0;
					// oߕ\iUPDATE_TIME ԑ҂ꍇȂ̂ŁAKo߂Ă͂j
					if (print_progress((int)((((__int64)prog_num + cover_num * j) * 1000) / prog_base))){
						err = 2;
						goto error_end;
					}
					time_last = GetTickCount();
				}
			}
			if (th->size != 0){	// G[
				i = th->size;
				printf("error, gpu-thread, %d, %d\n", i & 0xFF, i >> 8);
				err = 1;
				goto error_end;
			}

			// oߕ\
			prog_num += cover_num * block_lost;
			if (GetTickCount() - time_last >= UPDATE_TIME){
				if (print_progress((int)((prog_num * 1000) / prog_base))){
					err = 2;
					goto error_end;
				}
				time_last = GetTickCount();
			}

			cover_from += cover_num;
		}

#ifdef TIMER
		//printf("write block from %d to %d\n", write_id, block_lost - 1);
time_start = GetTickCount();
#endif
		// c̕ꂽubN
		work_buf = p_buf + (size_t)unit_size * write_id;
		for (i = write_id; i < block_lost; i++){
			for (j = recv_now + 1; j < source_num; j++){	// Ԃ̃\[XEubN
				if (s_blk[j].exist == 0){
					recv_now = j;
					break;
				}
			}
			//printf(" lost block[%d] = source block[%d]\n", i, recv_now);

			// ꂽ\[XEubÑ`FbNT؂
			checksum16_return(work_buf, hash, io_size);
			if (memcmp(work_buf + io_size, hash, HASH_SIZE) != 0){
				printf("checksum mismatch, recovered input slice %d\n", recv_now);
				err = 1;
				goto error_end;
			}
			if (s_blk[recv_now].size <= block_off){	// ݕsv
				work_buf += unit_size;
				continue;
			}
			// Ɨp̃\[XEt@Cɏ
			if (s_blk[recv_now].file != last_file){	// ʂ̃t@CȂJ
				last_file = s_blk[recv_now].file;
				if (hFile){
					CloseHandle(hFile);	// Õt@C
					hFile = NULL;
				}
				hFile = handle_temp_file(list_buf + files[last_file].name, file_path);
				if (hFile == INVALID_HANDLE_VALUE){
					hFile = NULL;
					err = 1;
					goto error_end;
				}
				//printf("file %d, open %S\n", last_file, file_path);
			}
			// \[XEt@Cł̃ubNfЂ̑傫ƈʒu
			len = s_blk[recv_now].size - block_off;
			if (len > io_size)
				len = io_size;
			if (file_write_data(hFile, (recv_now - files[last_file].b_off) * (__int64)block_size + block_off, work_buf, len)){
				printf("file_write_data, input slice %d\n", recv_now);
				err = 1;
				goto error_end;
			}
#ifdef TIMER
write_count++;
#endif
			work_buf += unit_size;
		}
#ifdef TIMER
time_write += GetTickCount() - time_start;
#endif

		block_off += io_size;
		// Ō̍ƃt@C
		CloseHandle(hFile);
		hFile = NULL;
	}
	print_progress_done();	// ubN̒fЉɂĂ 100% ŊƂ͌Ȃ

#ifdef TIMER
printf("read   %d.%03d sec\n", time_read / 1000, time_read % 1000);
j = ((block_size + io_size - 1) / io_size) * block_lost;
printf("write  %d.%03d sec, count = %d/%d\n", time_write / 1000, time_write % 1000, write_count, j);
#else
	info_OpenCL(buf, MEM_UNIT);	// foCX\
#endif

error_end:
	InterlockedExchange(&(th->now), INT_MAX / 2);	// TuEXbȟvZ𒆒f
	for (j = 0; j < cpu_num1; j++){
		if (hSub[j]){	// TuEXbhI
			SetEvent(hRun[j]);
			WaitForSingleObject(hSub[j], INFINITE);
			CloseHandle(hSub[j]);
		}
	}
	if (hFile)
		CloseHandle(hFile);
	if (buf)
		_aligned_free(buf);
	i = free_OpenCL();
	if (i != 0)
		printf("free_OpenCL, %d, %d", i & 0xFF, i >> 8);
	return err;
}

int decode_method5(	// ubNێꍇ (GPUΉ)
	wchar_t *file_path,
	int block_lost,			// ꂽ\[XEubN̐
	HANDLE *rcv_hFile,		// JoEt@C̃nh
	file_ctx_r *files,		// \[XEt@C̏
	source_ctx_r *s_blk,	// \[XEubN̏
	parity_ctx_r *p_blk,	// peBEubN̏
	unsigned short *mat)
{
	unsigned char *buf = NULL, *p_buf, *work_buf, *hash;
	unsigned short *id;
	int err = 0, i, j, last_file, source_off, read_num, recv_now, parity_now;
	int write_id, last_id_min, last_id_copy[(MAX_CPU + 1) * 2];
	volatile int *last_id;	// TuEXbh̓󋵂ێ
	int cpu_num1, cover_max, cover_from, cover_num;
	unsigned int unit_size, len;
	unsigned int time_last, prog_num = 0, prog_base;
	__int64 file_off;
	HANDLE hFile = NULL;
	HANDLE hSub[MAX_CPU + 1], hRun[MAX_CPU + 1], hEnd[MAX_CPU + 1];
	RS_TH th[1];

	memset(hSub, 0, sizeof(HANDLE) * (MAX_CPU + 1));
	id = mat + (block_lost * source_num);	// Ԗڂ̏\[XEubNǂ̃peBőւ邩
	last_id = last_id_copy + MAX_CPU + 1;
	unit_size = (block_size + HASH_SIZE + (MEM_UNIT - 1)) & ~(MEM_UNIT - 1);	// MEM_UNIT ̔{ɂ

	// ƃobt@[mۂiGPU ̍Ɨ̈Ƃ2̗]TĂj
	read_num = read_block_num(block_lost, 2, 1, MEM_UNIT);	// \[XEubNǂݍނ
	if (read_num == 0){
		//printf("cannot keep enough blocks, use another method\n");
		return -4;	// XCX𕪊ďȂƖ
	}
	//read_num = (read_num + 1) / 2 + 1;	// 2̎p
	//read_num = (read_num + 2) / 3 + 1;	// 3̎p
	file_off = (read_num + block_lost) * (size_t)unit_size + HASH_SIZE;
	buf = _aligned_malloc((size_t)file_off, 4096);	// GPU p̋E
	if (buf == NULL){
		printf("malloc, %I64d\n", file_off);
		err = 1;
		goto error_end;
	}
	p_buf = buf + (size_t)unit_size * read_num;	// peBEubNL^̈
	hash = p_buf + (size_t)unit_size * block_lost;
	prog_base = source_num * block_lost;	// ubN̍v|Z
#ifdef TIMER
	printf("\n read some blocks, and keep all recovering blocks (GPU)\n");
	printf("buffer size = %I64d MB, read_num = %d, round = %d\n", file_off >> 20, read_num, (source_num + read_num - 1) / read_num);
#endif

	// OpenCL ̏
	cpu_num1 = cpu_num + 1;	// GPU Ǘp̃Xbhǉ
	cover_max = read_num;	// ǂݍ߂镪ɂ
	i = init_OpenCL(unit_size, &cover_max, &len);
	if (i != 0){
		if (i != 3)	// GPU Ȃꍇ̓G[\Ȃ
			printf("init_OpenCL, %d, %d\n", i & 0xFF, i >> 8);
		i = free_OpenCL();
		if (i != 0)
			printf("free_OpenCL, %d, %d", i & 0xFF, i >> 8);
		//cpu_num1--;	// sȂꍇ͂̍sRgAEg
		// GPU g킸ɌvZ𑱍sꍇ͈ȉRgAEg
		err = -3;	// CPU ̕ɐ؂ւ
		goto error_end;
	}
	if (len == 0)	// GPULbVgȂACPUƎɃLbV̍œK݂
		len = try_cache_blocking(unit_size);
#ifdef TIMER
	printf("cache: limit size = %d, chunk_size = %d, split = %d\n", cpu_flag & 0x7FFF8000, len, (unit_size + len - 1) / len);
	printf("prog_base = %d, unit_size = %d, method = %d, cover_max = %d\n", prog_base, unit_size, OpenCL_method, cover_max);
#endif

	// }`EXbh̏
	th->mat = (unsigned short *)last_id;
	th->buf = p_buf;
	th->size = unit_size;
	th->count = block_lost;
	th->off = len;	// chunk size
	for (j = 0; j < cpu_num1; j++){	// TuEXbhƂ
		hRun[j] = CreateEvent(NULL, FALSE, FALSE, NULL);	// Auto Reset ɂ
		if (hRun[j] == NULL){
			print_win32_err();
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		hEnd[j] = CreateEvent(NULL, TRUE, FALSE, NULL);
		if (hEnd[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		// TuEXbhN
		th->run = hRun[j];
		th->end = hEnd[j];
		th->now = j;	// Xbhԍ
		_mm_sfence();	// [ւ݂̏ĂXbhN
		if (j == cpu_num){	// Ō̃Xbh GPU Ǘpɂ
			hSub[j] = (HANDLE)_beginthreadex(NULL, 65536, thread_decode_gpu, (LPVOID)th, 0, NULL);
		} else {
			hSub[j] = (HANDLE)_beginthreadex(NULL, 65536, thread_decode_each, (LPVOID)th, 0, NULL);
		}
		if (hSub[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			CloseHandle(hEnd[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		// IO ؂Ȃ悤ɁATuEXbh̗DxiGPUXbh͏Oj
		if ((j >= (cpu_num * 3) / 4) && (j < cpu_num))
			SetThreadPriority(hSub[j], THREAD_PRIORITY_BELOW_NORMAL);
		WaitForSingleObject(hEnd[j], INFINITE);	// ݒI̍}҂ (ZbgȂ)
	}

	// 񂩂ɕʂăubNǂݍŁAubN
	print_progress_text(0, "Recovering slice");
	time_last = GetTickCount();
	wcscpy(file_path, base_dir);
	write_id = 0;
	parity_now = 0;	// Ԗڂ̑փubN
	source_off = 0;	// ǂݍ݊JnXCXԍ
	while (source_off < source_num){
		if (read_num > source_num - source_off)
			read_num = source_num - source_off;
		th->size = 0xFFFFFFFF;	// 1st decode
		th->off = source_off - 1;	// ܂vZĖ

#ifdef TIMER
time_start = GetTickCount();
#endif
		last_file = -1;
		for (i = 0; i < read_num; i++){	// XCXǂݍŃ[ɔzuĂ
			switch(s_blk[source_off + i].exist){
			case 0:		// obt@[ɃpeBEubN̓eǂݍ
				if (file_read_data(rcv_hFile[p_blk[id[parity_now]].file], p_blk[id[parity_now]].off, buf + (size_t)unit_size * i, block_size)){
					printf("file_read_data, recovery slice %d\n", id[parity_now]);
					err = 1;
					goto error_end;
				}
				parity_now++;
				// peBEubÑ`FbNTvZ
				checksum16_altmap(buf + (size_t)unit_size * i, buf + ((size_t)unit_size * i + unit_size - HASH_SIZE), unit_size - HASH_SIZE);
#ifdef TIMER
read_count++;
#endif
				break;
			case 3:		// \[XEubN̓e͑S 0
				memset(buf + (size_t)unit_size * i, 0, unit_size);
				break;
			default:	// obt@[Ƀ\[XEubN̓eǂݍ
				if (s_blk[source_off + i].file != last_file){	// ʂ̃t@CȂJ
					last_file = s_blk[source_off + i].file;
					if (hFile){
						CloseHandle(hFile);	// Õt@C
						hFile = NULL;
					}
					if ((files[last_file].state & 7) != 0){	// 蒼ƃt@Cǂݍ
						get_temp_name(list_buf + files[last_file].name, file_path + base_len);
					} else if ((files[last_file].state & 32) != 0){	// Osɂ͕ʖt@Cǂݍ
						wcscpy(file_path + base_len, list_buf + files[last_file].name2);
					} else {	// Sȃ\[XEt@Cǂݍ (ǉs)
						wcscpy(file_path + base_len, list_buf + files[last_file].name);
					}
					hFile = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
					if (hFile == INVALID_HANDLE_VALUE){
						print_win32_err();
						hFile = NULL;
						printf_cp("cannot open file, %s\n", file_path);
						err = 1;
						goto error_end;
					}
				}
				len = s_blk[source_off + i].size;
				file_off = (source_off + i - files[last_file].b_off) * (__int64)block_size;
				if (file_read_data(hFile, file_off, buf + (size_t)unit_size * i, len)){
					printf("file_read_data, input slice %d\n", source_off + i);
					err = 1;
					goto error_end;
				}
				if (len < block_size)
					memset(buf + ((size_t)unit_size * i + len), 0, block_size - len);
				// \[XEubÑ`FbNTvZ
				checksum16_altmap(buf + (size_t)unit_size * i, buf + ((size_t)unit_size * i + unit_size - HASH_SIZE), unit_size - HASH_SIZE);
#ifdef TIMER
read_count++;
#endif
			}

			if (i + 1 < read_num){	// Ō̃ubNȊOȂ
				// TuEXbh̓󋵂𒲂ׂ
				j = WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, 0);
				if ((j != WAIT_TIMEOUT) && (j != WAIT_FAILED)){	// vZłȂȂ
					// oߕ\
					prog_num += block_lost;
					if (GetTickCount() - time_last >= UPDATE_TIME){
						if (print_progress(((__int64)prog_num * 1000) / prog_base)){
							err = 2;
							goto error_end;
						}
						time_last = GetTickCount();
					}
					// vZIubN̎vZJn
					th->off += 1;
					th->buf = buf + (size_t)unit_size * (th->off - source_off);
					th->mat = mat + th->off;
					th->now = -1;	// l - 1
					_mm_sfence();	// [ւ݂̏ĂXbhĊJ
					for (j = 0; j < (cpu_num + 1) / 2; j++){
						ResetEvent(hEnd[j]);	// ZbgĂ
						SetEvent(hRun[j]);	// TuEXbhɌvZJn
					}
				}
			}
		}
		if (hFile){	// Ō̃\[XEt@C
			CloseHandle(hFile);
			hFile = NULL;
		}
#ifdef TIMER
time_read += GetTickCount() - time_start;
#endif

		WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, INFINITE);	// TuEXbȟvZI̍}҂
		th->size = 0;	// 2nd decode
		th->off += 1;	// vZJn\[XEubN̔ԍ
		if (th->off == 0)	// G[ȊO th->off  0 ɂȂȂ
			memset(p_buf, 0, (size_t)unit_size * block_lost);
#ifdef TIMER
		j = (th->off - source_off) * 1000 / read_num;
		printf("partial decode = %d (%d.%d%%), read = %d\n", th->off - source_off, j / 10, j % 10, read_count);
#endif
		recv_now = -1;	// ubN̖{̃\[Xԍ
		last_file = -1;

		// VRAM ̃TCYɉĕ
		cover_from = th->off - source_off;
		i = (read_num - cover_from + cover_max - 1) / cover_max;	// ɕď邩
		cover_num = (read_num - cover_from + i - 1) / i;	// xɏʂ𕽋ω
		while (cover_from < read_num){
			// \[XEubN邩
			if (cover_from + cover_num > read_num)
				cover_num = read_num - cover_from;
			//printf("cover_from = %d, cover_num = %d\n", cover_from, cover_num);

			// GPU  CPU XbhƂɏubNvZ
			th->buf = buf + (size_t)unit_size * cover_from;
			th->mat = mat + (source_off + cover_from);
			th->off = source_off + cover_from;	// \[XEubN̔ԍɂ
			th->count = cover_num;
			for (j = 0; j < cpu_num1; j++)
				last_id[j] = 0;
			th->now = -1;	// l - 1
			_mm_sfence();	// [ւ݂̏ĂXbhĊJ
			for (j = 0; j < cpu_num1; j++){
				ResetEvent(hEnd[j]);	// ZbgĂ
				SetEvent(hRun[j]);	// TuEXbhɌvZJn
			}

			// Ō̃\[XEubNWȂAvZIubN
			if (source_off + cover_from + cover_num >= source_num){
				// TuEXbȟvZI̍} UPDATE_TIME / 2 ҂
				while (WaitForMultipleObjects(cpu_num1, hEnd, TRUE, UPDATE_TIME / 2) == WAIT_TIMEOUT){
					last_id_min = block_lost;	// ubNڂ܂ŌvZIĂ邩
					for (j = 0; j < cpu_num1; j++)	// CEXbḧ̗ɃRs[Ăr
						last_id_copy[j] = last_id[j];
					for (j = 0; j < cpu_num1; j++){
						if (last_id_min > last_id_copy[j])
							last_id_min = last_id_copy[j];
					}
					//printf("%d, %d, min = %d\n", last_id[0], last_id[1], last_id_min);

					//printf("\n write block from %d to %d\n", write_id, last_id_min - 1);
					// TuEXbĥǂꂩŏIubNɓB珑݂~߂
					while ((write_id < last_id_min) && (th->now < parity_num)){
#ifdef TIMER
time_start = GetTickCount();
#endif
						work_buf = p_buf + (size_t)unit_size * write_id;
						for (j = recv_now + 1; j < source_num; j++){	// Ԃ̃\[XEubN
							if (s_blk[j].exist == 0){
								recv_now = j;
								break;
							}
						}
						//printf(" lost block[%d] = source block[%d]\n", write_id, recv_now);

						// ꂽ\[XEubÑ`FbNT؂
						checksum16_return(work_buf, hash, unit_size - HASH_SIZE);
						if (memcmp(work_buf + unit_size - HASH_SIZE, hash, HASH_SIZE) != 0){
							printf("checksum mismatch, recovered input slice %d\n", recv_now);
							err = 1;
							goto error_end;
						}
						// Ɨp̃\[XEt@Cɏ
						if (s_blk[recv_now].file != last_file){	// ʂ̃t@CȂJ
							last_file = s_blk[recv_now].file;
							if (hFile){
								CloseHandle(hFile);	// Õt@C
								hFile = NULL;
							}
							hFile = handle_temp_file(list_buf + files[last_file].name, file_path);
							if (hFile == INVALID_HANDLE_VALUE){
								hFile = NULL;
								err = 1;
								goto error_end;
							}
							//printf("file %d, open %S\n", last_file, file_path);
						}
						if (file_write_data(hFile, (recv_now - files[last_file].b_off) * (__int64)block_size, work_buf, s_blk[recv_now].size)){
							printf("file_write_data, input slice %d\n", recv_now);
							err = 1;
							goto error_end;
						}
						write_id++;
#ifdef TIMER
time_write += GetTickCount() - time_start;
#endif
						if (GetTickCount() - time_last >= UPDATE_TIME){	// ݒoߕ\
							j = th->now - cpu_num1;
							if (j < 0)
								j = 0;
							if (print_progress((int)((((__int64)prog_num + cover_num * j) * 1000) / prog_base))){
								err = 2;
								goto error_end;
							}
							time_last = GetTickCount();
						}
					}
				}

			} else {	// rȂoߕ\ɂ
				// TuEXbȟvZI̍} UPDATE_TIME ҂
				while (WaitForMultipleObjects(cpu_num1, hEnd, TRUE, UPDATE_TIME) == WAIT_TIMEOUT){
					// th-now ōlȂ̂ŁAvZIĂ̂ th-now - cpu_num1 ƂȂ
					j = th->now - cpu_num1;
					if (j < 0)
						j = 0;
					// oߕ\iUPDATE_TIME ԑ҂ꍇȂ̂ŁAKo߂Ă͂j
					if (print_progress((int)((((__int64)prog_num + cover_num * j) * 1000) / prog_base))){
						err = 2;
						goto error_end;
					}
					time_last = GetTickCount();
				}
			}
			if (th->size != 0){	// G[
				i = th->size;
				printf("error, gpu-thread, %d, %d\n", i & 0xFF, i >> 8);
				err = 1;
				goto error_end;
			}

			// oߕ\
			prog_num += cover_num * block_lost;
			if (GetTickCount() - time_last >= UPDATE_TIME){
				if (print_progress((int)(((__int64)prog_num * 1000) / prog_base))){
					err = 2;
					goto error_end;
				}
				time_last = GetTickCount();
			}

			cover_from += cover_num;
		}

		source_off += read_num;
	}

#ifdef TIMER
	//printf("write block from %d to %d\n", write_id, block_lost - 1);
time_start = GetTickCount();
#endif
	// c̕ꂽubN
	work_buf = p_buf + (size_t)unit_size * write_id;
	for (i = write_id; i < block_lost; i++){
		for (j = recv_now + 1; j < source_num; j++){	// Ԃ̃\[XEubN
			if (s_blk[j].exist == 0){
				recv_now = j;
				break;
			}
		}
		//printf(" lost block[%d] = source block[%d]\n", i, recv_now);

		// ꂽ\[XEubÑ`FbNT؂
		checksum16_return(work_buf, hash, unit_size - HASH_SIZE);
		if (memcmp(work_buf + unit_size - HASH_SIZE, hash, HASH_SIZE) != 0){
			printf("checksum mismatch, recovered input slice %d\n", recv_now);
			err = 1;
			goto error_end;
		}
		// Ɨp̃\[XEt@Cɏ
		if (s_blk[recv_now].file != last_file){	// ʂ̃t@CȂJ
			last_file = s_blk[recv_now].file;
			if (hFile){
				CloseHandle(hFile);	// Õt@C
				hFile = NULL;
			}
			hFile = handle_temp_file(list_buf + files[last_file].name, file_path);
			if (hFile == INVALID_HANDLE_VALUE){
				hFile = NULL;
				err = 1;
				goto error_end;
			}
			//printf("file %d, open %S\n", last_file, file_path);
		}
		if (file_write_data(hFile, (recv_now - files[last_file].b_off) * (__int64)block_size, work_buf, s_blk[recv_now].size)){
			printf("file_write_data, input slice %d\n", recv_now);
			err = 1;
			goto error_end;
		}
		work_buf += unit_size;
	}
#ifdef TIMER
time_write += GetTickCount() - time_start;
#endif
	// Ō̍ƃt@C
	CloseHandle(hFile);
	hFile = NULL;
	print_progress_done();	// ubN̒fЉɂĂ 100% ŊƂ͌Ȃ

#ifdef TIMER
printf("read   %d.%03d sec\n", time_read / 1000, time_read % 1000);
printf("write  %d.%03d sec\n", time_write / 1000, time_write % 1000);
#else
	info_OpenCL(buf, MEM_UNIT);	// foCX\
#endif

error_end:
	InterlockedExchange(&(th->now), INT_MAX / 2);	// TuEXbȟvZ𒆒f
	for (j = 0; j < cpu_num1; j++){
		if (hSub[j]){	// TuEXbhI
			SetEvent(hRun[j]);
			WaitForSingleObject(hSub[j], INFINITE);
			CloseHandle(hSub[j]);
		}
	}
	if (hFile)
		CloseHandle(hFile);
	if (buf)
		_aligned_free(buf);
	i = free_OpenCL();
	if (i != 0)
		printf("free_OpenCL, %d, %d", i & 0xFF, i >> 8);
	return err;
}

