// rs_encode.c
// Copyright : 2018-12-30 Yutaka Sawada
// License : GPL

#ifndef _UNICODE
#define _UNICODE
#endif
#ifndef UNICODE
#define UNICODE
#endif
#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0501
#endif

#include <malloc.h>
#include <process.h>
#include <stdio.h>

#include <windows.h>

#include "common2.h"
#include "crc.h"
#include "gf16.h"
#include "phmd5.h"
#include "lib_opencl.h"
#include "reedsolomon.h"
#include "rs_encode.h"


#ifdef TIMER
static unsigned int time_start, time_read = 0, time_write = 0, time_calc = 0;
static unsigned int read_count = 0, write_count = 0, skip_count = 0;
#endif

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
// }`XbhCPUp̃TuEXbh

typedef struct {	// RS threading control struct
	unsigned short *mat;	// s
	unsigned char * volatile buf;
	volatile unsigned int size;		// oCg
	volatile int count;
	volatile int off;
	volatile int now;
	HANDLE run;
	HANDLE end;
} RS_TH;

static DWORD WINAPI thread_encode2(LPVOID lpParameter)
{
	unsigned char *buf, *p_buf, *work_buf;
	volatile unsigned short *factor;
	int i, j, th_id, part_num, chunk_num, max_num, left_num, left_max;
	unsigned int unit_size, len, off, chunk_size;
	HANDLE hRun, hEnd;
	RS_TH *th;
#ifdef TIMER
unsigned int loop_count2a = 0, loop_count2b = 0;
unsigned int time_start2, time_encode2a = 0, time_encode2b = 0;
#endif

	th = (RS_TH *)lpParameter;
	factor = th->mat;
	buf = th->buf;
	unit_size = th->size;
	part_num = th->count;
	chunk_size = th->off;
	th_id = th->now;	// Xbhԍ
	hRun = th->run;
	hEnd = th->end;
	_mm_sfence();
	SetEvent(hEnd);	// ݒ芮ʒm

	p_buf = buf + ((size_t)unit_size * (source_num + cpu_num - 1));
	chunk_num = (unit_size + chunk_size - 1) / chunk_size;
	max_num = chunk_num * part_num;
	if (th_id > 0)
		work_buf = buf + ((size_t)unit_size * (source_num + th_id - 1));

	WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	while (th->now < INT_MAX / 2){
#ifdef TIMER
time_start2 = GetTickCount();
#endif
		i = th->off;	// \[XEubNԍ
		len = chunk_size;
		if (th->buf == NULL){	// \[XEubNǂݍݒ
			// peBEubNƂɊ|ZĒǉĂ
			while ((j = InterlockedIncrement(&(th->now))) < max_num){	// j = ++th_now
				off = j / part_num;	// chunk ̔ԍ
				j = j % part_num;	// parity ̔ԍ
				off *= chunk_size;
				if (off + len > unit_size)
					len = unit_size - off;	// Ō chunk TCYقȂ邩
				if (i == 0)	// ŏ̃ubNvZۂ
					memset(p_buf + ((size_t)unit_size * j + off), 0, len);	// ubN 0Ŗ߂
				galois_align_multiply(buf + ((size_t)unit_size * i + off), p_buf + ((size_t)unit_size * j + off), len, factor[j]);
#ifdef TIMER
loop_count2a++;
#endif
			}
#ifdef TIMER
time_encode2a += GetTickCount() - time_start2;
#endif
		} else {	// peBEubNݒ
			if (th_id == 0){
				work_buf = th->buf;
				if (i == 0)	// ŏvZꍇ̓ubN 0Ŗ߂
					memset(work_buf, 0, unit_size);
			} else {
				memset(work_buf, 0, unit_size);	// ubN 0Ŗ߂
			}
			left_num = source_num - i;	// cubN
			left_max = chunk_num * left_num;
			// \[XEubNƂɊ|ZĒǉĂ
			while ((j = InterlockedIncrement(&(th->now))) < left_max){	// j = ++th_now
				off = j / left_num;	// chunk ̔ԍ
				j = j % left_num;	// source ̔ԍ
				off *= chunk_size;
				if (off + len > unit_size)
					len = unit_size - off;	// Ō chunk TCYقȂ邩
				galois_align_multiply(buf + ((size_t)unit_size * (i + j) + off), work_buf + off, len, factor[i + j]);
#ifdef TIMER
loop_count2b++;
#endif
			}
#ifdef TIMER
time_encode2b += GetTickCount() - time_start2;
#endif
		}
		_mm_sfence();	// [ւ݂̏
		SetEvent(hEnd);	// vZIʒm
		WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	}
#ifdef TIMER
printf("sub-thread[%d] : total loop = %d\n", th_id, loop_count2a + loop_count2b);
if (time_encode2a > 0){
	i = (int)((__int64)loop_count2a * chunk_size * 125 / ((__int64)time_encode2a * 131072));
} else {
	i = 0;
}
if (loop_count2a > 0)
	printf(" 1st encode %d.%03d sec, %d loop, %d MB/s\n", time_encode2a / 1000, time_encode2a % 1000, loop_count2a, i);
if (time_encode2b > 0){
	i = (int)((__int64)loop_count2b * chunk_size * 125 / ((__int64)time_encode2b * 131072));
} else {
	i = 0;
}
printf(" 2nd encode %d.%03d sec, %d loop, %d MB/s\n", time_encode2b / 1000, time_encode2b % 1000, loop_count2b, i);
#endif

	// I
	CloseHandle(hRun);
	CloseHandle(hEnd);
	return 0;
}

static DWORD WINAPI thread_encode3(LPVOID lpParameter)
{
	unsigned char *buf, *p_buf, *work_buf;
	volatile unsigned short *factor;
	int i, j, th_id, chunk_num, max_num, left_num, left_max;
	unsigned int unit_size, len, off, chunk_size;
	HANDLE hRun, hEnd;
	RS_TH *th;
#ifdef TIMER
unsigned int loop_count2a = 0, loop_count2b = 0;
unsigned int time_start2, time_encode2a = 0, time_encode2b = 0;
#endif

	th = (RS_TH *)lpParameter;
	factor = th->mat;
	buf = th->buf;
	unit_size = th->size;
	left_num = th->count;
	chunk_size = th->off;
	th_id = th->now;	// Xbhԍ
	hRun = th->run;
	hEnd = th->end;
	_mm_sfence();
	SetEvent(hEnd);	// ݒ芮ʒm

	p_buf = buf + (size_t)unit_size * (left_num + cpu_num - 1);
	chunk_num = (unit_size + chunk_size - 1) / chunk_size;
	max_num = chunk_num * parity_num;
	if (th_id > 0)
		work_buf = buf + (size_t)unit_size * (left_num + th_id - 1);

	WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	while (th->now < INT_MAX / 2){
#ifdef TIMER
time_start2 = GetTickCount();
#endif
		i = th->off;	// \[XEubNԍ
		len = chunk_size;
		if (th->buf == NULL){	// \[XEubNǂݍݒ
			left_num = th->count + i;
			// peBEubNƂɊ|ZĒǉĂ
			while ((j = InterlockedIncrement(&(th->now))) < max_num){	// j = ++th_now
				off = j / parity_num;	// chunk ̔ԍ
				j = j % parity_num;		// parity ̔ԍ
				off *= chunk_size;
				if (off + len > unit_size)
					len = unit_size - off;	// Ō chunk TCYقȂ邩
				if (left_num == 0)	// ŏ̃ubNvZۂɃubN 0Ŗ߂
					memset(p_buf + ((size_t)unit_size * j + off), 0, len);	// ubN 0Ŗ߂
				galois_align_multiply(buf + ((size_t)unit_size * i + off), p_buf + ((size_t)unit_size * j + off), len, factor[j]);
#ifdef TIMER
loop_count2a++;
#endif
			}
#ifdef TIMER
time_encode2a += GetTickCount() - time_start2;
#endif
		} else {	// peBEubNݒ
			if (th_id == 0){
				work_buf = th->buf;
			} else {
				memset(work_buf, 0, unit_size);	// ubN 0Ŗ߂
			}
			left_num = th->count - i;	// cubN
			left_max = chunk_num * left_num;
			// \[XEubNƂɊ|ZĒǉĂ
			while ((j = InterlockedIncrement(&(th->now))) < left_max){	// j = ++th_now
				off = j / left_num;	// chunk ̔ԍ
				j = j % left_num;	// source ̔ԍ
				off *= chunk_size;
				if (off + len > unit_size)
					len = unit_size - off;	// Ō chunk TCYقȂ邩
				galois_align_multiply(buf + ((size_t)unit_size * (i + j) + off), work_buf + off, len, factor[i + j]);
#ifdef TIMER
loop_count2b++;
#endif
			}
#ifdef TIMER
time_encode2b += GetTickCount() - time_start2;
#endif
		}
		_mm_sfence();	// [ւ݂̏
		SetEvent(hEnd);	// vZIʒm
		WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	}
#ifdef TIMER
printf("sub-thread[%d] : total loop = %d\n", th_id, loop_count2a + loop_count2b);
if (time_encode2a > 0){
	i = (int)((__int64)loop_count2a * chunk_size * 125 / ((__int64)time_encode2a * 131072));
} else {
	i = 0;
}
if (loop_count2a > 0)
	printf(" 1st encode %d.%03d sec, %d loop, %d MB/s\n", time_encode2a / 1000, time_encode2a % 1000, loop_count2a, i);
if (time_encode2b > 0){
	i = (int)((__int64)loop_count2b * chunk_size * 125 / ((__int64)time_encode2b * 131072));
} else {
	i = 0;
}
printf(" 2nd encode %d.%03d sec, %d loop, %d MB/s\n", time_encode2b / 1000, time_encode2b % 1000, loop_count2b, i);
#endif

	// I
	CloseHandle(hRun);
	CloseHandle(hEnd);
	return 0;
}

// ubNƂɌvZ邽߂̃Xbh
static DWORD WINAPI thread_encode_each(LPVOID lpParameter)
{
	unsigned char *s_buf, *p_buf, *work_buf;
	unsigned short *constant, *factor2;
	volatile unsigned short *factor1;
	int i, j, th_id, src_start, src_num, max_num;
	volatile int *last_id;
	unsigned int unit_size, len, off, chunk_size;
	HANDLE hRun, hEnd;
	RS_TH *th;
#ifdef TIMER
unsigned int loop_count2a = 0, loop_count2b = 0;
unsigned int time_start2, time_encode2a = 0, time_encode2b = 0;
#endif

	th = (RS_TH *)lpParameter;
	constant = th->mat;
	p_buf = th->buf;
	unit_size = th->size;
	th_id = th->now;	// Xbhԍ
	chunk_size = th->off;
	factor2 = (unsigned short *)(p_buf + ((size_t)unit_size * parity_num + HASH_SIZE));
	factor2 += th->count * th_id;	// XbhƂɕۑꏊς
	hRun = th->run;
	hEnd = th->end;
	_mm_sfence();
	SetEvent(hEnd);	// ݒ芮ʒm

	last_id = (int *)(constant + ((source_num + 1) & ~1));
	last_id += th_id;	// ŌɌvZubNԍL^Ă
	factor1 = constant + source_num;
	max_num = ((unit_size + chunk_size - 1) / chunk_size) * parity_num;

	WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	while (th->now < INT_MAX / 2){
#ifdef TIMER
time_start2 = GetTickCount();
#endif
		s_buf = th->buf;
		src_start = th->off;	// \[XEubNԍ

		if (th->size == 0xFFFFFFFF){	// \[XEubNǂݍݒ
			len = chunk_size;
			// peBEubNƂɊ|ZĒǉĂ
			while ((j = InterlockedIncrement(&(th->now))) < max_num){	// j = ++th_now
				off = j / parity_num;	// chunk ̔ԍ
				j = j % parity_num;		// parity ̔ԍ
				off *= chunk_size;
				if (off + len > unit_size)
					len = unit_size - off;	// Ō chunk TCYقȂ邩
				if (src_start == 0)	// ŏ̃ubNvZۂ
					memset(p_buf + ((size_t)unit_size * j + off), 0, len);	// ubN 0Ŗ߂
				galois_align_multiply(s_buf + off, p_buf + ((size_t)unit_size * j + off), len, factor1[j]);
#ifdef TIMER
loop_count2a++;
#endif
			}
#ifdef TIMER
time_encode2a += GetTickCount() - time_start2;
#endif
		} else {
			// XbhƂɍ쐬peBEubNς
			src_num = th->count;
			while ((j = InterlockedIncrement(&(th->now))) < parity_num){	// j = ++th_now
				work_buf = p_buf + (size_t)unit_size * j;

				// factor ͒萔s̏搔ɂȂ
				for (i = 0; i < src_num; i++)
					factor2[i] = galois_power(constant[src_start + i], first_num + j);

				// chunk ɕČvZ
				len = chunk_size;
				off = 0;
				while (off < unit_size){
					// \[XEubNƂɃpeBǉĂ
					for (i = 0; i < src_num; i++)
						galois_align_multiply(s_buf + ((size_t)unit_size * i + off), work_buf, len, factor2[i]);

					work_buf += len;
					off += len;
					if (off + len > unit_size)
						len = unit_size - off;
				}

				InterlockedExchange(last_id, j);	// vZIubN̔ԍ
#ifdef TIMER
loop_count2b += src_num;
#endif
			}
#ifdef TIMER
time_encode2b += GetTickCount() - time_start2;
#endif
		}
		_mm_sfence();	// [ւ݂̏
		SetEvent(hEnd);	// vZIʒm
		WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	}
#ifdef TIMER
loop_count2a /= (unit_size + chunk_size - 1) / chunk_size;	// chunkŊăubNɂ
printf("sub-thread[%d] : total loop = %d\n", th_id, loop_count2a + loop_count2b);
if (time_encode2a > 0){
	i = (int)((__int64)loop_count2a * unit_size * 125 / ((__int64)time_encode2a * 131072));
} else {
	i = 0;
}
if (loop_count2a > 0)
	printf(" 1st encode %d.%03d sec, %d loop, %d MB/s\n", time_encode2a / 1000, time_encode2a % 1000, loop_count2a, i);
if (time_encode2b > 0){
	i = (int)((__int64)loop_count2b * unit_size * 125 / ((__int64)time_encode2b * 131072));
} else {
	i = 0;
}
printf(" 2nd encode %d.%03d sec, %d loop, %d MB/s\n", time_encode2b / 1000, time_encode2b % 1000, loop_count2b, i);
#endif

	// I
	CloseHandle(hRun);
	CloseHandle(hEnd);
	return 0;
}

// GPU Ή̃TuEXbh (Xbhԍ͍ŌɂȂ)
static DWORD WINAPI thread_encode_gpu(LPVOID lpParameter)
{
	unsigned char *s_buf, *p_buf;
	unsigned short *constant, *factor2;
	int i, j, th_id, src_start, src_num;
	volatile int *last_id;
	unsigned int unit_size;
	HANDLE hRun, hEnd;
	RS_TH *th;
#ifdef TIMER
unsigned int time_start2, time_encode2 = 0, loop_count2 = 0;
#endif

	th = (RS_TH *)lpParameter;
	constant = th->mat;
	p_buf = th->buf;
	unit_size = th->size;
	th_id = th->now;	// Xbhԍ
	factor2 = (unsigned short *)(p_buf + ((size_t)unit_size * parity_num + HASH_SIZE));
	factor2 += th->count * th_id;	// XbhƂɕۑꏊς
	hRun = th->run;
	hEnd = th->end;
	_mm_sfence();
	SetEvent(hEnd);	// ݒ芮ʒm

	last_id = (int *)(constant + ((source_num + 1) & ~1));
	last_id += th_id;	// ŌɌvZubNԍL^Ă

	WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	while (th->now < INT_MAX / 2){
#ifdef TIMER
time_start2 = GetTickCount();
#endif
		// GPU̓\[XEubNǂݍݒɌĂ΂Ȃ
		s_buf = th->buf;
		src_start = th->off;	// \[XEubNԍ
		src_num = th->count;

		// ŏɃ\[XEubNVRAM֓]
		i = gpu_copy_blocks(s_buf, unit_size, src_num);
		if (i != 0){
			th->size = i;
			InterlockedExchange(&(th->now), INT_MAX / 2);	// TuEXbȟvZ𒆒f
		}

		// XbhƂɍ쐬peBEubNς
		while ((j = InterlockedIncrement(&(th->now))) < parity_num){	// j = ++th_now
			// factor ͒萔s̏搔ɂȂ
			for (i = 0; i < src_num; i++)
				factor2[i] = galois_power(constant[src_start + i], first_num + j);

			i = gpu_multiply_blocks(src_num, factor2, p_buf + (size_t)unit_size * j, unit_size);
			if (i != 0){
				th->size = i;
				break;
			}

			InterlockedExchange(last_id, j);	// vZIubN̔ԍ
#ifdef TIMER
loop_count2 += src_num;
#endif
		}
#ifdef TIMER
time_encode2 += GetTickCount() - time_start2;
#endif
		// ŌVRAM
		th->size = gpu_finish();

		_mm_sfence();	// [ւ݂̏
		SetEvent(hEnd);	// vZIʒm
		WaitForSingleObject(hRun, INFINITE);	// vZJn̍}҂
	}
#ifdef TIMER
printf("gpu-thread    : total loop = %d\n", loop_count2);
if (time_encode2 > 0){
	i = (int)((__int64)loop_count2 * unit_size * 125 / ((__int64)time_encode2 * 131072));
} else {
	i = 0;
}
printf(" 2nd encode %d.%03d sec, %d loop, %d MB/s\n", time_encode2 / 1000, time_encode2 % 1000, loop_count2, i);
#endif

	// I
	CloseHandle(hRun);
	CloseHandle(hEnd);
	return 0;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int encode_method1(	// \[XEubN̏ꍇ
	wchar_t *file_path,
	unsigned char *header_buf,	// Recovery Slice packet ̃pPbgEwb_[
	HANDLE *rcv_hFile,			// JoEt@C̃nh
	file_ctx_c *files,			// \[XEt@C̏
	source_ctx_c *s_blk,		// \[XEubN̏
	parity_ctx_c *p_blk)		// peBEubN̏
{
	unsigned char *buf = NULL, *work_buf, *hash;
	int err = 0, i, j;
	unsigned int io_size, unit_size, len, block_off;
	unsigned int time_last, prog_num = 0, prog_base;
	HANDLE hFile = NULL;
	PHMD5 md_ctx, *md_ptr = NULL;

	// ƃobt@[mۂ
	io_size = get_io_size(2, NULL, 1, sse_unit);
	//io_size = (((io_size + 2) / 3 + HASH_SIZE + (sse_unit - 1)) & ~(sse_unit - 1)) - HASH_SIZE;	// p
	unit_size = io_size + HASH_SIZE;	// `FbNT̕₷
	len = 2 * unit_size + HASH_SIZE;
	buf = _aligned_malloc(len, sse_unit);
	if (buf == NULL){
		printf("malloc, %d\n", len);
		err = 1;
		goto error_end;
	}
	work_buf = buf + unit_size;
	hash = work_buf + unit_size;
	prog_base = (block_size + io_size - 1) / io_size;
	prog_base *= parity_num;	// Ŝ̒fЂ̌
#ifdef TIMER
	printf("\n read one source block, and keep one parity block\n");
	printf("buffer size = %d MB, io_size = %d, split = %d\n", len >> 20, io_size, (block_size + io_size - 1) / io_size);
	j = try_cache_blocking(unit_size);
	printf("cache: limit size = %d, chunk_size = %d, split = %d\n", cpu_flag & 0x7FFF8000, j, (unit_size + j - 1) / j);
#endif

	if (io_size < block_size){	// XCXꍇAr܂ł̃nbVlێ
		len = sizeof(PHMD5) * parity_num;
		md_ptr = malloc(len);
		if (md_ptr == NULL){
			printf("malloc, %d\n", len);
			err = 1;
			goto error_end;
		}
		for (i = 0; i < parity_num; i++){
			Phmd5Begin(&(md_ptr[i]));
			j = first_num + i;	// ŏ̔ԍ̕
			memcpy(header_buf + 64, &j, 4);	// Recovery Slice ̔ԍ
			Phmd5Process(&(md_ptr[i]), header_buf + 32, 36);
		}
	}

	// \[XEt@CJ
	wcscpy(file_path, base_dir);
	wcscpy(file_path + base_len, list_buf + files[s_blk[0].file].name);
	hFile = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (hFile == INVALID_HANDLE_VALUE){
		print_win32_err();
		hFile = NULL;
		printf_cp("cannot open file, %s\n", list_buf + files[s_blk[0].file].name);
		err = 1;
		goto error_end;
	}

	// obt@[ETCYƂɃpeBEubN쐬
	time_last = GetTickCount();
	s_blk[0].crc = 0xFFFFFFFF;
	block_off = 0;
	while (block_off < block_size){
#ifdef TIMER
time_start = GetTickCount();
#endif
		// \[XEubNǂݍ
		len = s_blk[0].size - block_off;
		if (len > io_size)
			len = io_size;
		if (file_read_data(hFile, (__int64)block_off, buf, len)){
			printf("file_read_data, input slice %d\n", i);
			err = 1;
			goto error_end;
		}
		if (len < io_size)
			memset(buf + len, 0, io_size - len);
		// \[XEubÑ`FbNTvZ
		s_blk[0].crc = crc_update(s_blk[0].crc, buf, len);	// without pad
		checksum16_altmap(buf, buf + io_size, io_size);
#ifdef TIMER
time_read += GetTickCount() - time_start;
#endif

		// JoEt@CɏރTCY
		if (block_size - block_off < io_size){
			len = block_size - block_off;
		} else {
			len = io_size;
		}

		// peBEubNƂ
		for (i = 0; i < parity_num; i++){
#ifdef TIMER
time_start = GetTickCount();
#endif
			memset(work_buf, 0, unit_size);
			// factor  2̏搔ɂȂ
			galois_align_multiply(buf, work_buf, unit_size, galois_power(2, first_num + i));
#ifdef TIMER
time_calc += GetTickCount() - time_start;
#endif

			// oߕ\
			prog_num++;
			if (GetTickCount() - time_last >= UPDATE_TIME){
				if (print_progress((prog_num * 1000) / prog_base)){
					err = 2;
					goto error_end;
				}
				time_last = GetTickCount();
			}

#ifdef TIMER
time_start = GetTickCount();
#endif
			// peBEubÑ`FbNT؂
			checksum16_return(work_buf, hash, io_size);
			if (memcmp(work_buf + io_size, hash, HASH_SIZE) != 0){
				printf("checksum mismatch, recovery slice %d\n", i);
				err = 1;
				goto error_end;
			}
			// nbVlvZāAJoEt@Cɏ
			if (io_size >= block_size){	// 1ŏ݂IȂ
				Phmd5Begin(&md_ctx);
				j = first_num + i;	// ŏ̔ԍ̕
				memcpy(header_buf + 64, &j, 4);	// Recovery Slice ̔ԍ
				Phmd5Process(&md_ctx, header_buf + 32, 36);
				Phmd5Process(&md_ctx, work_buf, len);
				Phmd5End(&md_ctx);
				memcpy(header_buf + 16, md_ctx.hash, 16);
				// wb_[
				if (file_write_data(rcv_hFile[p_blk[i].file], p_blk[i].off + block_off - 68, header_buf, 68)){
					printf("file_write_data, recovery slice %d\n", i);
					err = 1;
					goto error_end;
				}
			} else {
				Phmd5Process(&(md_ptr[i]), work_buf, len);
			}
			if (file_write_data(rcv_hFile[p_blk[i].file], p_blk[i].off + block_off, work_buf, len)){
				printf("file_write_data, recovery slice %d\n", i);
				err = 1;
				goto error_end;
			}
#ifdef TIMER
time_write += GetTickCount() - time_start;
#endif
		}

		block_off += io_size;
	}
	print_progress_done();	// sčs̐擪ɖ߂Ă

	// t@CƂɃubN CRC-32 ؂
	if (s_blk[0].size < block_size){	// c 0 ŃpfBO
		len = block_size - s_blk[0].size;
		memset(buf, 0, len);
		s_blk[0].crc = crc_update(s_blk[0].crc, buf, len);
	}
	s_blk[0].crc ^= 0xFFFFFFFF;
	if (((unsigned int *)(files[s_blk[0].file].hash))[0] != s_blk[0].crc){
		printf("checksum mismatch, input file %d\n", s_blk[0].file);
		err = 1;
		goto error_end;
	}

	if (io_size < block_size){	// 1ŏ݂IȂȂ
		if (GetTickCount() - time_last >= UPDATE_TIME){	// LZ󂯕t
			if (cancel_progress()){
				err = 2;
				goto error_end;
			}
		}

#ifdef TIMER
time_start = GetTickCount();
#endif
		// Ō Recovery Slice packet ̃wb_[
		for (i = 0; i < parity_num; i++){
			Phmd5End(&(md_ptr[i]));
			memcpy(header_buf + 16, md_ptr[i].hash, 16);
			j = first_num + i;	// ŏ̃peBEubNԍ̕
			memcpy(header_buf + 64, &j, 4);	// Recovery Slice ̔ԍ
			// JoEt@Cɏ
			if (file_write_data(rcv_hFile[p_blk[i].file], p_blk[i].off - 68, header_buf, 68)){	// wb_[̃TCY炷
				printf("file_write_data, packet header\n");
				err = 1;
				goto error_end;
			}
		}
#ifdef TIMER
time_write += GetTickCount() - time_start;
#endif
	}

#ifdef TIMER
printf("read   %d.%03d sec\n", time_read / 1000, time_read % 1000);
printf("write  %d.%03d sec\n", time_write / 1000, time_write % 1000);
printf("encode %d.%03d sec\n", time_calc / 1000, time_calc % 1000);
#endif

error_end:
	if (md_ptr)
		free(md_ptr);
	if (hFile)
		CloseHandle(hFile);
	if (buf)
		_aligned_free(buf);
	return err;
}

int encode_method2(	// \[XEf[^Sēǂݍޏꍇ
	wchar_t *file_path,
	unsigned char *header_buf,	// Recovery Slice packet ̃pPbgEwb_[
	HANDLE *rcv_hFile,			// JoEt@C̃nh
	file_ctx_c *files,			// \[XEt@C̏
	source_ctx_c *s_blk,		// \[XEubN̏
	parity_ctx_c *p_blk,		// peBEubN̏
	unsigned short *constant)
{
	unsigned char *buf = NULL, *p_buf, *work_buf, *sub_buf, *hash;
	unsigned short *factor;
	int err = 0, i, j, last_file, part_num;
	unsigned int io_size, unit_size, len, block_off;
	unsigned int time_last;
	__int64 file_off, prog_num = 0, prog_base;
	HANDLE hFile = NULL;
	HANDLE hSub[MAX_CPU], hRun[MAX_CPU], hEnd[MAX_CPU];
	RS_TH th[1];
	PHMD5 md_ctx, *md_ptr = NULL;

	memset(hSub, 0, sizeof(HANDLE) * MAX_CPU);
	factor = constant + source_num;

	// ƃobt@[mۂ
	part_num = source_num >> PART_MAX_RATE;	// \[XEubNɑ΂銄ōőʂ߂
	if (part_num > parity_num)
		part_num = parity_num;
	i = source_num + (cpu_num - 1);	// \[XEubNƃTuEXbh̍Ɨ̈悪Kv
	io_size = get_io_size(i, &part_num, 1, sse_unit);
	//io_size = (((io_size + 2) / 3 + HASH_SIZE + (sse_unit - 1)) & ~(sse_unit - 1)) - HASH_SIZE;	// p
	unit_size = io_size + HASH_SIZE;	// `FbNT̕₷
	file_off = (i + part_num) * (size_t)unit_size + HASH_SIZE;
	buf = _aligned_malloc((size_t)file_off, sse_unit);
	if (buf == NULL){
		printf("malloc, %I64d\n", file_off);
		err = 1;
		goto error_end;
	}
	sub_buf = buf + (size_t)unit_size * source_num;	// TuEXbh̍Ɨ̈
	p_buf = sub_buf + (size_t)unit_size * (cpu_num - 1);	// peBEubN𕔕IɋL^̈
	hash = p_buf + (size_t)unit_size * part_num;
	prog_base = (block_size + io_size - 1) / io_size;
	prog_base *= source_num * parity_num;	// Ŝ̒fЂ̌
#ifdef TIMER
	printf("\n read all source blocks, and keep some parity blocks\n");
	printf("buffer size = %I64d MB, io_size = %d, split = %d\n", file_off >> 20, io_size, (block_size + io_size - 1) / io_size);
	len = try_cache_blocking(unit_size);
	printf("cache: limit size = %d, chunk_size = %d, split = %d\n", cpu_flag & 0x7FFF8000, len, (unit_size + len - 1) / len);
	printf("prog_base = %I64d, unit_size = %d, part_num = %d\n", prog_base, unit_size, part_num);
#endif

	if (io_size < block_size){	// XCXꍇAr܂ł̃nbVlێ
		len = sizeof(PHMD5) * parity_num;
		md_ptr = malloc(len);
		if (md_ptr == NULL){
			printf("malloc, %d\n", len);
			err = 1;
			goto error_end;
		}
		for (i = 0; i < parity_num; i++){
			Phmd5Begin(&(md_ptr[i]));
			j = first_num + i;	// ŏ̔ԍ̕
			memcpy(header_buf + 64, &j, 4);	// Recovery Slice ̔ԍ
			Phmd5Process(&(md_ptr[i]), header_buf + 32, 36);
		}
	}

	// }`EXbh̏
	th->mat = factor;
	th->buf = buf;
	th->size = unit_size;
	th->count = part_num;
	th->off = try_cache_blocking(unit_size);	// LbV̍œK݂
	for (j = 0; j < cpu_num; j++){	// TuEXbhƂ
		hRun[j] = CreateEvent(NULL, FALSE, FALSE, NULL);	// Auto Reset ɂ
		if (hRun[j] == NULL){
			print_win32_err();
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		hEnd[j] = CreateEvent(NULL, TRUE, FALSE, NULL);
		if (hEnd[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		// TuEXbhN
		th->run = hRun[j];
		th->end = hEnd[j];
		th->now = j;	// Xbhԍ
		_mm_sfence();	// [ւ݂̏ĂXbhN
		hSub[j] = (HANDLE)_beginthreadex(NULL, 65536, thread_encode2, (LPVOID)th, 0, NULL);
		if (hSub[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			CloseHandle(hEnd[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		if (j >= (cpu_num * 3) / 4)	// IO ؂Ȃ悤ɁATuEXbh̗Dx
			SetThreadPriority(hSub[j], THREAD_PRIORITY_BELOW_NORMAL);
		WaitForSingleObject(hEnd[j], INFINITE);	// ݒI̍}҂ (ZbgȂ)
	}

	// \[XEubNfЂǂݍŁApeBEubNfЂ쐬
	time_last = GetTickCount();
	wcscpy(file_path, base_dir);
	block_off = 0;
	while (block_off < block_size){
		th->buf = NULL;
		th->off = -1;	// ܂vZĖ

		// \[XEubNǂݍ
#ifdef TIMER
time_start = GetTickCount();
#endif
		last_file = -1;
		for (i = 0; i < source_num; i++){
			if (s_blk[i].file != last_file){	// ʂ̃t@CȂJ
				last_file = s_blk[i].file;
				if (hFile){
					CloseHandle(hFile);	// Õt@C
					hFile = NULL;
				}
				wcscpy(file_path + base_len, list_buf + files[last_file].name);
				hFile = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
				if (hFile == INVALID_HANDLE_VALUE){
					print_win32_err();
					hFile = NULL;
					printf_cp("cannot open file, %s\n", list_buf + files[last_file].name);
					err = 1;
					goto error_end;
				}
				file_off = block_off;
			} else {	// t@CȂubNETCY炷
				file_off += block_size;
			}
			if (s_blk[i].size > block_off){	// obt@[Ƀ\[XEt@C̓eǂݍ
				len = s_blk[i].size - block_off;
				if (len > io_size)
					len = io_size;
				if (file_read_data(hFile, file_off, buf + (size_t)unit_size * i, len)){
					printf("file_read_data, input slice %d\n", i);
					err = 1;
					goto error_end;
				}
				if (len < io_size)
					memset(buf + ((size_t)unit_size * i + len), 0, io_size - len);
				// \[XEubÑ`FbNTvZ
				if (block_off == 0)
					s_blk[i].crc = 0xFFFFFFFF;
				s_blk[i].crc = crc_update(s_blk[i].crc, buf + (size_t)unit_size * i, len);	// without pad
				checksum16_altmap(buf + (size_t)unit_size * i, buf + ((size_t)unit_size * i + io_size), io_size);
#ifdef TIMER
read_count++;
#endif

				if (i + 1 < source_num){	// Ō̃ubNȊOȂ
					// TuEXbh̓󋵂𒲂ׂ
					j = WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, 0);
					if ((j != WAIT_TIMEOUT) && (j != WAIT_FAILED)){	// vZłȂȂ
						// oߕ\
						prog_num += part_num;
						if (GetTickCount() - time_last >= UPDATE_TIME){
							if (print_progress((int)((prog_num * 1000) / prog_base))){
								err = 2;
								goto error_end;
							}
							time_last = GetTickCount();
						}
						// vZIubN̎vZJn
						th->off += 1;
						if (th->off > 0){	// obt@[ɓǂݍ񂾎vZ
							while (s_blk[th->off].size <= block_off){
								prog_num += part_num;
								th->off += 1;
#ifdef TIMER
skip_count++;
#endif
							}
						}
						for (j = 0; j < part_num; j++)
							factor[j] = galois_power(constant[th->off], first_num + j);	// factor ͒萔s̏搔ɂȂ
						th->now = -1;	// l - 1
						_mm_sfence();	// [ւ݂̏ĂXbhĊJ
						for (j = 0; j < (cpu_num + 1) / 2; j++){	// ǂݍݒ̓Xbh炷
							ResetEvent(hEnd[j]);	// ZbgĂ
							SetEvent(hRun[j]);	// TuEXbhɌvZJn
						}
					}
				}
			} else {
				memset(buf + (size_t)unit_size * i, 0, unit_size);
			}
		}
		// Ō̃\[XEt@C
		CloseHandle(hFile);
		hFile = NULL;
#ifdef TIMER
time_read += GetTickCount() - time_start;
#endif

		WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, INFINITE);	// TuEXbȟvZI̍}҂
		th->off += 1;	// vZJn\[XEubN̔ԍ
		if (th->off > 0){
			while (s_blk[th->off].size <= block_off){	// vZsvȃ\[XEubN͂Ƃ΂
				prog_num += part_num;
				th->off += 1;
#ifdef TIMER
skip_count++;
#endif
			}
		}
#ifdef TIMER
		j = (int)(((__int64)(th->off * part_num) * 1000) / (source_num * parity_num));
		printf("partial encode = %d (%d.%d%%), read = %d, skip = %d\n", th->off, j / 10, j % 10, read_count, skip_count);
		// ܂ł̃peBEubÑ`FbNT؂
		for (j = 0; j < part_num; j++){
			checksum16_return(p_buf + (size_t)unit_size * j, hash, io_size);
			if (memcmp(p_buf + ((size_t)unit_size * j + io_size), hash, HASH_SIZE) != 0){
				printf("checksum mismatch, recovery slice %d after 1st encode\n", j);
				err = 1;
				goto error_end;
			}
			galois_altmap_change(p_buf + (size_t)unit_size * j, unit_size);
		}
#endif

		// oߕ\
		prog_num += source_num - th->off;
		if (GetTickCount() - time_last >= UPDATE_TIME){
			if (print_progress((int)((prog_num * 1000) / prog_base))){
				err = 2;
				goto error_end;
			}
			time_last = GetTickCount();
		}

		// ŏ̃peB̌vZn߂Ă
		work_buf = p_buf;
		for (j = 0; j < source_num; j++){
			if (s_blk[j].size <= block_off){
				factor[j] = 0;	// |ZȂ
			} else {
				factor[j] = galois_power(constant[j], first_num);	// factor ͒萔s̏搔ɂȂ
			}
		}
		th->buf = work_buf;
		th->now = -1;	// l - 1
		_mm_sfence();
		for (j = 0; j < cpu_num; j++){
			ResetEvent(hEnd[j]);	// ZbgĂ
			SetEvent(hRun[j]);	// TuEXbhɌvZJn
		}

		// JoEt@CɏރTCY
		if (block_size - block_off < io_size){
			len = block_size - block_off;
		} else {
			len = io_size;
		}

		// peBEubNƂ
		for (i = 0; i < parity_num; i++){
			// peB̌vZÎ҂Ċ
			WaitForMultipleObjects(cpu_num, hEnd, TRUE, INFINITE);	// TuEXbȟvZI̍}҂
			for (j = 1; j < cpu_num; j++){
				// ŌɃTuEXbh̃obt@[ XOR 
				galois_align_xor(sub_buf + (size_t)unit_size * (j - 1), work_buf, unit_size);
			}
			if (i + 1 < parity_num){	// Ō̃ubNȊOȂ
				if (i + 1 >= part_num){	// IȃGR[h{ӏȍ~Ȃ
					if (th->buf != p_buf){	// GR[hƏ݂Ń_uEobt@O
						th->buf = p_buf;
						th->off = 0;	// ŏ̃\[XEubNvZ
					} else {
						th->buf += unit_size;
					}
				} else {
					th->buf += unit_size;
				}
				// oߕ\
				prog_num += source_num - th->off;
				if (GetTickCount() - time_last >= UPDATE_TIME){
					if (print_progress((int)((prog_num * 1000) / prog_base))){
						err = 2;
						goto error_end;
					}
					time_last = GetTickCount();
				}
				// ̃peB̌vZJnĂ
				for (j = 0; j < source_num; j++){
					if (s_blk[j].size <= block_off){
						factor[j] = 0;	// |ZȂ
					} else {
						factor[j] = galois_power(constant[j], first_num + i + 1);	// factor ͒萔s̏搔ɂȂ
					}
				}
				th->now = -1;	// l - 1
				_mm_sfence();	// [ւ݂̏ĂXbhĊJ
				for (j = 0; j < cpu_num; j++){
					ResetEvent(hEnd[j]);	// ZbgĂ
					SetEvent(hRun[j]);	// TuEXbhɌvZJn
				}
			}

#ifdef TIMER
time_start = GetTickCount();
#endif
			// peBEubÑ`FbNT؂
			checksum16_return(work_buf, hash, io_size);
			if (memcmp(work_buf + io_size, hash, HASH_SIZE) != 0){
				printf("checksum mismatch, recovery slice %d\n", i);
				err = 1;
				goto error_end;
			}
			// nbVlvZāAJoEt@Cɏ
			if (io_size >= block_size){	// 1ŏ݂IȂ
				Phmd5Begin(&md_ctx);
				j = first_num + i;	// ŏ̔ԍ̕
				memcpy(header_buf + 64, &j, 4);	// Recovery Slice ̔ԍ
				Phmd5Process(&md_ctx, header_buf + 32, 36);
				Phmd5Process(&md_ctx, work_buf, len);
				Phmd5End(&md_ctx);
				memcpy(header_buf + 16, md_ctx.hash, 16);
				// wb_[
				if (file_write_data(rcv_hFile[p_blk[i].file], p_blk[i].off + block_off - 68, header_buf, 68)){
					printf("file_write_data, recovery slice %d\n", i);
					err = 1;
					goto error_end;
				}
			} else {
				Phmd5Process(&(md_ptr[i]), work_buf, len);
			}
			//printf("%d, buf = %p, size = %u, off = %I64d\n", i, work_buf, len, p_blk[i].off + block_off);
			if (file_write_data(rcv_hFile[p_blk[i].file], p_blk[i].off + block_off, work_buf, len)){
				printf("file_write_data, recovery slice %d\n", i);
				err = 1;
				goto error_end;
			}
#ifdef TIMER
time_write += GetTickCount() - time_start;
#endif

			if (i + 1 >= part_num) {	// IȃGR[h{ӏȍ~Ȃ
				if (work_buf != p_buf){	// GR[hƏ݂Ń_uEobt@O
					work_buf = p_buf;
				} else {
					work_buf += unit_size;
				}
			} else {
				work_buf += unit_size;
			}
		}

		block_off += io_size;
	}
	print_progress_done();	// sčs̐擪ɖ߂Ă

	// t@CƂɃubN CRC-32 ؂
	memset(buf, 0, io_size);
	j = 0;
	while (j < source_num){
		last_file = s_blk[j].file;
		part_num = (int)((files[last_file].size + (__int64)block_size - 1) / block_size);
		i = j + part_num - 1;	// ubN̔ԍ
		if (s_blk[i].size < block_size){	// c 0 ŃpfBO
			len = block_size - s_blk[i].size;
			while (len > io_size){
				len -= io_size;
				s_blk[i].crc = crc_update(s_blk[i].crc, buf, io_size);
			}
			s_blk[i].crc = crc_update(s_blk[i].crc, buf, len);
		}
		memset(hash, 0, 16);
		for (i = 0; i < part_num; i++)	// XOR  16oCgɌ炷
			((unsigned int *)hash)[i & 3] ^= s_blk[j + i].crc ^ 0xFFFFFFFF;
		if (memcmp(files[last_file].hash, hash, 16) != 0){
			printf("checksum mismatch, input file %d\n", last_file);
			err = 1;
			goto error_end;
		}
		j += part_num;
	}

	if (io_size < block_size){	// 1ŏ݂IȂȂ
		if (GetTickCount() - time_last >= UPDATE_TIME){	// LZ󂯕t
			if (cancel_progress()){
				err = 2;
				goto error_end;
			}
		}

#ifdef TIMER
time_start = GetTickCount();
#endif
		// Ō Recovery Slice packet ̃wb_[
		for (i = 0; i < parity_num; i++){
			Phmd5End(&(md_ptr[i]));
			memcpy(header_buf + 16, md_ptr[i].hash, 16);
			j = first_num + i;	// ŏ̃peBEubNԍ̕
			memcpy(header_buf + 64, &j, 4);	// Recovery Slice ̔ԍ
			// JoEt@Cɏ
			if (file_write_data(rcv_hFile[p_blk[i].file], p_blk[i].off - 68, header_buf, 68)){	// wb_[̃TCY炷
				printf("file_write_data, packet header\n");
				err = 1;
				goto error_end;
			}
		}
#ifdef TIMER
time_write += GetTickCount() - time_start;
#endif
	}

#ifdef TIMER
printf("read   %d.%03d sec\n", time_read / 1000, time_read % 1000);
printf("write  %d.%03d sec\n", time_write / 1000, time_write % 1000);
#endif

error_end:
	InterlockedExchange(&(th->now), INT_MAX / 2);	// TuEXbȟvZ𒆒f
	for (j = 0; j < cpu_num; j++){
		if (hSub[j]){	// TuEXbhI
			SetEvent(hRun[j]);
			WaitForSingleObject(hSub[j], INFINITE);
			CloseHandle(hSub[j]);
		}
	}
	if (md_ptr)
		free(md_ptr);
	if (hFile)
		CloseHandle(hFile);
	if (buf)
		_aligned_free(buf);
	return err;
}

int encode_method3(	// peBEubNSĕێłꍇ
	wchar_t *file_path,
	unsigned char *header_buf,	// Recovery Slice packet ̃pPbgEwb_[
	HANDLE *rcv_hFile,			// JoEt@C̃nh
	file_ctx_c *files,			// \[XEt@C̏
	source_ctx_c *s_blk,		// \[XEubN̏
	parity_ctx_c *p_blk,		// peBEubN̏
	unsigned short *constant)
{
	unsigned char *buf = NULL, *p_buf, *work_buf, *sub_buf, *hash;
	unsigned short *factor;
	int err = 0, i, j, last_file, source_off, read_num;
	unsigned int unit_size, len;
	unsigned int time_last, prog_num = 0, prog_base;
	size_t mem_size;
	HANDLE hFile = NULL;
	HANDLE hSub[MAX_CPU], hRun[MAX_CPU], hEnd[MAX_CPU];
	RS_TH th[1];
	PHMD5 md_ctx;

	memset(hSub, 0, sizeof(HANDLE) * MAX_CPU);
	factor = constant + source_num;
	unit_size = (block_size + HASH_SIZE + (sse_unit - 1)) & ~(sse_unit - 1);	// `FbNT̕₷

	// ƃobt@[mۂ
	read_num = read_block_num(parity_num, cpu_num - 1, 1, sse_unit);	// \[XEubNǂݍނ
	if (read_num == 0){
		//printf("cannot keep enough blocks, use another method\n");
		return -2;	// XCX𕪊ďȂƖ
	}
	//read_num = (read_num + 2) / 3 + 1;	// p
	mem_size = read_num + parity_num + (cpu_num - 1);	// peBEubNƃTuEXbh̍Ɨ̈悪Kv
	mem_size = mem_size * unit_size + HASH_SIZE;
	buf = _aligned_malloc(mem_size, sse_unit);
	if (buf == NULL){
		printf("malloc, %Id\n", mem_size);
		err = 1;
		goto error_end;
	}
	sub_buf = buf + (size_t)unit_size * read_num;	// TuEXbh̍Ɨ̈
	p_buf = sub_buf + (size_t)unit_size * (cpu_num - 1);	// peBEubNL^̈
	hash = p_buf + (size_t)unit_size * parity_num;
	prog_base = source_num * parity_num;	// ubN̍v|Z
#ifdef TIMER
	printf("\n read some source blocks, and keep all parity blocks\n");
	printf("buffer size = %Id MB, read_num = %d, round = %d\n", mem_size >> 20, read_num, (source_num + read_num - 1) / read_num);
	len = try_cache_blocking(unit_size);
	printf("cache: limit size = %d, chunk_size = %d, split = %d\n", cpu_flag & 0x7FFF8000, len, (unit_size + len - 1) / len);
	printf("prog_base = %d, unit_size = %d\n", prog_base, unit_size);
#endif

	// }`EXbh̏
	th->mat = factor;
	th->buf = buf;
	th->size = unit_size;
	th->count = read_num;
	th->off = try_cache_blocking(unit_size);	// LbV̍œK݂
	for (j = 0; j < cpu_num; j++){	// TuEXbhƂ
		hRun[j] = CreateEvent(NULL, FALSE, FALSE, NULL);	// Auto Reset ɂ
		if (hRun[j] == NULL){
			print_win32_err();
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		hEnd[j] = CreateEvent(NULL, TRUE, FALSE, NULL);
		if (hEnd[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		// TuEXbhN
		th->run = hRun[j];
		th->end = hEnd[j];
		th->now = j;	// Xbhԍ
		_mm_sfence();	// [ւ݂̏ĂXbhN
		hSub[j] = (HANDLE)_beginthreadex(NULL, 65536, thread_encode3, (LPVOID)th, 0, NULL);
		if (hSub[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			CloseHandle(hEnd[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		if (j >= (cpu_num * 3) / 4)	// IO ؂Ȃ悤ɁATuEXbh̗Dx
			SetThreadPriority(hSub[j], THREAD_PRIORITY_BELOW_NORMAL);
		WaitForSingleObject(hEnd[j], INFINITE);	// ݒI̍}҂ (ZbgȂ)
	}

	// 񂩂ɕʂă\[XEubNǂݍŁApeBEubN쐬
	time_last = GetTickCount();
	wcscpy(file_path, base_dir);
	last_file = -1;
	source_off = 0;	// ǂݍ݊JnXCXԍ
	while (source_off < source_num){
		if (read_num > source_num - source_off)
			read_num = source_num - source_off;
		th->buf = NULL;
		th->count = source_off;
		th->off = -1;	// ܂vZĖ

#ifdef TIMER
time_start = GetTickCount();
#endif
		for (i = 0; i < read_num; i++){	// XCXǂݍŃ[ɔzuĂ
			// \[XEubNǂݍ
			if (s_blk[source_off + i].file != last_file){	// ʂ̃t@CȂJ
				last_file = s_blk[source_off + i].file;
				if (hFile){
					CloseHandle(hFile);	// Õt@C
					hFile = NULL;
				}
				wcscpy(file_path + base_len, list_buf + files[last_file].name);
				hFile = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
				if (hFile == INVALID_HANDLE_VALUE){
					print_win32_err();
					hFile = NULL;
					printf_cp("cannot open file, %s\n", list_buf + files[last_file].name);
					err = 1;
					goto error_end;
				}
			}
			// obt@[Ƀ\[XEt@C̓eǂݍ
			len = s_blk[source_off + i].size;
			if (!ReadFile(hFile, buf + (size_t)unit_size * i, len, &j, NULL) || (len != j)){
				print_win32_err();
				err = 1;
				goto error_end;
			}
			if (len < block_size)
				memset(buf + ((size_t)unit_size * i + len), 0, block_size - len);
			// \[XEubÑ`FbNTvZ
			s_blk[source_off + i].crc = crc_update(0xFFFFFFFF, buf + (size_t)unit_size * i, block_size) ^ 0xFFFFFFFF;	// include pad
			checksum16_altmap(buf + (size_t)unit_size * i, buf + ((size_t)unit_size * i + unit_size - HASH_SIZE), unit_size - HASH_SIZE);

			if (i + 1 < read_num){	// Ō̃ubNȊOȂ
				// TuEXbh̓󋵂𒲂ׂ
				j = WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, 0);
				if ((j != WAIT_TIMEOUT) && (j != WAIT_FAILED)){	// vZłȂȂ
					// oߕ\
					prog_num += parity_num;
					if (GetTickCount() - time_last >= UPDATE_TIME){
						if (print_progress(((__int64)prog_num * 1000) / prog_base)){
							err = 2;
							goto error_end;
						}
						time_last = GetTickCount();
					}
					// vZIubN̎vZJn
					th->off += 1;
					for (j = 0; j < parity_num; j++)
						factor[j] = galois_power(constant[source_off + th->off], first_num + j);	// factor ͒萔s̏搔ɂȂ
					th->now = -1;	// l - 1
					_mm_sfence();	// [ւ݂̏ĂXbhĊJ
					for (j = 0; j < (cpu_num + 1) / 2; j++){	// ǂݍݒ̓Xbh炷
						ResetEvent(hEnd[j]);	// ZbgĂ
						SetEvent(hRun[j]);	// TuEXbhɌvZJn
					}
				}
			}
		}
		if (source_off + i == source_num - 1){	// Ō̃\[XEt@C
			CloseHandle(hFile);
			hFile = NULL;
		}
#ifdef TIMER
time_read += GetTickCount() - time_start;
#endif

		WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, INFINITE);	// TuEXbȟvZI̍}҂
		th->off += 1;	// vZJn\[XEubN̔ԍ
		if (source_off + th->off == 0)	// G[ȊO th->off  0 ɂȂȂ
			memset(p_buf, 0, (size_t)unit_size * parity_num);
#ifdef TIMER
		j = (th->off * 1000) / read_num;
		printf("partial encode = %d (%d.%d%%)\n", th->off, j / 10, j % 10);
		// ܂ł̃peBEubÑ`FbNT؂
		for (j = 0; j < parity_num; j++){
			checksum16_return(p_buf + (size_t)unit_size * j, hash, unit_size - HASH_SIZE);
			if (memcmp(p_buf + ((size_t)unit_size * j + unit_size - HASH_SIZE), hash, HASH_SIZE) != 0){
				printf("checksum mismatch, recovery slice %d after 1st encode\n", j);
				err = 1;
				goto error_end;
			}
			galois_altmap_change(p_buf + (size_t)unit_size * j, unit_size);
		}
#endif

		// oߕ\
		prog_num += read_num - th->off;
		if (GetTickCount() - time_last >= UPDATE_TIME){
			if (print_progress(((__int64)prog_num * 1000) / prog_base)){
				err = 2;
				goto error_end;
			}
			time_last = GetTickCount();
		}

		// ŏ̃peB̌vZn߂Ă
		work_buf = p_buf;
		th->count = read_num;
		for (j = 0; j < read_num; j++)
			factor[j] = galois_power(constant[source_off + j], first_num);	// factor ͒萔s̏搔ɂȂ
		th->buf = work_buf;
		th->now = -1;	// l - 1
		_mm_sfence();
		for (j = 0; j < cpu_num; j++){
			ResetEvent(hEnd[j]);	// ZbgĂ
			SetEvent(hRun[j]);	// TuEXbhɌvZJn
		}

		// peBEubNƂ
		for (i = 0; i < parity_num; i++){
			// peB̌vZÎ҂Ċ
			WaitForMultipleObjects(cpu_num, hEnd, TRUE, INFINITE);	// TuEXbȟvZI̍}҂
			for (j = 1; j < cpu_num; j++){
				// ŌɃTuEXbh̃obt@[ XOR 
				galois_align_xor(sub_buf + (size_t)unit_size * (j - 1), work_buf, unit_size);
			}
			if (i + 1 < parity_num){	// Ō̃ubNȊOȂ
				// oߕ\
				prog_num += read_num - th->off;
				if (GetTickCount() - time_last >= UPDATE_TIME){
					if (print_progress(((__int64)prog_num * 1000) / prog_base)){
						err = 2;
						goto error_end;
					}
					time_last = GetTickCount();
				}
				// ̃peB̌vZJnĂ
				for (j = 0; j < read_num; j++)
					factor[j] = galois_power(constant[source_off + j], first_num + i + 1);	// factor ͒萔s̏搔ɂȂ
				th->buf += unit_size;
				th->now = -1;	// l - 1
				_mm_sfence();	// [ւ݂̏ĂXbhĊJ
				for (j = 0; j < cpu_num; j++){
					ResetEvent(hEnd[j]);	// ZbgĂ
					SetEvent(hRun[j]);	// TuEXbhɌvZJn
				}
			}

			if (source_off + read_num >= source_num){	// Ō݂̏Ȃ
#ifdef TIMER
time_start = GetTickCount();
#endif
				// peBEubÑ`FbNT؂
				checksum16_return(work_buf, hash, unit_size - HASH_SIZE);
				if (memcmp(work_buf + unit_size - HASH_SIZE, hash, HASH_SIZE) != 0){
					printf("checksum mismatch, recovery slice %d\n", i);
					err = 1;
					goto error_end;
				}
				// nbVlvZāAJoEt@Cɏ
				Phmd5Begin(&md_ctx);
				j = first_num + i;	// ŏ̃peBEubNԍ̕
				memcpy(header_buf + 64, &j, 4);	// Recovery Slice ̔ԍ
				Phmd5Process(&md_ctx, header_buf + 32, 36);
				Phmd5Process(&md_ctx, work_buf, block_size);
				Phmd5End(&md_ctx);
				memcpy(header_buf + 16, md_ctx.hash, 16);
				// wb_[
				if (file_write_data(rcv_hFile[p_blk[i].file], p_blk[i].off - 68, header_buf, 68)){
					printf("file_write_data, recovery slice %d\n", i);
					err = 1;
					goto error_end;
				}
				//printf("%d, buf = %p, off = %I64d\n", i, work_buf, p_blk[i].off);
				if (file_write_data(rcv_hFile[p_blk[i].file], p_blk[i].off, work_buf, block_size)){
					printf("file_write_data, recovery slice %d\n", i);
					err = 1;
					goto error_end;
				}
#ifdef TIMER
time_write += GetTickCount() - time_start;
			} else {	// peBEubÑ`FbNT؂
				checksum16_return(work_buf, hash, unit_size - HASH_SIZE);
				if (memcmp(work_buf + unit_size - HASH_SIZE, hash, HASH_SIZE) != 0){
					printf("checksum mismatch, recovery slice %d\n", i);
					err = 1;
					goto error_end;
				}
				galois_altmap_change(work_buf, unit_size);
#endif
			}

			work_buf += unit_size;
		}

		source_off += read_num;
	}
	print_progress_done();	// sčs̐擪ɖ߂Ă

	// t@CƂɃubN CRC-32 ؂
	j = 0;
	while (j < source_num){
		last_file = s_blk[j].file;
		read_num = (int)((files[last_file].size + (__int64)block_size - 1) / block_size);
		memset(hash, 0, 16);
		for (i = 0; i < read_num; i++)	// XOR  16oCgɌ炷
			((unsigned int *)hash)[i & 3] ^= s_blk[j + i].crc;
		if (memcmp(files[last_file].hash, hash, 16) != 0){
			printf("checksum mismatch, input file %d\n", last_file);
			err = 1;
			goto error_end;
		}
		j += read_num;
	}

#ifdef TIMER
printf("read   %d.%03d sec\n", time_read / 1000, time_read % 1000);
printf("write  %d.%03d sec\n", time_write / 1000, time_write % 1000);
#endif

error_end:
	InterlockedExchange(&(th->now), INT_MAX / 2);	// TuEXbȟvZ𒆒f
	for (j = 0; j < cpu_num; j++){
		if (hSub[j]){	// TuEXbhI
			SetEvent(hRun[j]);
			WaitForSingleObject(hSub[j], INFINITE);
			CloseHandle(hSub[j]);
		}
	}
	if (hFile)
		CloseHandle(hFile);
	if (buf)
		_aligned_free(buf);
	return err;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int encode_method4(	// SẴubNfГIɕێꍇ (GPUΉ)
	wchar_t *file_path,
	unsigned char *header_buf,	// Recovery Slice packet ̃pPbgEwb_[
	HANDLE *rcv_hFile,			// JoEt@C̃nh
	file_ctx_c *files,			// \[XEt@C̏
	source_ctx_c *s_blk,		// \[XEubN̏
	parity_ctx_c *p_blk,		// peBEubN̏
	unsigned short *constant)	// ubN̗̈mۂĂH
{
	unsigned char *buf = NULL, *p_buf, *work_buf, *hash;
	unsigned short *factor1;
	int err = 0, i, j, last_file;
	int write_id, last_id_min, *last_id_copy;
	volatile int *last_id;	// TuEXbh̓󋵂ێ
	int cpu_num1, cover_max, cover_from, cover_num;
	unsigned int io_size, unit_size, len, block_off;
	unsigned int time_last;
	__int64 file_off, prog_num = 0, prog_base;
	HANDLE hFile = NULL;
	HANDLE hSub[MAX_CPU + 1], hRun[MAX_CPU + 1], hEnd[MAX_CPU + 1];
	RS_TH th[1];
	PHMD5 md_ctx, *md_ptr = NULL;

	memset(hSub, 0, sizeof(HANDLE) * (MAX_CPU + 1));
	factor1 = constant + source_num;
	last_id = (int *)(constant + ((source_num + 1) & ~1));	// MAX_CPU + 1 ȏ̗vf͂͂
	last_id_copy = (int *)(constant + ((source_num + 1) & ~1)) + (cpu_num + 1);

	// ƃobt@[mۂiGPU ̍Ɨ̈Ƃ2̗]TĂj
	// part_num g킸ASẴubNێ鏊encode_method2ƈقȂ邱ƂɒӁI
	io_size = get_io_size(source_num + parity_num + 2, NULL, 1, MEM_UNIT);
	//io_size = (((io_size + 1) / 2 + HASH_SIZE + (MEM_UNIT - 1)) & ~(MEM_UNIT - 1)) - HASH_SIZE;	// 2̎p
	//io_size = (((io_size + 2) / 3 + HASH_SIZE + (MEM_UNIT - 1)) & ~(MEM_UNIT - 1)) - HASH_SIZE;	// 3̎p
	unit_size = io_size + HASH_SIZE;	// `FbNT̕₷
	file_off = (source_num + parity_num) * (size_t)unit_size + HASH_SIZE
			+ (source_num * sizeof(unsigned short) * (cpu_num + 1));
	buf = _aligned_malloc((size_t)file_off, 4096);	// GPU p̋E
	if (buf == NULL){
		printf("malloc, %I64d\n", file_off);
		err = 1;
		goto error_end;
	}
	p_buf = buf + (size_t)unit_size * source_num;	// peBEubNL^̈
	hash = p_buf + (size_t)unit_size * parity_num;
	prog_base = (block_size + io_size - 1) / io_size;
	prog_base *= source_num * parity_num;	// Ŝ̒fЂ̌
#ifdef TIMER
	printf("\n read all source blocks, and keep all parity blocks (GPU)\n");
	printf("buffer size = %I64d MB, io_size = %d, split = %d\n", file_off >> 20, io_size, (block_size + io_size - 1) / io_size);
#endif

	if (io_size < block_size){	// XCXꍇAr܂ł̃nbVlێ
		block_off = sizeof(PHMD5) * parity_num;
		md_ptr = malloc(block_off);
		if (md_ptr == NULL){
			printf("malloc, %d\n", block_off);
			err = 1;
			goto error_end;
		}
		for (i = 0; i < parity_num; i++){
			Phmd5Begin(&(md_ptr[i]));
			j = first_num + i;	// ŏ̔ԍ̕
			memcpy(header_buf + 64, &j, 4);	// Recovery Slice ̔ԍ
			Phmd5Process(&(md_ptr[i]), header_buf + 32, 36);
		}
	}

	// OpenCL ̏
	cpu_num1 = cpu_num + 1;	// GPU Ǘp̃Xbhǉ
	cover_max = source_num;
	i = init_OpenCL(unit_size, &cover_max, &len);
	if (i != 0){
		if (i != 3)	// GPU Ȃꍇ̓G[\Ȃ
			printf("init_OpenCL, %d, %d\n", i & 0xFF, i >> 8);
		i = free_OpenCL();
		if (i != 0)
			printf("free_OpenCL, %d, %d", i & 0xFF, i >> 8);
		//cpu_num1--;	// sȂꍇ͂̍sRgAEg
		// GPU g킸ɌvZ𑱍sꍇ͈ȉRgAEg
		err = -2;	// CPU ̕ɐ؂ւ
		goto error_end;
	}
	if (len == 0)	// GPULbVgȂACPUƎɃLbV̍œK݂
		len = try_cache_blocking(unit_size);
#ifdef TIMER
	printf("cache: limit size = %d, chunk_size = %d, split = %d\n", cpu_flag & 0x7FFF8000, len, (unit_size + len - 1) / len);
	printf("prog_base = %I64d, unit_size = %d, method = %d, cover_max = %d\n", prog_base, unit_size, OpenCL_method, cover_max);
#endif

	// }`EXbh̏
	th->mat = constant;
	th->buf = p_buf;
	th->size = unit_size;
	th->count = source_num;
	th->off = len;	// chunk size
	for (j = 0; j < cpu_num1; j++){	// TuEXbhƂ
		hRun[j] = CreateEvent(NULL, FALSE, FALSE, NULL);	// Auto Reset ɂ
		if (hRun[j] == NULL){
			print_win32_err();
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		hEnd[j] = CreateEvent(NULL, TRUE, FALSE, NULL);
		if (hEnd[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		// TuEXbhN
		th->run = hRun[j];
		th->end = hEnd[j];
		th->now = j;	// Xbhԍ
		_mm_sfence();	// [ւ݂̏ĂXbhN
		if (j == cpu_num){	// Ō̃Xbh GPU Ǘpɂ
			hSub[j] = (HANDLE)_beginthreadex(NULL, 65536, thread_encode_gpu, (LPVOID)th, 0, NULL);
		} else {
			hSub[j] = (HANDLE)_beginthreadex(NULL, 65536, thread_encode_each, (LPVOID)th, 0, NULL);
		}
		if (hSub[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			CloseHandle(hEnd[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		// IO ؂Ȃ悤ɁATuEXbh̗DxiGPUXbh͏Oj
		if ((j >= (cpu_num * 3) / 4) && (j < cpu_num))
			SetThreadPriority(hSub[j], THREAD_PRIORITY_BELOW_NORMAL);
		WaitForSingleObject(hEnd[j], INFINITE);	// ݒI̍}҂ (ZbgȂ)
	}

	// \[XEubNfЂǂݍŁApeBEubNfЂ쐬
	time_last = GetTickCount();
	wcscpy(file_path, base_dir);
	block_off = 0;
	while (block_off < block_size){
		th->size = 0xFFFFFFFF;	// 1st encode
		th->off = -1;	// ܂vZĖ
		write_id = 0;

		// \[XEubNǂݍ
#ifdef TIMER
time_start = GetTickCount();
#endif
		last_file = -1;
		for (i = 0; i < source_num; i++){
			if (s_blk[i].file != last_file){	// ʂ̃t@CȂJ
				last_file = s_blk[i].file;
				if (hFile){
					CloseHandle(hFile);	// Õt@C
					hFile = NULL;
				}
				wcscpy(file_path + base_len, list_buf + files[last_file].name);
				hFile = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
				if (hFile == INVALID_HANDLE_VALUE){
					print_win32_err();
					hFile = NULL;
					printf_cp("cannot open file, %s\n", list_buf + files[last_file].name);
					err = 1;
					goto error_end;
				}
				file_off = block_off;
			} else {	// t@CȂubNETCY炷
				file_off += block_size;
			}
			if (s_blk[i].size > block_off){	// obt@[Ƀ\[XEt@C̓eǂݍ
				len = s_blk[i].size - block_off;
				if (len > io_size)
					len = io_size;
				if (file_read_data(hFile, file_off, buf + (size_t)unit_size * i, len)){
					printf("file_read_data, input slice %d\n", i);
					err = 1;
					goto error_end;
				}
				if (len < io_size)
					memset(buf + ((size_t)unit_size * i + len), 0, io_size - len);
				// \[XEubÑ`FbNTvZ
				if (block_off == 0)
					s_blk[i].crc = 0xFFFFFFFF;
				s_blk[i].crc = crc_update(s_blk[i].crc, buf + (size_t)unit_size * i, len);	// without pad
				checksum16_altmap(buf + (size_t)unit_size * i, buf + ((size_t)unit_size * i + io_size), io_size);
#ifdef TIMER
read_count++;
#endif

				if (i + 1 < source_num){	// Ō̃ubNȊOȂ
					// TuEXbh̓󋵂𒲂ׂ
					j = WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, 0);
					if ((j != WAIT_TIMEOUT) && (j != WAIT_FAILED)){	// vZłȂȂ
						// oߕ\
						prog_num += parity_num;
						if (GetTickCount() - time_last >= UPDATE_TIME){
							if (print_progress((int)((prog_num * 1000) / prog_base))){
								err = 2;
								goto error_end;
							}
							time_last = GetTickCount();
						}
						// vZIubN̎vZJn
						th->off += 1;
						if (th->off > 0){	// obt@[ɓǂݍ񂾎vZ
							while (s_blk[th->off].size <= block_off){
								prog_num += parity_num;
								th->off += 1;
#ifdef TIMER
skip_count++;
#endif
							}
						}
						th->buf = buf + (size_t)unit_size * th->off;
						for (j = 0; j < parity_num; j++)
							factor1[j] = galois_power(constant[th->off], first_num + j);	// factor ͒萔s̏搔ɂȂ
						th->now = -1;	// l - 1
						_mm_sfence();
						for (j = 0; j < (cpu_num + 1) / 2; j++){
							ResetEvent(hEnd[j]);	// ZbgĂ
							SetEvent(hRun[j]);	// TuEXbhɌvZJn
						}
					}
				}
			} else {
				memset(buf + (size_t)unit_size * i, 0, unit_size);
			}
		}
		// Ō̃\[XEt@C
		CloseHandle(hFile);
		hFile = NULL;
#ifdef TIMER
time_read += GetTickCount() - time_start;
#endif

		WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, INFINITE);	// TuEXbȟvZI̍}҂
		th->size = 0;	// 2nd encode
		th->off += 1;	// vZJn\[XEubN̔ԍ
		if (th->off > 0){
			while (s_blk[th->off].size <= block_off){	// vZsvȃ\[XEubN͂Ƃ΂
				prog_num += parity_num;
				th->off += 1;
#ifdef TIMER
skip_count++;
#endif
			}
		} else {	// G[ȊO th->off  0 ɂȂȂ
			memset(p_buf, 0, (size_t)unit_size * parity_num);
		}
#ifdef TIMER
		j = (int)(((__int64)(th->off * parity_num) * 1000) / (source_num * parity_num));
		printf("partial encode = %d (%d.%d%%), read = %d, skip = %d\n", th->off, j / 10, j % 10, read_count, skip_count);
#endif

		// JoEt@CɏރTCY
		if (block_size - block_off < io_size){
			len = block_size - block_off;
		} else {
			len = io_size;
		}

		// VRAM ̃TCYɉĕ
		cover_from = th->off;
		i = (source_num - cover_from + cover_max - 1) / cover_max;	// ɕď邩
		cover_num = (source_num - cover_from + i - 1) / i;	// xɏʂ𕽋ω
		//printf("cover range = %d, cover_num = %d\n", source_num - cover_from, cover_num);
		while (cover_from < source_num){
			// \[XEubN邩
			if (cover_from + cover_num > source_num)
				cover_num = source_num - cover_from;
			//printf("cover_from = %d, cover_num = %d\n", cover_from, cover_num);

			// GPU  CPU XbhƂɃpeBEubNvZ
			th->buf = buf + (size_t)unit_size * cover_from;
			th->off = cover_from;	// \[XEubN̔ԍɂ
			th->count = cover_num;
			for (j = 0; j < cpu_num1; j++)
				last_id[j] = 0;
			th->now = -1;	// l - 1
			_mm_sfence();
			for (j = 0; j < cpu_num1; j++){
				ResetEvent(hEnd[j]);	// ZbgĂ
				SetEvent(hRun[j]);	// TuEXbhɌvZJn
			}

			// Ō̃\[XEubNWȂAvZIubN
			if (cover_from + cover_num >= source_num){
				// TuEXbȟvZI̍} UPDATE_TIME / 2 ҂
				while (WaitForMultipleObjects(cpu_num, hEnd, TRUE, UPDATE_TIME / 2) == WAIT_TIMEOUT){
					last_id_min = parity_num;	// ubNڂ܂ŌvZIĂ邩
					for (j = 0; j < cpu_num; j++)	// CEXbḧ̗ɃRs[Ăr
						last_id_copy[j] = last_id[j];
					for (j = 0; j < cpu_num1; j++){
						if (last_id_min > last_id_copy[j])
							last_id_min = last_id_copy[j];
					}
					//printf("%d, %d, min = %d\n", last_id[0], last_id[1], last_id_min);

					// TuEXbĥǂꂩŏIubNɓB珑݂~߂
					while ((write_id < last_id_min) && (th->now < parity_num)){
#ifdef TIMER
time_start = GetTickCount();
#endif
						work_buf = p_buf + (size_t)unit_size * write_id;
						// peBEubÑ`FbNT؂
						checksum16_return(work_buf, hash, io_size);
						if (memcmp(work_buf + io_size, hash, HASH_SIZE) != 0){
							printf("checksum mismatch, recovery slice %d\n", write_id);
							err = 1;
							goto error_end;
						}
						// nbVlvZāAJoEt@Cɏ
						if (io_size >= block_size){	// 1ŏ݂IȂ
							Phmd5Begin(&md_ctx);
							j = first_num + write_id;	// ŏ̔ԍ̕
							memcpy(header_buf + 64, &j, 4);	// Recovery Slice ̔ԍ
							Phmd5Process(&md_ctx, header_buf + 32, 36);
							Phmd5Process(&md_ctx, work_buf, len);
							Phmd5End(&md_ctx);
							memcpy(header_buf + 16, md_ctx.hash, 16);
							// wb_[
							if (file_write_data(rcv_hFile[p_blk[write_id].file], p_blk[write_id].off + block_off - 68, header_buf, 68)){
								printf("file_write_data, recovery slice %d\n", write_id);
								err = 1;
								goto error_end;
							}
						} else {
							Phmd5Process(&(md_ptr[write_id]), work_buf, len);
						}
						//printf("%d, buf = %p, size = %u, off = %I64d\n", write_id, work_buf, len, p_blk[write_id].off + block_off);
						if (file_write_data(rcv_hFile[p_blk[write_id].file], p_blk[write_id].off + block_off, work_buf, len)){
							printf("file_write_data, recovery slice %d\n", write_id);
							err = 1;
							goto error_end;
						}
						write_id++;
#ifdef TIMER
time_write += GetTickCount() - time_start;
#endif
						if (GetTickCount() - time_last >= UPDATE_TIME){	// ݒoߕ\
							j = th->now - cpu_num1;
							if (j < 0)
								j = 0;
							if (print_progress((int)((((__int64)prog_num + cover_num * j) * 1000) / prog_base))){
								err = 2;
								goto error_end;
							}
							time_last = GetTickCount();
						}
					}
				}

			} else {	// rȂoߕ\ɂ
				// TuEXbȟvZI̍} UPDATE_TIME ҂
				while (WaitForMultipleObjects(cpu_num1, hEnd, TRUE, UPDATE_TIME) == WAIT_TIMEOUT){
					// th-now ōlȂ̂ŁAvZIĂ̂ th-now - cpu_num1 ƂȂ
					j = th->now - cpu_num1;
					if (j < 0)
						j = 0;
					// oߕ\iUPDATE_TIME ԑ҂ꍇȂ̂ŁAKo߂Ă͂j
					if (print_progress((int)((((__int64)prog_num + cover_num * j) * 1000) / prog_base))){
						err = 2;
						goto error_end;
					}
					time_last = GetTickCount();
				}
			}
			if (th->size != 0){	// G[
				i = th->size;
				printf("error, gpu-thread, %d, %d\n", i & 0xFF, i >> 8);
				err = 1;
				goto error_end;
			}

			// oߕ\
			prog_num += cover_num * parity_num;
			if (GetTickCount() - time_last >= UPDATE_TIME){
				if (print_progress((int)(((__int64)prog_num * 1000) / prog_base))){
					err = 2;
					goto error_end;
				}
				time_last = GetTickCount();
			}

			cover_from += cover_num;
		}

#ifdef TIMER
		//printf("write block from %d to %d\n", write_id, parity_num - 1);
time_start = GetTickCount();
#endif
		// c̃peBEubN
		work_buf = p_buf + (size_t)unit_size * write_id;
		for (i = write_id; i < parity_num; i++){
			// peBEubÑ`FbNT؂
			checksum16_return(work_buf, hash, io_size);
			if (memcmp(work_buf + io_size, hash, HASH_SIZE) != 0){
				printf("checksum mismatch, recovery slice %d\n", i);
				err = 1;
				goto error_end;
			}
			// nbVlvZāAJoEt@Cɏ
			if (io_size >= block_size){	// 1ŏ݂IȂ
				Phmd5Begin(&md_ctx);
				j = first_num + i;	// ŏ̔ԍ̕
				memcpy(header_buf + 64, &j, 4);	// Recovery Slice ̔ԍ
				Phmd5Process(&md_ctx, header_buf + 32, 36);
				Phmd5Process(&md_ctx, work_buf, len);
				Phmd5End(&md_ctx);
				memcpy(header_buf + 16, md_ctx.hash, 16);
				// wb_[
				if (file_write_data(rcv_hFile[p_blk[i].file], p_blk[i].off + block_off - 68, header_buf, 68)){
					printf("file_write_data, recovery slice %d\n", i);
					err = 1;
					goto error_end;
				}
			} else {
				Phmd5Process(&(md_ptr[i]), work_buf, len);
			}
			//printf("%d, buf = %p, size = %u, off = %I64d\n", i, work_buf, len, p_blk[i].off + block_off);
			if (file_write_data(rcv_hFile[p_blk[i].file], p_blk[i].off + block_off, work_buf, len)){
				printf("file_write_data, recovery slice %d\n", i);
				err = 1;
				goto error_end;
			}
			work_buf += unit_size;
		}
#ifdef TIMER
time_write += GetTickCount() - time_start;
#endif

		block_off += io_size;
	}
	print_progress_done();	// sčs̐擪ɖ߂Ă

	// t@CƂɃubN CRC-32 ؂
	memset(buf, 0, io_size);
	j = 0;
	while (j < source_num){
		last_file = s_blk[j].file;
		cover_num = (int)((files[last_file].size + (__int64)block_size - 1) / block_size);
		i = j + cover_num - 1;	// ubN̔ԍ
		if (s_blk[i].size < block_size){	// c 0 ŃpfBO
			len = block_size - s_blk[i].size;
			while (len > io_size){
				len -= io_size;
				s_blk[i].crc = crc_update(s_blk[i].crc, buf, io_size);
			}
			s_blk[i].crc = crc_update(s_blk[i].crc, buf, len);
		}
		memset(hash, 0, 16);
		for (i = 0; i < cover_num; i++)	// XOR  16oCgɌ炷
			((unsigned int *)hash)[i & 3] ^= s_blk[j + i].crc ^ 0xFFFFFFFF;
		if (memcmp(files[last_file].hash, hash, 16) != 0){
			printf("checksum mismatch, input file %d\n", last_file);
			err = 1;
			goto error_end;
		}
		j += cover_num;
	}

	//printf("io_size = %d, block_size = %d\n", io_size, block_size);
	if (io_size < block_size){	// 1ŏ݂IȂȂ
		if (GetTickCount() - time_last >= UPDATE_TIME){	// LZ󂯕t
			if (cancel_progress()){
				err = 2;
				goto error_end;
			}
		}

#ifdef TIMER
time_start = GetTickCount();
#endif
		// Ō Recovery Slice packet ̃wb_[
		for (i = 0; i < parity_num; i++){
			Phmd5End(&(md_ptr[i]));
			memcpy(header_buf + 16, md_ptr[i].hash, 16);
			j = first_num + i;	// ŏ̃peBEubNԍ̕
			memcpy(header_buf + 64, &j, 4);	// Recovery Slice ̔ԍ
			// JoEt@Cɏ
			if (file_write_data(rcv_hFile[p_blk[i].file], p_blk[i].off - 68, header_buf, 68)){	// wb_[̃TCY炷
				printf("file_write_data, packet header\n");
				err = 1;
				goto error_end;
			}
		}
#ifdef TIMER
time_write += GetTickCount() - time_start;
#endif
	}

#ifdef TIMER
printf("read   %d.%03d sec\n", time_read / 1000, time_read % 1000);
printf("write  %d.%03d sec\n", time_write / 1000, time_write % 1000);
#else
	info_OpenCL(buf, MEM_UNIT);	// foCX\
#endif

error_end:
	InterlockedExchange(&(th->now), INT_MAX / 2);	// TuEXbȟvZ𒆒f
	for (j = 0; j < cpu_num1; j++){
		if (hSub[j]){	// TuEXbhI
			SetEvent(hRun[j]);
			WaitForSingleObject(hSub[j], INFINITE);
			CloseHandle(hSub[j]);
		}
	}
	if (md_ptr)
		free(md_ptr);
	if (hFile)
		CloseHandle(hFile);
	if (buf)
		_aligned_free(buf);
	i = free_OpenCL();
	if (i != 0)
		printf("free_OpenCL, %d, %d", i & 0xFF, i >> 8);
	return err;
}

int encode_method5(	// \[XEubN̈ꕔƃpeBEubNێꍇ (GPUΉ)
	wchar_t *file_path,
	unsigned char *header_buf,	// Recovery Slice packet ̃pPbgEwb_[
	HANDLE *rcv_hFile,			// JoEt@C̃nh
	file_ctx_c *files,			// \[XEt@C̏
	source_ctx_c *s_blk,		// \[XEubN̏
	parity_ctx_c *p_blk,		// peBEubN̏
	unsigned short *constant)
{
	unsigned char *buf = NULL, *p_buf, *work_buf, *hash;
	unsigned short *factor1;
	int err = 0, i, j, last_file, source_off, read_num;
	int write_id, last_id_min, *last_id_copy;
	volatile int *last_id;	// TuEXbh̓󋵂ێ
	int cpu_num1, cover_max, cover_from, cover_num;
	unsigned int unit_size, len;
	unsigned int time_last, prog_num = 0, prog_base;
	size_t mem_size;
	HANDLE hFile = NULL;
	HANDLE hSub[MAX_CPU + 1], hRun[MAX_CPU + 1], hEnd[MAX_CPU + 1];
	RS_TH th[1];
	PHMD5 md_ctx;

	memset(hSub, 0, sizeof(HANDLE) * (MAX_CPU + 1));
	factor1 = constant + source_num;
	last_id = (int *)(constant + ((source_num + 1) & ~1));	// MAX_CPU + 1 ȏ̗vf͂͂
	last_id_copy = (int *)(constant + ((source_num + 1) & ~1)) + (cpu_num + 1);
	unit_size = (block_size + HASH_SIZE + (MEM_UNIT - 1)) & ~(MEM_UNIT - 1);	// MEM_UNIT ̔{ɂ

	// ƃobt@[mۂiGPU ̍Ɨ̈Ƃ2̗]TĂj
	read_num = read_block_num(parity_num, 2, 1, MEM_UNIT);	// \[XEubNǂݍނ
	if (read_num == 0){
		//printf("cannot keep enough blocks, use another method\n");
		return -4;	// XCX𕪊ďȂƖ
	}
	//read_num = (read_num + 1) / 2 + 1;	// 2̎p
	//read_num = (read_num + 2) / 3 + 1;	// 3̎p
	mem_size = (size_t)(read_num + parity_num) * unit_size + HASH_SIZE
			+ (read_num * sizeof(unsigned short) * (cpu_num + 1));
	buf = _aligned_malloc(mem_size, 4096);	// GPU p̋E
	if (buf == NULL){
		printf("malloc, %Id\n", mem_size);
		err = 1;
		goto error_end;
	}
	p_buf = buf + (size_t)unit_size * read_num;	// peBEubNL^̈
	hash = p_buf + (size_t)unit_size * parity_num;
	prog_base = source_num * parity_num;	// ubN̍v|Z
#ifdef TIMER
	printf("\n read some source blocks, and keep all parity blocks (GPU)\n");
	printf("buffer size = %Id MB, read_num = %d, round = %d\n", mem_size >> 20, read_num, (source_num + read_num - 1) / read_num);
#endif

	// OpenCL ̏
	cpu_num1 = cpu_num + 1;	// GPU Ǘp̃Xbhǉ
	cover_max = read_num;	// ǂݍ߂镪ɂ
	i = init_OpenCL(unit_size, &cover_max, &len);
	if (i != 0){
		if (i != 3)	// GPU Ȃꍇ̓G[\Ȃ
			printf("init_OpenCL, %d, %d\n", i & 0xFF, i >> 8);
		i = free_OpenCL();
		if (i != 0)
			printf("free_OpenCL, %d, %d", i & 0xFF, i >> 8);
		//cpu_num1--;	// sȂꍇ͂̍sRgAEg
		// GPU g킸ɌvZ𑱍sꍇ͈ȉRgAEg
		err = -3;	// CPU ̕ɐ؂ւ
		goto error_end;
	}
	if (len == 0)	// GPULbVgȂACPUƎɃLbV̍œK݂
		len = try_cache_blocking(unit_size);
#ifdef TIMER
	printf("cache: limit size = %d, chunk_size = %d, split = %d\n", cpu_flag & 0x7FFF8000, len, (unit_size + len - 1) / len);
	printf("prog_base = %d, unit_size = %d, method = %d, cover_max = %d\n", prog_base, unit_size, OpenCL_method, cover_max);
#endif

	// }`EXbh̏
	th->mat = constant;
	th->buf = p_buf;
	th->size = unit_size;
	th->count = read_num;
	th->off = len;	// chunk size
	for (j = 0; j < cpu_num1; j++){	// TuEXbhƂ
		hRun[j] = CreateEvent(NULL, FALSE, FALSE, NULL);	// Auto Reset ɂ
		if (hRun[j] == NULL){
			print_win32_err();
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		hEnd[j] = CreateEvent(NULL, TRUE, FALSE, NULL);
		if (hEnd[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		// TuEXbhN
		th->run = hRun[j];
		th->end = hEnd[j];
		th->now = j;	// Xbhԍ
		_mm_sfence();	// [ւ݂̏ĂXbhN
		if (j == cpu_num){	// Ō̃Xbh GPU Ǘpɂ
			hSub[j] = (HANDLE)_beginthreadex(NULL, 65536, thread_encode_gpu, (LPVOID)th, 0, NULL);
		} else {
			hSub[j] = (HANDLE)_beginthreadex(NULL, 65536, thread_encode_each, (LPVOID)th, 0, NULL);
		}
		if (hSub[j] == NULL){
			print_win32_err();
			CloseHandle(hRun[j]);
			CloseHandle(hEnd[j]);
			printf("error, sub-thread\n");
			err = 1;
			goto error_end;
		}
		// IO ؂Ȃ悤ɁATuEXbh̗DxiGPUXbh͏Oj
		if ((j >= (cpu_num * 3) / 4) && (j < cpu_num))
			SetThreadPriority(hSub[j], THREAD_PRIORITY_BELOW_NORMAL);
		WaitForSingleObject(hEnd[j], INFINITE);	// ݒI̍}҂ (ZbgȂ)
	}

	// 񂩂ɕʂă\[XEubNǂݍŁApeBEubN쐬
	time_last = GetTickCount();
	wcscpy(file_path, base_dir);
	last_file = -1;
	write_id = 0;
	source_off = 0;	// ǂݍ݊JnXCXԍ
	while (source_off < source_num){
		if (read_num > source_num - source_off)
			read_num = source_num - source_off;
		th->size = 0xFFFFFFFF;	// 1st encode
		th->off = source_off - 1;	// ܂vZĖ

#ifdef TIMER
time_start = GetTickCount();
#endif
		for (i = 0; i < read_num; i++){	// XCXǂݍŃ[ɔzuĂ
			// \[XEubNǂݍ
			if (s_blk[source_off + i].file != last_file){	// ʂ̃t@CȂJ
				last_file = s_blk[source_off + i].file;
				if (hFile){
					CloseHandle(hFile);	// Õt@C
					hFile = NULL;
				}
				wcscpy(file_path + base_len, list_buf + files[last_file].name);
				hFile = CreateFile(file_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
				if (hFile == INVALID_HANDLE_VALUE){
					print_win32_err();
					hFile = NULL;
					printf_cp("cannot open file, %s\n", list_buf + files[last_file].name);
					err = 1;
					goto error_end;
				}
			}
			// obt@[Ƀ\[XEt@C̓eǂݍ
			len = s_blk[source_off + i].size;
			if (!ReadFile(hFile, buf + (size_t)unit_size * i, len, &j, NULL) || (len != j)){
				print_win32_err();
				err = 1;
				goto error_end;
			}
			if (len < block_size)
				memset(buf + ((size_t)unit_size * i + len), 0, block_size - len);
			// \[XEubÑ`FbNTvZ
			s_blk[source_off + i].crc = crc_update(0xFFFFFFFF, buf + (size_t)unit_size * i, block_size) ^ 0xFFFFFFFF;	// include pad
			checksum16_altmap(buf + (size_t)unit_size * i, buf + ((size_t)unit_size * i + unit_size - HASH_SIZE), unit_size - HASH_SIZE);

			if (i + 1 < read_num){	// Ō̃ubNȊOȂ
				// TuEXbh̓󋵂𒲂ׂ
				j = WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, 0);
				if ((j != WAIT_TIMEOUT) && (j != WAIT_FAILED)){	// vZłȂȂ
					// oߕ\
					prog_num += parity_num;
					if (GetTickCount() - time_last >= UPDATE_TIME){
						if (print_progress(((__int64)prog_num * 1000) / prog_base)){
							err = 2;
							goto error_end;
						}
						time_last = GetTickCount();
					}
					// vZIubN̎vZJn
					th->off += 1;
					th->buf = buf + (size_t)unit_size * (th->off - source_off);
					for (j = 0; j < parity_num; j++)
						factor1[j] = galois_power(constant[th->off], first_num + j);	// factor ͒萔s̏搔ɂȂ
					th->now = -1;	// l - 1
					_mm_sfence();
					for (j = 0; j < (cpu_num + 1) / 2; j++){
						ResetEvent(hEnd[j]);	// ZbgĂ
						SetEvent(hRun[j]);	// TuEXbhɌvZJn
					}
				}
			}
		}
		if (source_off + i == source_num - 1){	// Ō̃\[XEt@C
			CloseHandle(hFile);
			hFile = NULL;
		}
#ifdef TIMER
time_read += GetTickCount() - time_start;
#endif

		WaitForMultipleObjects((cpu_num + 1) / 2, hEnd, TRUE, INFINITE);	// TuEXbȟvZI̍}҂
		th->size = 0;	// 2nd encode
		th->off += 1;	// vZJn\[XEubN̔ԍ
		if (th->off == 0)	// G[ȊO th->off  0 ɂȂȂ
			memset(p_buf, 0, (size_t)unit_size * parity_num);
#ifdef TIMER
		j = (th->off - source_off) * 1000 / read_num;
		printf("partial encode = %d (%d.%d%%)\n", th->off - source_off, j / 10, j % 10);
#endif

		// VRAM ̃TCYɉĕ
		cover_from = th->off - source_off;
		i = (read_num - cover_from + cover_max - 1) / cover_max;	// ɕď邩
		cover_num = (read_num - cover_from + i - 1) / i;	// xɏʂ𕽋ω
		//printf("cover range = %d, cover_num = %d\n", read_num - cover_from, cover_num);
		while (cover_from < read_num){
			// \[XEubN邩
			if (cover_from + cover_num > read_num)
				cover_num = read_num - cover_from;
			//printf("cover_from = %d, cover_num = %d\n", cover_from, cover_num);

			// GPU  CPU XbhƂɃpeBEubNvZ
			th->buf = buf + (size_t)unit_size * cover_from;
			th->off = source_off + cover_from;	// \[XEubN̔ԍɂ
			th->count = cover_num;
			for (j = 0; j < cpu_num1; j++)
				last_id[j] = 0;
			th->now = -1;	// l - 1
			_mm_sfence();
			for (j = 0; j < cpu_num1; j++){
				ResetEvent(hEnd[j]);	// ZbgĂ
				SetEvent(hRun[j]);	// TuEXbhɌvZJn
			}

			// Ō̃\[XEubNWȂAvZIubN
			if (source_off + cover_from + cover_num >= source_num){
				// TuEXbȟvZI̍} UPDATE_TIME / 2 ҂
				while (WaitForMultipleObjects(cpu_num1, hEnd, TRUE, UPDATE_TIME / 2) == WAIT_TIMEOUT){
					last_id_min = parity_num;	// ubNڂ܂ŌvZIĂ邩
					for (j = 0; j < cpu_num1; j++)	// CEXbḧ̗ɃRs[Ăr
						last_id_copy[j] = last_id[j];
					for (j = 0; j < cpu_num1; j++){
						if (last_id_min > last_id_copy[j])
							last_id_min = last_id_copy[j];
					}
					//printf("%d, %d, min = %d\n", last_id[0], last_id[1], last_id_min);

					//printf("\n write block from %d to %d\n", write_id, last_id_min - 1);
					// TuEXbĥǂꂩŏIubNɓB珑݂~߂
					while ((write_id < last_id_min) && (th->now < parity_num)){
#ifdef TIMER
time_start = GetTickCount();
#endif
						work_buf = p_buf + (size_t)unit_size * write_id;
						// peBEubÑ`FbNT؂
						checksum16_return(work_buf, hash, unit_size - HASH_SIZE);
						if (memcmp(work_buf + unit_size - HASH_SIZE, hash, HASH_SIZE) != 0){
							printf("checksum mismatch, recovery slice %d\n", write_id);
							err = 1;
							goto error_end;
						}
						// nbVlvZāAJoEt@Cɏ
						Phmd5Begin(&md_ctx);
						j = first_num + write_id;	// ŏ̔ԍ̕
						memcpy(header_buf + 64, &j, 4);	// Recovery Slice ̔ԍ
						Phmd5Process(&md_ctx, header_buf + 32, 36);
						Phmd5Process(&md_ctx, work_buf, block_size);
						Phmd5End(&md_ctx);
						memcpy(header_buf + 16, md_ctx.hash, 16);
						// wb_[
						if (file_write_data(rcv_hFile[p_blk[write_id].file], p_blk[write_id].off - 68, header_buf, 68)){
							printf("file_write_data, recovery slice %d\n", write_id);
							err = 1;
							goto error_end;
						}
						//printf("%d, buf = %p, off = %I64d\n", write_id, work_buf, p_blk[write_id].off);
						if (file_write_data(rcv_hFile[p_blk[write_id].file], p_blk[write_id].off, work_buf, block_size)){
							printf("file_write_data, recovery slice %d\n", write_id);
							err = 1;
							goto error_end;
						}
						write_id++;
#ifdef TIMER
time_write += GetTickCount() - time_start;
#endif
						if (GetTickCount() - time_last >= UPDATE_TIME){	// ݒoߕ\
							j = th->now - cpu_num1;
							if (j < 0)
								j = 0;
							if (print_progress((int)((((__int64)prog_num + cover_num * j) * 1000) / prog_base))){
								err = 2;
								goto error_end;
							}
							time_last = GetTickCount();
						}
					}
				}

			} else {	// rȂoߕ\ɂ
				// TuEXbȟvZI̍} UPDATE_TIME ҂
				while (WaitForMultipleObjects(cpu_num1, hEnd, TRUE, UPDATE_TIME) == WAIT_TIMEOUT){
					// th-now ōlȂ̂ŁAvZIĂ̂ th-now - cpu_num1 ƂȂ
					j = th->now - cpu_num1;
					if (j < 0)
						j = 0;
					// oߕ\iUPDATE_TIME ԑ҂ꍇȂ̂ŁAKo߂Ă͂j
					if (print_progress((int)((((__int64)prog_num + cover_num * j) * 1000) / prog_base))){
						err = 2;
						goto error_end;
					}
					time_last = GetTickCount();
				}
			}
			if (th->size != 0){	// G[
				i = th->size;
				printf("error, gpu-thread, %d, %d\n", i & 0xFF, i >> 8);
				err = 1;
				goto error_end;
			}

			// oߕ\
			prog_num += cover_num * parity_num;
			if (GetTickCount() - time_last >= UPDATE_TIME){
				if (print_progress((int)(((__int64)prog_num * 1000) / prog_base))){
					err = 2;
					goto error_end;
				}
				time_last = GetTickCount();
			}

			cover_from += cover_num;
		}

		source_off += read_num;
	}

#ifdef TIMER
	//printf("write block from %d to %d\n", write_id, parity_num - 1);
time_start = GetTickCount();
#endif
	// c̃peBEubN
	work_buf = p_buf + (size_t)unit_size * write_id;
	for (i = write_id; i < parity_num; i++){
		// peBEubÑ`FbNT؂
		checksum16_return(work_buf, hash, unit_size - HASH_SIZE);
		if (memcmp(work_buf + unit_size - HASH_SIZE, hash, HASH_SIZE) != 0){
			printf("checksum mismatch, recovery slice %d\n", i);
			err = 1;
			goto error_end;
		}
		// nbVlvZāAJoEt@Cɏ
		Phmd5Begin(&md_ctx);
		j = first_num + i;	// ŏ̔ԍ̕
		memcpy(header_buf + 64, &j, 4);	// Recovery Slice ̔ԍ
		Phmd5Process(&md_ctx, header_buf + 32, 36);
		Phmd5Process(&md_ctx, work_buf, block_size);
		Phmd5End(&md_ctx);
		memcpy(header_buf + 16, md_ctx.hash, 16);
		// wb_[
		if (file_write_data(rcv_hFile[p_blk[i].file], p_blk[i].off - 68, header_buf, 68)){
			printf("file_write_data, recovery slice %d\n", i);
			err = 1;
			goto error_end;
		}
		//printf("%d, buf = %p, off = %I64d\n", i, work_buf, p_blk[i].off);
		if (file_write_data(rcv_hFile[p_blk[i].file], p_blk[i].off, work_buf, block_size)){
			printf("file_write_data, recovery slice %d\n", i);
			err = 1;
			goto error_end;
		}
		work_buf += unit_size;
	}
#ifdef TIMER
time_write += GetTickCount() - time_start;
#endif
	print_progress_done();	// sčs̐擪ɖ߂Ă

	// t@CƂɃubN CRC-32 ؂
	j = 0;
	while (j < source_num){
		last_file = s_blk[j].file;
		read_num = (int)((files[last_file].size + (__int64)block_size - 1) / block_size);
		memset(hash, 0, 16);
		for (i = 0; i < read_num; i++)	// XOR  16oCgɌ炷
			((unsigned int *)hash)[i & 3] ^= s_blk[j + i].crc;
		if (memcmp(files[last_file].hash, hash, 16) != 0){
			printf("checksum mismatch, input file %d\n", last_file);
			err = 1;
			goto error_end;
		}
		j += read_num;
	}

#ifdef TIMER
printf("read   %d.%03d sec\n", time_read / 1000, time_read % 1000);
printf("write  %d.%03d sec\n", time_write / 1000, time_write % 1000);
#else
	info_OpenCL(buf, MEM_UNIT);	// foCX\
#endif

error_end:
	InterlockedExchange(&(th->now), INT_MAX / 2);	// TuEXbȟvZ𒆒f
	for (j = 0; j < cpu_num1; j++){
		if (hSub[j]){	// TuEXbhI
			SetEvent(hRun[j]);
			WaitForSingleObject(hSub[j], INFINITE);
			CloseHandle(hSub[j]);
		}
	}
	if (hFile)
		CloseHandle(hFile);
	if (buf)
		_aligned_free(buf);
	i = free_OpenCL();
	if (i != 0)
		printf("free_OpenCL, %d, %d", i & 0xFF, i >> 8);
	return err;
}

